classdef getSimulinkMapping < handle
% autosar.api.getSimulinkMapping API class.
%
% MAPPINGOBJ = autosar.api.getSimulinkMapping(MDLNAME) creates an object
% MAPPINGOBJ that can be used to query and set the Simulink mapping of
% Simulink model MDLNAME
% getSimulinkMapping methods:
%    getDataTransfer   - get the mapping information for a Simulink data transfer line
%    getFunction       - get the mapping information for a Simulink entry point function
%    getInport         - get the mapping information for a Simulink inport
%    getOutport        - get the mapping information for a Simulink outport
%    getFunctionCaller - get the mapping information for a Simulink function caller block
%    getLookupTable    - get the mapping information for a Simulink lookup table
%    getDataDefaults   - get the memory type configured for Signals or Discrete States
%    mapDataTransfer   - map a Simulink data transfer line
%    mapFunction       - map a Simulink entry point function
%    mapInport         - map an inport
%    mapOutport        - map an outport
%    mapFunctionCaller - map a function caller
%    mapLookupTable    - map a lookup table
%    mapDataDefaults    - map a memory type for Signals or Discrete States
%
% See also autosar.api.syncModel autosar.api.create autosar.api.getAUTOSARProperties
    
%   Copyright 2012-2017 The MathWorks, Inc.
    
    properties(Access=private, Transient=true)
        ModelName;
        ChangeLogger;
    end
    
    properties(Constant, Access = private)
        ValidReceiverDAMs = {'ImplicitReceive'
                             'ExplicitReceive'
                             'ExplicitReceiveByVal'
                             'QueuedExplicitReceive'
                             'ErrorStatus'
                             'IsUpdated'
                             'EndToEndRead'
                             'EndToEndQueuedReceive'};
        ValidSenderDAMs   = {'ImplicitSend'
                             'ImplicitSendByRef'
                             'ExplicitSend'
                             'EndToEndWrite'
                             'QueuedExplicitSend'
                             'EndToEndQueuedSend'};
        ValidNvReceiverDAMs = {'ImplicitReceive'
                               'ExplicitReceive'
                               'ExplicitReceiveByVal'};
        ValidNvSenderDAMs   = {'ImplicitSend'
                               'ImplicitSendByRef'
                               'ExplicitSend'};
        ValidParameterAccessModes = {'PortParameter'
                               'PerInstance'
                               'Shared'
                               'ConstantMemory'};
        ValidDataDefaultsData = {'Signals'
                                'States'};
        ValidDataDefaultsMemory = {'None'
                               'ArTypedPIM'};
    end
    
    
    methods(Access=public)
        function this = getSimulinkMapping(modelName, changeLogger)
            this.ModelName = modelName;
            
            if nargin>1
                this.ChangeLogger = changeLogger;
            else
                this.ChangeLogger = [];
            end
        end
           
        function [ARIRVName, ARDataAccessMode] = getDataTransfer(this, SLDataTransferName)
           % [ARIRVNAME, ARDATAACCESSMODE] = GETDATATRANSFER(THIS, SLDATATRANSFERNAME) are
           % the corresponding AUTOSAR InterRunnableVariable ARIRVNAME and
           % DataAccessMode ARDATAACCESSMODE values mapped to the Simulink data
           % transfer line SLDATATRANSFERNAME
           %
           % See also mapDataTransfer

           if ~(ischar(SLDataTransferName) || isStringScalar(SLDataTransferName))
               DAStudio.error('autosarstandard:validation:invalidDataTransferName');
           end
           
           try
               modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
           catch Me
               %Trim stack to suppress mapping internal functions
               Me.throwAsCaller();
           end
           findDataTransfer = false;
           %check whether the data transfer is mapped from Implicit
           %RTBs/signal lines
           SLDataTransfer = modelMapping.DataTransfers.findobj('SignalName', SLDataTransferName);
           if length(SLDataTransfer) == 1   % found data transfer from the signal line mapping
               findDataTransfer = true;
           else
               if ~autosar.api.getSimulinkMapping.isValidRateTransitionBlock(SLDataTransferName)
                   DAStudio.error('RTW:autosar:invalidMappingDataTransfer', SLDataTransferName);
               end
               SLDataTransfer = modelMapping.RateTransition.findobj('Block',SLDataTransferName);
               
               if length(SLDataTransfer) == 1   % found data transfer from the rtb mapping 
                   findDataTransfer = true;
               end
           end
           
           % no Data Transfer mapping is found
           if ~findDataTransfer
                DAStudio.error('RTW:autosar:invalidMappingDataTransfer', SLDataTransferName);
           end
           
           ARIRVName = SLDataTransfer.MappedTo.IrvName;
           ARDataAccessMode = SLDataTransfer.MappedTo.IrvAccessMode; 
        end
        
        function ARRunnableName = getFunction(this, SLFcnName)
        % ARRUNNABLENAME = GETFUNCTION(THIS, SLFCNNAME) is the AUTOSAR 
        % Runnable name ARRUNNABLENAME mapped to Simulink entry point
        % function or Simulink function SLFCNNAME.
        %
        % See also mapFunction
           
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to suppress mapping internal functions
                Me.throwAsCaller();
            end
            SLFcnName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLFcnName);
            isExportStyle = autosar.validation.ExportFcnValidator.isTopModelExportFcn(this.ModelName);
            
            isMultiTasking = length(modelMapping.StepFunctions) > 1;
            
            switch SLFcnName
                case 'InitializeFunction'
                    SLFunction = modelMapping.InitializeFunctions(1);
                case 'TerminateFunction'
                    SLFunction = modelMapping.TerminateFunctions(1);
                case 'StepFunction'
                    if isExportStyle
                        DAStudio.error('autosarstandard:validation:invalidMappingFcnForExportFcnModel', SLFcnName);
                    else
                        SLFunction = modelMapping.StepFunctions(1);
                    end
                otherwise
                    resetFunction = autosar.api.getSimulinkMapping.getResetFunction(modelMapping, SLFcnName);
                    if ~isempty(resetFunction)
                        SLFunction = resetFunction;
                    else
                        SLFunction = modelMapping.FcnCallInports.findobj( ...
                            'Block', [modelMapping.Name, '/', SLFcnName]);
                        
                        if ~isMultiTasking && length(SLFunction) ~= 1
                            SLBlock = autosar.api.Utils.getBlockNameFromFunctionName(this.ModelName,...
                                SLFcnName, false);
                            SLFunction = modelMapping.ServerFunctions.findobj('Block', SLBlock);
                        end
                        if isMultiTasking && length(SLFunction) ~= 1
                            tid = regexp(SLFcnName,'^StepFunction(\d+)','tokens');
                            if ~isempty(tid)
                                tid = str2double(tid{1});
                                if ((tid >= 1) && ...
                                        (tid < length(modelMapping.StepFunctions)) )
                                    SLFunction = modelMapping.StepFunctions(tid+1);
                                else
                                    DAStudio.error('autosarstandard:validation:invalidMappingFcnForRateBasedModel', SLFcnName);
                                end
                            else
                                DAStudio.error('autosarstandard:validation:invalidMappingFcnForRateBasedModel', SLFcnName);
                            end
                        end
                        if length(SLFunction) ~= 1
                            DAStudio.error('autosarstandard:validation:invalidMappingFcnForExportFcnModel', SLFcnName);
                        end
                    end
            end
            
            ARRunnableName = SLFunction.MappedTo.Runnable;
        end
        
        
        
        function [ARPortName, ARElementName, ARDataAccessMode] = getInport(this, SLPortName)
           % [ARPORTNAME, ARDATAELEMENTNAME, ARDATAACCESSMODE] = GETINPORT(THIS,
           % SLPORTNAME) are the corresponding AUTOSAR Receiver (or
           % SenderReceiver) port ARPORTNAME, DataElement ARDATAELEMENTNAME and
           % DataAccessMode ARDATAACCESSMODE values mapped to Simulink Inport
           % SLPORTNAME, SLPORTNAME should be the one given by the
           % function get_param(block, 'Name') to ensure no errors
           %
           % See also mapInport
           
               try
                   modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
               catch Me
                   %Trim stack to suppress mapping internal functions
                   Me.throwAsCaller();
               end
               SLPortName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLPortName);
               
               SLPort = modelMapping.Inports.findobj('Block', [modelMapping.Name, '/', SLPortName]);
           
               if length(SLPort) ~= 1
                   DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'Simulink Port', SLPortName);
               end

               ARPortName = SLPort.MappedTo.Port;
               ARElementName = SLPort.MappedTo.Element;
               ARDataAccessMode = SLPort.MappedTo.DataAccessMode;
           
           end
        
         function [ARPortName, ARElementName, ARDataAccessMode] = getOutport(this, SLPortName)
           % [ARPORTNAME, ARDATAELEMENTNAME, ARDATAACCESSMODE] = GETOUTPORT(THIS,
           % SLPORTNAME) are the corresponding AUTOSAR Sender (or
           % SenderReceiver) port ARPORTNAME, DataElement ARDATAELEMENTNAME and
           % DataAccessMode ARDATAACCESSMODE values mapped to Simulink Outport
           % SLPORTNAME, SLPORTNAME should be the one given by the
           % function get_param(block, 'Name') to ensure no errors
           %
           % See also mapOutport

               try
                   modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
               catch Me
                   %Trim stack to suppress mapping internal functions
                   Me.throwAsCaller();
               end
               SLPortName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLPortName);
               SLPort = modelMapping.Outports.findobj('Block', [modelMapping.Name, '/', SLPortName]);
           
               if length(SLPort) ~= 1
                   DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'Simulink Port', SLPortName);
               end

               ARPortName = SLPort.MappedTo.Port;
               ARElementName = SLPort.MappedTo.Element;
               ARDataAccessMode = SLPort.MappedTo.DataAccessMode;
                          
         end
         
         function [ARPortName, AROperationName] = getFunctionCaller(this, SLFcnName)
             % [ARPORTNAME, AROPERATIONNAME] = GETFUNCTIONCALLER(THIS,
             % SLFCNNAME) are the corresponding AUTOSAR Client
             % port ARPORTNAME, Operation AROPERATIONNAME values
             % mapped to Simulink function caller SLFCNNAME.
             % 
             % See also mapFunctionCaller
             
             try
                 modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
             catch Me
                 %Trim stack to suppress mapping internal functions
                 Me.throwAsCaller();
             end
             SLFcnName = strrep(SLFcnName, newline, ' ');
             SLBlock = autosar.api.Utils.getBlockNameFromFunctionName(this.ModelName,...
                 SLFcnName, true);
             SLPort = modelMapping.FunctionCallers.findobj('Block', SLBlock);
             
             if length(SLPort) ~= 1
                 DAStudio.error('RTW:autosar:invalidFunctionCallerBlock', SLFcnName);
             end
             
             ARPortName = SLPort.MappedTo.ClientPort;
             AROperationName = SLPort.MappedTo.Operation;
             
         end
         
         function ARMemoryType = getDataDefaults(this, SLData) 
         % ARMEMORYTYPE = GETDATADEFAULTS(THIS, SLData) returns the AUTOSAR 
         % memory type set for the Simulink data SLData('Signals' or 'States'). 
         % The returned AUTOSAR memory type is either 'ArTypedPIM' or 'None', 
         % which only applies to signals or states whose storage classes are 
         % configured to SimulinkGlobal.
             
             try
                 modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
             catch Me
                 %Trim stack to suppress mapping internal functions
                 Me.throwAsCaller();
             end
             switch SLData
                 case 'Signals'
                    ARMemoryType = modelMapping.DataDefaultsMapping.Signals.Memory;
                 case 'States'
                    ARMemoryType = modelMapping.DataDefaultsMapping.DiscreteStates.Memory;
                 otherwise
                    DAStudio.error('autosarstandard:api:invalidDataDefaultsData', SLData);
             end
         end 

         
        function mapDataTransfer(this, SLDataTransferName, ARIRVName, ARDataAccessMode)
        % MAPDATATRANSFER(THIS, SLDATATRANSFERNAME, ARIRVNAME, ARDATAACCESSMODE) maps
        % Simulink data transfer line SLDATATRANSFERNAME to
        % InterRunnableVariable ARIRVNAME. DataAccessMODE ARDATAACCESSMODE
        % can be one of 'Implicit', 'Explicit'
        %
        % See also getDataTransfer
           
            if ~(ischar(SLDataTransferName)|| isStringScalar(SLDataTransferName))
                DAStudio.error('autosarstandard:validation:invalidDataTransferName');
            end
        
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                dataObj = autosar.api.getAUTOSARProperties(this.ModelName, true);
            catch Me
                %Trim stack to suppress internal mapping functions
                Me.throwAsCaller();
            end
            findSignal = false;
            findRTB = false;
            
            % check the data transfers mapped from signal lines/hidden
            % RTBs.
            SLDataTransfer = modelMapping.DataTransfers.findobj('SignalName', SLDataTransferName);
            if length(SLDataTransfer) == 1  % found map from signal lines/Hidden RTBs
                findSignal = true;
                
                % check IRV path
                componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
                ARIRVPaths = dataObj.find(componentQualifiedName, 'IrvData', ...
                    'Name', ARIRVName, 'PathType', 'FullyQualified');
                if length(ARIRVPaths) ~= 1
                    DAStudio.error('RTW:autosar:invalidMappingDataTransferIrv', ...
                        SLDataTransferName, ARIRVName);
                end
            else % check the data transfers mapped from explicit RTBs
                if ~autosar.api.getSimulinkMapping.isValidRateTransitionBlock(SLDataTransferName)
                    DAStudio.error('RTW:autosar:invalidMappingDataTransfer', SLDataTransferName);
                end
                
                if ~(strcmp(get_param(SLDataTransferName,'Integrity'), 'on') && ...
                       strcmp(get_param(SLDataTransferName,'Deterministic'), 'off'))
                   DAStudio.error(...
                       'autosarstandard:validation:RateTranBlkForIRV',...
                       SLDataTransferName);
                end
                
                SLDataTransfer = modelMapping.RateTransition.findobj('Block',SLDataTransferName);
                
                if length(SLDataTransfer) == 1  % found map from explicit RTBs
                    findRTB = true;
                    
                    % check IRV path
                    componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
                    ARIRVPaths = dataObj.find(componentQualifiedName, 'IrvData', ...
                        'Name', ARIRVName, 'PathType', 'FullyQualified');
                    if length(ARIRVPaths) ~= 1
                        DAStudio.error('autosarstandard:validation:IRVNotExistForDataTransfer', ...
                            SLDataTransferName, ARIRVName);
                    end
                end 
            end
            
            if ~findSignal && ~findRTB
                DAStudio.error('RTW:autosar:invalidMappingDataTransfer', SLDataTransferName);
            end
            
            if ~isempty(this.ChangeLogger)
                if ~strcmp(SLDataTransfer.MappedTo.IrvName, ARIRVName) &&...
                        ~isempty(SLDataTransfer.MappedTo.IrvName)
                    this.ChangeLogger.logModification('Automatic', 'IRVName', ...
                        'Simulink data transfer', ...
                        SLDataTransferName, ...
                        SLDataTransfer.MappedTo.IrvName, ...
                        ARIRVName);
                end
                    
                if ~strcmp(SLDataTransfer.MappedTo.IrvAccessMode, ARDataAccessMode)
                    this.ChangeLogger.logModification('Automatic', 'DataAccessMode', ...
                        'Simulink data transfer', ...
                        SLDataTransferName, ...
                        SLDataTransfer.MappedTo.IrvAccessMode, ...
                        ARDataAccessMode);
                end
            end
            
            if findSignal
                SLDataTransfer.mapInterRunnableVariable(SLDataTransferName, ARIRVName, ARDataAccessMode, '');   
            else
                SLDataTransfer.mapInterRunnableVariable(ARIRVName, ARDataAccessMode, ''); 
            end
            
        end
        
        function mapInport(this, SLPortName, ARPortName, ARDataName, ARDataAccessModeStr)
            % MAPINPORT(THIS, SLPORTNAME, ARPORTNAME, ARDATAELEMENTNAME, ARDATAACCESSMODE)
            % maps Simulink Inport SLPORTNAME to DataElement ARDATAELEMENTNAME
            % of AUTOSAR Receiver (or SenderReceiver) port
            % ARPORTNAME. SLPORTNAME should be the one given by the
            % function get_param(block, 'Name') to ensure no errors
            % DataAccessMode ARDATAACCESSMODESTR can be one of
            % 'ImplicitReceive', 'ExplicitReceive', 'ExplicitReceiveByVal', 
            % 'QueuedExplicitReceive', 'ErrorStatus', 'IsUpdated', 
            % 'EndToEndRead'.
            %
            % See also getInport
            
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                dataObj = autosar.api.getAUTOSARProperties(this.ModelName, true);
            catch Me
                %Trim stack to suppress internal mapping functions
                Me.throwAsCaller();
            end
            SLPortName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLPortName);
            SLPort = modelMapping.Inports.findobj('Block', [modelMapping.Name, '/', SLPortName]);
           
            % SLPort Check
            if length(SLPort) ~= 1
                    DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'Simulink Port', SLPortName);
            end

            % ARPort Check
            if strcmp(ARDataAccessModeStr, 'ModeReceive')
                portType = 'ModeReceiverPort';
            else
                portType = 'DataReceiverPort';
            end
            componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
            ARPortPaths = dataObj.find(componentQualifiedName, portType, ...
                 'Name', ARPortName, 'PathType', 'FullyQualified');
            
            % May be trying to map a SenderReceiverPort
            if isempty(ARPortPaths) && strcmp(portType, 'DataReceiverPort')
                 ARPortPaths = dataObj.find(componentQualifiedName, 'DataSenderReceiverPort', ...
                 'Name', ARPortName, 'PathType', 'FullyQualified');
                 if ~isempty(ARPortPaths)
                     % Found a DataSenderReceiverPort
                     portType = 'DataSenderReceiverPort';
                 end
            end
                
            % May be trying to map a NvDataReceiverPort
            if isempty(ARPortPaths) && strcmp(portType, 'DataReceiverPort')
                 ARPortPaths = dataObj.find(componentQualifiedName, 'NvDataReceiverPort', ...
                 'Name', ARPortName, 'PathType', 'FullyQualified');
                 if ~isempty(ARPortPaths)
                     % Found a NvDataReceiverPort
                     portType = 'NvDataReceiverPort';
                 end
            end            
            
            % May be trying to map a NvDataSenderReceiverPort
            if isempty(ARPortPaths) && strcmp(portType, 'DataReceiverPort')
                 ARPortPaths = dataObj.find(componentQualifiedName, 'NvDataSenderReceiverPort', ...
                 'Name', ARPortName, 'PathType', 'FullyQualified');
                 if ~isempty(ARPortPaths)
                     % Found a NvDataSenderReceiverPort
                     portType = 'NvDataSenderReceiverPort';
                 end
            end            
            
            if isempty(ARPortPaths) && strcmp(portType, 'ModeReceiverPort')
                % After 3.x import SenderReceiverInterface referenced by a DataReceiverPort may have ModeDeclarationGroupElement
                ARPortPaths = dataObj.find(componentQualifiedName, 'DataReceiverPort', ...
                    'Name', ARPortName, 'PathType', 'FullyQualified');
            end
            
            if length(ARPortPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, portType, ARPortName);
            end
            
            % ARDataElement Check           
            ifPath = dataObj.get(ARPortPaths{1}, 'Interface', 'PathType', 'FullyQualified');
            if strcmp(ARDataAccessModeStr, 'ModeReceive')
                metaModelClassName = 'ModeDeclarationGroupElement';
            else
                metaModelClassName = 'FlowData';
            end
            ARDataElementPaths = dataObj.find(ifPath, metaModelClassName, ...
                'Name', ARDataName, 'PathType', 'FullyQualified');
            if isempty(ifPath) || length(ARDataElementPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'DataElement', ARDataName);
            end
            
            switch portType
              case 'DataReceiverPort'
                validDataAccessModes = this.ValidReceiverDAMs;
              case 'DataSenderReceiverPort'
                validDataAccessModes = {'ImplicitReceive', ...
                                    'ExplicitReceive', ...
                                    'ExplicitReceiveByVal', ...
                                    'ErrorStatus', ...
                                    'IsUpdated', ...
                                    'EndToEndRead'};
              case {'NvDataReceiverPort', 'NvDataSenderReceiverPort'}
                validDataAccessModes = this.ValidNvReceiverDAMs;
              case 'ModeReceiverPort'
                validDataAccessModes = {'ModeReceive'};
              otherwise
                assert(false, 'Did not recognize portType %s', portType);
            end
                
            
            % ARDataAccessModeStr Check
            if ~any(strcmp(ARDataAccessModeStr, validDataAccessModes))
                validDataAccessModesStr = autosar.api.Utils.cell2str(validDataAccessModes);
                DAStudio.error('RTW:autosar:invalidMappingDAM', ARDataAccessModeStr, SLPortName, validDataAccessModesStr);
            end

            if ~isempty(this.ChangeLogger)
                if ~strcmp(SLPort.MappedTo.Port, ARPortName)
                    this.ChangeLogger.logModification('Automatic', 'ARPortName', ...
                                                      'Simulink inport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.Port, ...
                                                      ARPortName);
                end
                    
                if ~strcmp(SLPort.MappedTo.Element, ARDataName)
                    this.ChangeLogger.logModification('Automatic', 'ARElementName', ...
                                                      'Simulink inport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.Element, ...
                                                      ARDataName);
                end
                
                if ~strcmp(SLPort.MappedTo.DataAccessMode, ARDataAccessModeStr)
                    this.ChangeLogger.logModification('Automatic', 'DataAccessMode', ...
                                                      'Simulink inport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.DataAccessMode, ...
                                                      ARDataAccessModeStr);
                end
            end
            
            % add or Update ComSpec Object according to Data Access Mode
            try
                autosar.mm.sl2mm.ComSpecBuilder.addOrUpdateM3IComSpec(ARPortName, ...
                    ARDataName, ARDataAccessModeStr, this.ModelName);
                
                 SLPort.mapPortElement(ARPortName, ARDataName, ARDataAccessModeStr, '');
            catch Me
                %Trim stack to suppress internal comSpec mapping functions
                Me.throwAsCaller();
            end


            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        function mapOutport(this, SLPortName, ARPortName, ARDataName, ARDataAccessModeStr)
            % MAPOUTPORT(THIS, SLPORTNAME, ARPORTNAME, ARDATAELEMENTNAME, ARDATAACCESSMODE)
            % maps Simulink Outport SLPORTNAME to DataElement ARDATAELEMENTNAME
            % of AUTOSAR Sender (or SenderReceiver) port
            % ARPORTNAME. SLPORTNAME should be the one given by the
            % function get_param(block, 'Name') to ensure no errors
            % DataAccessMode ARDATAACCESSMODESTR can be one of
            % 'ImplicitSend', 'ImplicitSendByRef', 'ExplicitSend', 
            % 'EndToEndWrite'.
            %
            % See also getOutport
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                dataObj = autosar.api.getAUTOSARProperties(this.ModelName, true);
            catch Me
                %Trim stack to suppress internal mapping functions
                Me.throwAsCaller();
            end
            SLPortName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLPortName);
            SLPort = modelMapping.Outports.findobj('Block', [modelMapping.Name, '/', SLPortName]);
            
            if length(SLPort) ~= 1                  
                DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'Simulink Port', SLPortName);               
            end

            if strcmp(ARDataAccessModeStr, 'ModeSend')
                portType = 'ModeSenderPort';
            else
                portType = 'DataSenderPort';
            end
            
            % ARPort Check
            componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
            ARPortPaths = dataObj.find(componentQualifiedName, portType, ...
                'Name', ARPortName, 'PathType', 'FullyQualified');
            
            % May be trying to map a SenderReceiverPort
            if isempty(ARPortPaths)
                 ARPortPaths = dataObj.find(componentQualifiedName, 'DataSenderReceiverPort', ...
                 'Name', ARPortName, 'PathType', 'FullyQualified');
            end
            
            % May be trying to map a NvDataSenderPort
            if isempty(ARPortPaths)
                 ARPortPaths = dataObj.find(componentQualifiedName, 'NvDataSenderPort', ...
                 'Name', ARPortName, 'PathType', 'FullyQualified');
                 if ~isempty(ARPortPaths)
                     % Found a NvDataSenderPort
                     portType = 'NvDataSenderPort';
                 end
            end     
            
            % May be trying to map a NvDataSenderReceiverPort
            if isempty(ARPortPaths)
                 ARPortPaths = dataObj.find(componentQualifiedName, 'NvDataSenderReceiverPort', ...
                 'Name', ARPortName, 'PathType', 'FullyQualified');
                 if ~isempty(ARPortPaths)
                     % Found a NvDataSenderPort
                     portType = 'NvDataSenderPort';
                 end
            end
            
            if isempty(ARPortPaths) && strcmp(portType, 'ModeSenderPort')
                % After 3.x import SenderReceiverInterface referenced by
                % a DataSenderPort may have ModeDeclarationGroupElement??
                
                ARPortPaths = dataObj.find(componentQualifiedName, 'DataSenderPort', ...
                    'Name', ARPortName, 'PathType', 'FullyQualified');
            end
            
            if length(ARPortPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'DataSenderPort', ARPortName);
            end
            
            % ARDataElement Check           
            ifPath = dataObj.get(ARPortPaths{1}, 'Interface', 'PathType', 'FullyQualified');
            if strcmp(ARDataAccessModeStr, 'ModeSend')
                metaModelClassName = 'ModeDeclarationGroupElement';
            else
                metaModelClassName = 'FlowData';
            end            
            ARDataElementPaths = dataObj.find(ifPath, metaModelClassName, ...
                'Name', ARDataName, 'PathType', 'FullyQualified');
            if isempty(ifPath) || length(ARDataElementPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'DataElement', ARDataName);
            end
            
            % ARDataAccessModeStr Check
            switch portType
              case 'DataSenderPort'
                validDataAccessModes = this.ValidSenderDAMs;
              case 'NvDataSenderPort'
                validDataAccessModes = this.ValidNvSenderDAMs;
              case 'ModeSenderPort'
                validDataAccessModes = {'ModeSend'};                
              otherwise
                assert(false, 'Did not recognize portType %s', portType);
            end
            
            
            % ARDataAccessModeStr Check
            if ~any(strcmp(ARDataAccessModeStr, validDataAccessModes))
                validDataAccessModesStr = autosar.api.Utils.cell2str(validDataAccessModes);
                DAStudio.error('RTW:autosar:invalidMappingDAM', ARDataAccessModeStr, SLPortName, validDataAccessModesStr);
            end
            
            if ~isempty(this.ChangeLogger)
                if ~strcmp(SLPort.MappedTo.Port, ARPortName)
                    this.ChangeLogger.logModification('Automatic', 'ARPortName', ...
                                                      'Simulink outport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.Port, ...
                                                      ARPortName);
                end
                    
                if ~strcmp(SLPort.MappedTo.Element, ARDataName)
                    this.ChangeLogger.logModification('Automatic', 'ARElementName', ...
                                                      'Simulink outport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.Element, ...
                                                      ARDataName);
                end
                
                % ------- Keeping Implicit send selection on update -------
                % In the AUTOSAR ARXML both ImplicitSend and 
                % ImplicitSendByRef are the expressed in the same way. When
                % updating we want to keep the previous user selection in
                % this case.
                implicitSendAccessModes = {'ImplicitSend', 'ImplicitSendByRef'};
                currIsImplicitSendAm = ismember(SLPort.MappedTo.DataAccessMode, ...
                                                implicitSendAccessModes);
                newIsImplicitSendAm = ismember(ARDataAccessModeStr, ...
                                               implicitSendAccessModes);
                if (currIsImplicitSendAm && newIsImplicitSendAm)
                    ARDataAccessModeStr = SLPort.MappedTo.DataAccessMode;
                end
                % ---------------------------------------------------------
                
                if ~strcmp(SLPort.MappedTo.DataAccessMode, ARDataAccessModeStr)
                    this.ChangeLogger.logModification('Automatic', 'DataAccessMode', ...
                                                      'Simulink outport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.DataAccessMode, ...
                                                      ARDataAccessModeStr);
                end
            end
            
            try
                % add or Update ComSpec Object according to Data Access Mode
                autosar.mm.sl2mm.ComSpecBuilder.addOrUpdateM3IComSpec(ARPortName, ...
                    ARDataName, ARDataAccessModeStr, this.ModelName);
                
                SLPort.mapPortElement(ARPortName, ARDataName, ARDataAccessModeStr, '');
            catch Me
                %Trim stack to suppress internal ComSpec mapping functions
                Me.throwAsCaller();
            end
            
            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        
        function mapFunction(this, SLFcnName, ARRunnableName)
            % MAPFUNCTION(THIS, SLFCNNAME, ARRUNNABLENAME) maps Simulink entry point
            % function or Simulink function SLFCNNAME to AUTOSAR Runnable 
            % ARRUNNABLENAME.
            %
            % See also getFunction
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to suppress mapping internal functions
                Me.throwAsCaller();
            end
            % ARRunnableName cannot be empty.
            if isempty(ARRunnableName)
                DAStudio.error('RTW:autosar:invalidRunnableName',...
                    ARRunnableName);
            end
            
            SLFcnName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLFcnName);
            isExportStyle = autosar.validation.ExportFcnValidator.isTopModelExportFcn(this.ModelName);
            isMultiTasking = length(modelMapping.StepFunctions) > 1;
            
            mappingObj = [];
            switch SLFcnName
                case 'InitializeFunction'
                    mappingObj = modelMapping.InitializeFunctions(1);
                case 'TerminateFunction'
                    mappingObj = modelMapping.TerminateFunctions(1);
                case 'StepFunction'
                    if ~isExportStyle
                        mappingObj = modelMapping.StepFunctions(1);
                    end
                otherwise
                    SLFcnCallInport = modelMapping.FcnCallInports.findobj( ...
                        'Block', [modelMapping.Name, '/', SLFcnName]);
                    if length(SLFcnCallInport) == 1
                        mappingObj = SLFcnCallInport;
                    end
                    if ~isMultiTasking && isempty(mappingObj)
                        SLBlock = autosar.api.Utils.getBlockNameFromFunctionName(this.ModelName,...
                            SLFcnName, false);
                        SLServerFunction = modelMapping.ServerFunctions.findobj( ...
                            'Block', SLBlock);
                        if length(SLServerFunction) == 1
                            mappingObj = SLServerFunction;
                        end
                    end
                    if isempty(mappingObj)
                        mappingObj = autosar.api.getSimulinkMapping.getResetFunction(modelMapping, SLFcnName);
                    end
                    if isMultiTasking && isempty(mappingObj)
                        tid = regexp(SLFcnName,'^StepFunction(\d+)','tokens');
                        if ~isempty(tid)
                            tid = str2double(tid{1});
                            if ((tid >= 1) && ...
                                    (tid < length(modelMapping.StepFunctions)) )
                                mappingObj = modelMapping.StepFunctions(tid+1);
                            else
                                DAStudio.error('autosarstandard:validation:invalidMappingFcnForRateBasedModel', SLFcnName);
                            end
                        else
                            DAStudio.error('autosarstandard:validation:invalidMappingFcnForRateBasedModel', SLFcnName);
                        end
                    end
            end
            
            if isempty(mappingObj)
                DAStudio.error('autosarstandard:validation:invalidMappingFcnForExportFcnModel', SLFcnName);
            end
            
            try
                autosar.api.Utils.mapFunction(this.ModelName,...
                    mappingObj, ARRunnableName);
            catch Me
                %Trim stack to suppress internal functions
                Me.throwAsCaller();
            end
            
            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        function mapFunctionCaller(this, SLFcnName, ARPortName, AROperationName)
            % MAPFUNCTIONCALLER(THIS, SLFCNNAME, ARPORTNAME, AROPERATIONNAME)
            % maps Simulink Function caller SLFCNNAME to Operation AROPERATIONNAME
            % of AUTOSAR Client port ARPORTNAME.
            %
            % See also getFunctionCaller
            
            % convert strings to chars
            if isstring(ARPortName)
                ARPortName = convertStringsToChars(ARPortName);
            end
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                dataObj = autosar.api.getAUTOSARProperties(this.ModelName, true);
            catch Me
                %Trim stack to suppress mapping internal functions
                Me.throwAsCaller();
            end
                       
            SLFcnName = strrep(SLFcnName, newline, ' ');
            SLBlock = autosar.api.Utils.getBlockNameFromFunctionName(this.ModelName,...
                            SLFcnName, true);
            SLBlockMap = modelMapping.FunctionCallers.findobj('Block', SLBlock);              
            
            % SLBlockMap Check
            if length(SLBlockMap) ~= 1
                DAStudio.error('RTW:autosar:invalidCallerMapping', SLFcnName, 'function caller', SLFcnName);
            end

            % ARPort Check
            portType = 'ClientPort';
            componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
            ARPortPaths = dataObj.find(componentQualifiedName, portType, ...
                 'Name', ARPortName, 'PathType', 'FullyQualified');            
            
            if length(ARPortPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidCallerMapping', SLFcnName, portType, ARPortName);
            end
            
            % AROperationName Check           
            ifPath = dataObj.get(ARPortPaths{1}, 'Interface', 'PathType', 'FullyQualified');
            metaModelClassName = 'Operation';
            AROperationPaths = dataObj.find(ifPath, metaModelClassName, ...
                'Name', AROperationName, 'PathType', 'FullyQualified');
            if isempty(ifPath) || length(AROperationPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidCallerMapping', SLFcnName, 'Operation', AROperationName);
            end                
            

            if ~isempty(this.ChangeLogger)
                if ~strcmp(SLBlockMap.MappedTo.ClientPort, ARPortName)
                    this.ChangeLogger.logModification('Automatic', 'ARPortName', ...
                                                      'Simulink Function Caller block', ...
                                                      SLBlockMap.Block, ...
                                                      SLBlockMap.MappedTo.ClientPort, ...
                                                      ARPortName);
                end
                    
                if ~strcmp(SLBlockMap.MappedTo.Operation, AROperationName)
                    this.ChangeLogger.logModification('Automatic', 'AROperationName', ...
                                                      'Simulink Function Caller block', ...
                                                      SLBlockMap.Block, ...
                                                      SLBlockMap.MappedTo.Operation, ...
                                                      AROperationName);
                end
            end
            for index=1:length(modelMapping.FunctionCallers)
                if strcmp(get_param(modelMapping.FunctionCallers(index).Block, 'FunctionPrototype'),...
                        get_param(SLBlockMap.Block, 'FunctionPrototype'))
                    % Call function in utils to do the mapping
                    try
                        autosar.api.Utils.mapCaller(this.ModelName, modelMapping.FunctionCallers(index), ...
                            ARPortName, AROperationName);
                    catch Me
                        Me.throwAsCaller();
                    end
                end
            end

            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        function [ARParameterAccessMode, ARPortName, ARParameterData] = getLookupTable(this, SLParam)
            % [ARPARAMETERACCESSMODE, ARPORTNAME, ARPARAMETERDATA] = GETLOOKUPTABLE(THIS,
            % SLPARAM) are the corresponding ParameterAccessMode
            % ARPARAMETERACCESSMODE, AUTOSAR Parameter Receiver port
            % ARPORTNAME and Parameter Data ARPARAMETERDATA values mapped
            % to Simulink Lookup Table SLPARAM.
            %
            % See also mapLookupTable
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to remove internal mapping functions
                Me.throwAsCaller();
            end
            SLLut = modelMapping.LookupTables.findobj('LookupTableName', SLParam);
            
            if length(SLLut) ~= 1
                DAStudio.error('autosarstandard:api:invalidMapping', 'Lookup Table',...
                    SLParam, 'Simulink Lookup Table', SLParam);
            end
            
            ARParameterAccessMode = SLLut.MappedTo.ParameterAccessMode;
            ARPortName = SLLut.MappedTo.Port;
            ARParameterData = SLLut.MappedTo.Parameter;
        end
        
        function mapLookupTable(this, SLParam, ARParameterAccessMode, ARPortName, ARParameterData)
            % MAPLOOKUPTABLE(THIS, SLPARAM, ARPARAMETERACCESSMODE, ARPORTNAME, ARPARAMETERDATA)
            % maps Simulink Lookup Table SLPARAM to ParameterAccessMode
            % ARPARAMETERACCESSMODE, AUTOSAR Parameter Receiver Port
            % ARPORTNAME and Parameter Data ARPARAMETERDATA.
            % ParameterAccessMode ARPARAMETERACCESSMODE can be one of
            % PortParameter, Shared, PerInstance and Const.
            %
            % Examples: 
            % Map Simulink lookup table L_4_single to AUTOSAR Parameter
            % with ParameterAccessMode Shared.
            %
            % mappingObj = autosar.api.getSimulinkMapping(modelName);
            % mappingObj.mapLookupTable('L_4_single', 'Shared', '',...
            %   'L_4_single_ar');
            %
            % Map Simulink lookup table L_4_single to AUTOSAR Parameter
            % to AUTOSAR Parameter Receiver Port paramRPort and Parameter
            % paramElem1.
            %
            % mappingObj = autosar.api.getSimulinkMapping(modelName);
            % mappingObj.mapLookupTable('L_4_single', 'PortParameter',...
            %   'paramRPort', 'paramElem1');
            %
            % See also getLookupTable
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                dataObj = autosar.api.getAUTOSARProperties(this.ModelName, true);
            catch Me
                %Trim to suppress mapping internal functions
                Me.throwAsCaller();
            end
            
            SLLUTMap = modelMapping.LookupTables.findobj('LookupTableName', SLParam);
            
            % SL LookupTable Check
            if length(SLLUTMap) ~= 1
                DAStudio.error('autosarstandard:api:invalidMapping', 'Lookup Table',...
                    SLParam, 'Simulink Lookup Table', SLParam);
            end
            
            if strcmp(ARParameterAccessMode, 'PortParameter')
                % ARPort Check
                portType = 'ParameterReceiverPort';
                componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
                ARPortPaths = dataObj.find(componentQualifiedName, portType, ...
                    'Name', ARPortName, 'PathType', 'FullyQualified');
                
                if length(ARPortPaths) ~= 1
                    DAStudio.error('autosarstandard:api:invalidMapping', 'Lookup Table', ...
                        SLParam, portType, ARPortName);
                end
                
                % ARParameterName Check
                ifPath = dataObj.get(ARPortPaths{1}, 'Interface', 'PathType', 'FullyQualified');
                metaModelClassName = 'ParameterData';
                ARParameterDataPaths = dataObj.find(ifPath, metaModelClassName, ...
                    'Name', ARParameterData, 'PathType', 'FullyQualified');
                if isempty(ifPath) || length(ARParameterDataPaths) ~= 1
                    DAStudio.error('autosarstandard:api:invalidMapping', 'Lookup Table', ...
                        SLParam, 'Parameter Data', ARParameterData);
                    
                end
            end
            % ARParameterAccessMode Check
            if ~any(strcmp(ARParameterAccessMode, this.ValidParameterAccessModes))
                validParameterDataRolesStr = autosar.api.Utils.cell2str(this.ValidParameterAccessModes);
                DAStudio.error('autosarstandard:api:invalidParameterMappingRole',...
                    ARParameterAccessMode, 'Simulink Lookup Table', SLLUTMap.LookupTableName,...
                    validParameterDataRolesStr);
            end
            SLLUTMap.mapLookupTable(SLLUTMap.LookupTableName, ARParameterAccessMode,...
                ARPortName, ARParameterData, '');
            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        function mapDataDefaults(this, SLData, ARMemoryType) 
        % MAPDATADEFAULTS(THIS, SLData, ARMEMORYTYPE) maps Simulink data
        % SLData('Signals' or 'States') to AUTOSAR memory type ARMemoryType. 
        % The valid AUTOSAR memory type is 'None' or 'ArTypedPIM', and only
        % applies to signals or states whose storage classes are set to
        % SimulinkGlobal.
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to remove internal mapping functions
                Me.throwAsCaller();
            end
            
            if ~any(strcmp(SLData, this.ValidDataDefaultsData))
                DAStudio.error('autosarstandard:api:invalidDataDefaultsData', SLData);
            end
            
            if ~any(strcmp(ARMemoryType, this.ValidDataDefaultsMemory))
                DAStudio.error('autosarstandard:api:invalidDataMemoryType', ARMemoryType, SLData);
            end 
            
            % verify that schema 4.0 is used for ArTypedPIM
            if strcmp(ARMemoryType, 'ArTypedPIM') 
                autosarVer = get_param(this.ModelName, 'AutosarSchemaVersion');
                if (str2double(autosarVer) < 4.0)
                    autosar.mm.util.MessageReporter.createWarning(...
                        'autosarstandard:validation:invalidSchemaForArTypedPIM',...
                        autosarVer, SLData);
                end
            end 
            
            switch SLData
                case 'Signals'
                    modelMapping.DataDefaultsMapping.Signals.Memory = ARMemoryType;
                case 'States'
                    modelMapping.DataDefaultsMapping.DiscreteStates.Memory = ARMemoryType;
                otherwise
                    DAStudio.error('autosarstandard:api:invalidDataDefaultsData', SLData);
            end        
        end 
    end
    
    methods(Hidden)
        
        function mapPort(this, SLPortName, ARPortName, ARDataName, ARDataAccessModeStr)
            % MAPPORT(THIS, SLPORTNAME, ARPORTNAME, ARDATAELEMENTNAME, ARDATAACCESSMODE)
            % maps Simulink port SLPORTNAME to DataElement ARDATAELEMENTNAME
            % of AUTOSAR Sender/Receiver (or SenderReceiver) port
            % ARPORTNAME. SLPORTNAME should be the one given by the
            % function get_param(block, 'Name') to ensure no errors
            % 
            % See also mapInport, mapOutport
            modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            SLPortName2 = autosar.api.getSimulinkMapping.escapeSimulinkName(SLPortName);
            SLInport = modelMapping.Inports.findobj('Block', [modelMapping.Name, '/', SLPortName2]);

            if length(SLInport) == 1
                this.mapInport(SLPortName, ARPortName,...
                    ARDataName, ARDataAccessModeStr);
            else
                this.mapOutport(SLPortName, ARPortName,...
                    ARDataName, ARDataAccessModeStr);
            end
        end
        
        function mapModelBlock(this, SLModelBlockName, ARPrototypeName, uuid)
            % MAPMODELBLOCK(THIS, SLMODELBLOCKNAME, ARPROTOTYPENAME)
            % maps Simulink model block SLMODELBLOCKNAME to component prototype
            % ARPrototypeName            
            
            modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            SLModelBlockName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLModelBlockName);
            SLModelBlock = modelMapping.ModelBlocks.findobj('Block', [modelMapping.Name, '/', SLModelBlockName]);
            
            if ~isempty(this.ChangeLogger)
                if ~isempty(SLModelBlock.MappedTo) &&...
                        ~strcmp(SLModelBlock.MappedTo.PrototypeName, ARPrototypeName)
                    this.ChangeLogger.logModification(...
                        'Automatic', 'ARPrototypeName', ...
                        'Simulink model block', ...
                        SLModelBlock.Block, ...
                        SLModelBlock.MappedTo.PrototypeName, ...
                        ARPrototypeName);
                end
            end
            
            SLModelBlock.mapComponentPrototype(ARPrototypeName, uuid);
            
            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
          
    end
    
    methods(Static, Hidden)
        
        function validNvReceiverDAMs = getValidNvReceiverDAMs()
            % VALIDNVRECEIVERDAMS = GETVALIDNVRECEIVERDAMS returns a cell-array of valid
            % NvReceiverPort Data Access Modes.
            validNvReceiverDAMs = autosar.api.getSimulinkMapping.ValidNvReceiverDAMs;
        end
        
        function validNvSenderDAMs = getValidNvSenderDAMs()
            % VALIDNVSENDERDAMS = GETVALIDNVSENDERDAMS returns a cell-array of valid
            % NvSenderPort Data Access Modes.
            validNvSenderDAMs = autosar.api.getSimulinkMapping.ValidNvSenderDAMs;
        end
        
    end
    
    methods(Static, Access=private)
        function res = escapeSimulinkName(pathStr)
        % Convert new line to space as this is what
        % BlockMapping.getBlockPath() does, and change / into // as this is
        % what getFullName() does when a port name is obtained
        
           %convert strings to chars
           if isstring(pathStr)
               pathStr = convertStringsToChars(pathStr);
           end
           pathStr = strrep(pathStr, newline, ' ');
           res = strrep(pathStr, '/', '//');
        end      
        
        function isValid = isValidRateTransitionBlock(blkPath)
            % Check if input string is a valid rate transition block
            isValid = false;
            try
               if Simulink.ID.isValid(Simulink.ID.getSID(blkPath))
                   if strcmp(get_param(blkPath,'BlockType'), 'RateTransition')
                        isValid = true;                      
                   end
               end
            catch ex %#ok<NASGU>
            end
        end
        
        function resetFunction = getResetFunction(modelMapping, SLResetFcnName)
            for idx = 1:length(modelMapping.ResetFunctions)
                currResetFcn = modelMapping.ResetFunctions(idx);
                if strcmp(SLResetFcnName, currResetFcn.Name)
                    resetFunction = currResetFcn;
                    return;
                end
            end
            resetFunction = [];
            return;
        end        
    end
    
end

% LocalWords:  MAPPINGOBJ MDLNAME ARIRVNAME ARDATAACCESSMODE GETDATATRANSFER Bs
% LocalWords:  SLDATATRANSFERNAME ARRUNNABLENAME GETFUNCTION SLFCNNAME utils
% LocalWords:  ARPORTNAME ARDATAELEMENTNAME GETINPORT SLPORTNAME GETOUTPORT RTB
% LocalWords:  MAPDATATRANSFER MAPINPORT ARDATAACCESSMODESTR MAPOUTPORT MAPPORT
% LocalWords:  MAPFUNCTION AROPERATIONNAME GETFUNCTIONCALLER MAPFUNCTIONCALLER
% LocalWords:  ARCOMPUMETHODNAME SLDATATYPENAME Compu MAPDATATYPE Nv RPort rtb
% LocalWords:  autosarstandard ARLOOKUPTABLENAME ARPARAMETERACCESSMODE GETLOOKUPTABLE
% LocalWords:  ARPARAMETERDATANAME MAPLOOKUPTABLE SLPARAM ARPARAMETERDATA
% LocalWords:  VALIDNVRECEIVERDAMS GETVALIDNVRECEIVERDAMS VALIDNVSENDERDAMS
% LocalWords:  GETVALIDNVSENDERDAMS MAPMODELBLOCK SLMODELBLOCKNAME Specsfor
% LocalWords:  ARPROTOTYPENAME ARCOMPONENTNAME GETSIGNALMEMORY MAPSIGNALMEMORY
% LocalWords:  SLMEMORYTYPE ARMEMORYTYPE GETDATADEFAULTS MAPDATADEFAULTS
