classdef ServiceComponent < handle
    % SERVICECOMPONENT Auto-creates RTE layer between application and BSW services
    
    %   Copyright 2017 The MathWorks, Inc.
    
    properties (Constant, Access = private)
        ReferenceImplLibNames = {'Dem Module', 'NvM Module', 'FiM Module'};
        
        DemServiceBlockPath = 'autosarspkglib/Basic Software/Diagnostic Event Manager (Dem)/Diagnostic Service Component';
        NvMServiceBlockPath = 'autosarspkglib/Basic Software/NVRAM Manager (NvM)/NVRAM Service Component';
        
        DemServiceBlockMaskType = 'Diagnostic Service Component';
        NvMServiceBlockMaskType = 'NVRAM Service Component';
    end
    
    methods (Static)
       function serviceBlocks = find(sys)
            % Find service component blocks
            serviceBlocks = find_system( sys, ...
                                        'Regexp', 'on', ...
                                         'FollowLinks', 'on', ...
                                         'LookUnderMasks', 'on', ...
                                         'BlockType', 'SubSystem', ...
                                         'MaskType', ['(^', autosar.bsw.ServiceComponent.NvMServiceBlockMaskType, '$|', ...
                                                       '^', autosar.bsw.ServiceComponent.DemServiceBlockMaskType, '$)']);
        end
        
        function updateApplicationPorts(bswCompPath)
            
            isLib = strcmp( get_param( bdroot(bswCompPath), 'BlockDiagramType' ) , 'library' );
            if isLib
                return
            end
            
            modelName = bdroot(bswCompPath);
            if ~autosar.validation.CompiledModelUtils.isCompiled(modelName)
                return
            end
            
            % checkout embedded coder license
            autosar.bsw.ServiceComponent.checkoutEcoderLicense();
            
            % Check that block is singleton
            autosar.bsw.ServiceComponent.verifySingletonBlock(bswCompPath);
            
            % get the application function names
            allApplicationFcnNames = autosar.bsw.ServiceComponent.getApplicationFcnNames(modelName);
            
            % get the service function names for these application names
            serviceFunctionNames = {};
            applicationFcnNames = {};
            
            clientPortNames = {};
            
            bswCompType = autosar.bsw.ServiceComponent.getBswCompType(bswCompPath);
            for fcnIdx = 1:length(allApplicationFcnNames)
                % Split function names into Port_Operation
                applicationFcnName = allApplicationFcnNames{fcnIdx};
                [operationName, portName] = autosar.bsw.ServiceComponent.parseOperationAndPortNames(...
                    applicationFcnName);
                
                if isempty(operationName)
                    % not a supported BSW function
                    continue;
                end
                
                switch (bswCompType)
                    case 'Dem'
                        if autosar.bsw.ServiceComponent.isDemOperation(operationName)
                            applicationFcnNames{end+1} = applicationFcnName; %#ok<AGROW>
                            serviceFunctionNames{end+1} = [bswCompType '_' operationName]; %#ok<AGROW>
                            clientPortNames{end+1} = portName; %#ok<AGROW>
                            defaultPortDefinedArgument = '1';
                        end
                    case 'NvM'
                        if autosar.bsw.ServiceComponent.isNvMOperation(operationName)
                            applicationFcnNames{end+1} = applicationFcnName; %#ok<AGROW>
                            serviceFunctionNames{end+1} = [bswCompType '_' operationName]; %#ok<AGROW>
                            clientPortNames{end+1} = portName; %#ok<AGROW>
                            defaultPortDefinedArgument = '-1'; % unique IDs
                        end
                    otherwise
                        assert(false, 'Unsupported bsw component type: %s', bswCompType);
                end
            end
            
            clientPortPortDefinedArguments = cell(1, length(clientPortNames));

            % Retain existing port defined arguments
            oldClientPortNames = eval(get_param(bswCompPath, 'ClientPortNames'));
            oldPortDefinedArguments = eval(get_param(bswCompPath, 'ClientPortPortDefinedArguments'));

            for ii = 1:length(clientPortNames)
                portName = clientPortNames{ii};
                portDefinedArgumentCell = oldPortDefinedArguments(strcmp(oldClientPortNames, portName));
                if isempty(portDefinedArgumentCell)
                    clientPortPortDefinedArguments{ii} = defaultPortDefinedArgument;
                else
                    clientPortPortDefinedArguments{ii} = portDefinedArgumentCell{1};
                end
            end
            
            % Change 'auto' (-1) IDs to unique IDs
            if any(strcmp(clientPortPortDefinedArguments, '-1'))
                
                % Get a list of unique IDs that we can use as necessary
                uniqueIDs = cell(size(clientPortPortDefinedArguments));
                for ii=1:length(clientPortPortDefinedArguments)
                    uniqueIDs{ii} = num2str(ii);
                end

                % Replace '-1' entries with unique IDs
                replaceIdx = find(strcmp(clientPortPortDefinedArguments, '-1'));
                replaceClientPortNames = clientPortNames(replaceIdx);
                [~, ~, replaceClientPortNamesIdx] = unique(replaceClientPortNames);
                for ii=1:length(replaceIdx)
                    clientPortPortDefinedArguments(replaceIdx(ii)) = uniqueIDs(replaceClientPortNamesIdx(ii));
                end
            end
            
            % set parameters on the RTE Connector block
            connecterPath = [bswCompPath '/RTE Service Connector'];
            set_param(connecterPath, 'ApplicationFunctionNames', ['{', autosar.api.Utils.cell2str(applicationFcnNames), '}']);
            set_param(connecterPath, 'ServiceFunctionNames', ['{', autosar.api.Utils.cell2str(serviceFunctionNames), '}']);
            set_param(connecterPath, 'ServicePortDefinedArguments', ['{', autosar.api.Utils.cell2str(clientPortPortDefinedArguments), '}']);

            % Make port names unique
            [clientPortNames, index] = unique(clientPortNames);
            clientPortPortDefinedArguments = clientPortPortDefinedArguments(index);
            
            % Set ClientPortNames
            set_param(bswCompPath, 'ClientPortNames', ['{', autosar.api.Utils.cell2str(clientPortNames), '}']);
                
            % Set ClientPort PortDefinedArguments
            set_param(bswCompPath, 'ClientPortPortDefinedArguments', ['{', autosar.api.Utils.cell2str(clientPortPortDefinedArguments), '}']);
            
            % now that all parameters are set, configure the RTE connector
            % block
            autosar.bsw.ServiceComponent.configureRTE(connecterPath);
        end
        
        % error out if user is trying to add the service component block to
        % a model which is mapped to AUTOSAR Component or Composition. We
        % don't want to allow this because it causes issues as the Simulink
        % Function and caller blocks clobber the mapping.
        function preCopyCallback(connectorPath, dstModel)
            if autosar.api.Utils.isMappedToComponent(dstModel)
                connectorName = strrep(connectorPath, [dstModel '/'], '');
                error(message('autosarstandard:bsw:BSWComponentBlockAdditionNotAllowed',...
                    connectorName, dstModel));
            end
        end
        
        % return service component block which can implement the specified
        % application function name
        function serviceBlockPath = getServiceBlockForAppFcnName(applicationFcnName)
            
            % parse operation name
            operationName = autosar.bsw.ServiceComponent.parseOperationAndPortNames(...
                applicationFcnName);
            
            % return service component block which can implement this
            % operation
            if autosar.bsw.ServiceComponent.isDemOperation(operationName)
                serviceBlockPath = autosar.bsw.ServiceComponent.DemServiceBlockPath;
            elseif autosar.bsw.ServiceComponent.isNvMOperation(operationName)
                serviceBlockPath = autosar.bsw.ServiceComponent.NvMServiceBlockPath;
            else
                % not a supported BSW function
                serviceBlockPath = '';
            end
        end
    end
    
    methods (Static)

        function bswCompType = getBswCompType(bswCompPath)
            blkMaskType = get_param(bswCompPath, 'MaskType');
            switch (blkMaskType)
                case autosar.bsw.ServiceComponent.DemServiceBlockMaskType
                    bswCompType = 'Dem';
                case autosar.bsw.ServiceComponent.NvMServiceBlockMaskType
                    bswCompType = 'NvM';
                otherwise
                    assert(false, 'Unsupported BSW block mask type: %s', blkMaskType);
            end
        end

        function blkPath = createBlock(sys, serviceImplStr)
        % Create service component block
            
            serviceImpl = eval(serviceImplStr);
            
            blkName = serviceImpl.getType();
            blkPath = [sys, '/', blkName];
            blkH = add_block('built-in/Subsystem', blkPath, 'MakeNameUnique', 'on');
            blkPath = getfullname(blkH);
            
            rteConnectorBlockPath = 'autosarspkglib_internal/RTE Service Connector';
            add_block(rteConnectorBlockPath, [blkPath, '/RTE Service Connector'], 'Position', [265 49 405 86]);
            
            % Add service specific blocks
            serviceImpl.addBlocks(blkPath);
            
            % Mask block
            mo = get_param(blkPath, 'MaskObject');
            if isempty(mo)
                mo = Simulink.Mask.create(blkPath);
            end

            % Mask help
            mo.Description    = serviceImpl.getDescription();
            mo.Help           = serviceImpl.getHelp();
            mo.Type           = serviceImpl.getType();

            % Mask initialization
            mo.Initialization = 'autosar.bsw.ServiceComponent.updateApplicationPorts(gcb);';
            
            % Mask needs to be self modifiable to modifying underlying
            % Function Caller
            mo.SelfModifiable = 'on';
            
            % Remove and then add mask parameters
            mo.removeAllParameters();
            mo.addParameter('Type', 'edit', ...
                            'Name', 'ClientPortNames',...
                            'Prompt', 'Application port names:', ...
                            'Value', '{}', ...
                            'Tunable','on',...
                            'Evaluate','off',...
                            'Hidden','off');
            mo.addParameter('Type', 'edit', ...
                            'Name', 'ClientPortPortDefinedArguments',...
                            'Prompt', 'Port defined arguments:', ...
                            'Value', '{}', ...
                            'Tunable','on',...
                            'Evaluate','off',...
                            'Hidden','off');
            serviceImpl.addMaskObjectParameters(mo);

            % Attach UI (DDG)
            set_param(blkPath, 'DialogController', 'bsw_create_dialog');
            set_param(blkPath, 'DialogControllerArgs', {'Component'});  
            
            % Set PreCopyFcn
            set_param(blkPath, 'PreCopyFcn', 'autosar.bsw.ServiceComponent.preCopyCallback(gcb, bdroot);');
            
            % Display just "type" rather than "type (mask)(link)"
            dialogControls = mo.getDialogControls();
            for ii=1:length(dialogControls)
                dc = dialogControls(ii);
                if strcmp(dc.Name, 'DescGroupVar')
                    dc.Prompt = serviceImpl.getType();
                    break
                end
            end
            
            blockIconType     = serviceImpl.getBlockIconType();
            if ~isempty(blockIconType)
                blockDVGIcon = ['BSWBlockIcon.', blockIconType];
                mo.BlockDVGIcon   = blockDVGIcon;

                % BSW SVG icons optimized for following block size.
                width = 140; 
                height = 65;
                
                % BSW SVG icons need to be pre-registered
                set_param(bdroot(sys), 'PreLoadFcn', 'autosar.bsw.BasicSoftwareCaller.registerIcons();');
            else
                mo.Display        = serviceImpl.getDisplay();
                width = 140;
                height = 65;
            end
            
            % Resize block
            pos = get_param(blkPath, 'Position');
            pos(3) = pos(1) + width;
            pos(4) = pos(2) + height;
            set_param(blkPath, 'Position', pos);

            
            % Run the initialization commands on the block so block
            % parameters are fully initialized
            Simulink.Block.eval(get_param(blkPath, 'Handle'));
        end
        
    end
    
    methods (Static, Access = private)
        
        function [operationName,portName] = parseOperationAndPortNames(fcnName)
            operationName = '';
            portName = '';
            
            pat = '(?<PortName>\w+)\_(?<OperationName>\w+)';
            names = regexp(fcnName, pat, 'names');
            if ~isempty(names)
                operationName = names.OperationName;
                portName = names.PortName;
            end
        end
        
        % Check out embedded coder license
        function checkoutEcoderLicense()
            % autosarlicensed will checkout embedded coder license
            [islicensed, errorargs] = autosarlicensed();
            if ~islicensed
                DAStudio.error(errorargs{:});
            end
        end
        
        % Check for multiple service component blocks
        function verifySingletonBlock(blkPath)

            maskType = get_param(blkPath, 'MaskType');
            serviceBlocks = find_system( bdroot(blkPath), ...
                                         'FollowLinks', 'on', ...
                                         'LookUnderMasks', 'on', ...
                                         'BlockType', 'SubSystem', ...
                                         'MaskType', maskType);
            if length(serviceBlocks) > 1
                DAStudio.error('autosarstandard:bsw:multipleServiceComponents', maskType, serviceBlocks{1}, serviceBlocks{2});
            end

        end
        
        
        % auto-create the RTE layer between the application and BSW
        % services
        function configureRTE(connectorPath)
            % don't update library
            isLib = strcmp(get_param(bdroot(connectorPath), 'BlockDiagramType'), 'library' );
            if isLib
                return
            end
            
            % sync'ing operation is merely a delete/add of RTE layer
            autosar.bsw.ServiceComponent.deleteRTELayer(connectorPath);
            
            applicationFcnNamesStr = get_param(connectorPath, 'ApplicationFunctionNames');
            applicationFcnNames = eval(applicationFcnNamesStr);
            
            serviceFunctionNamesStr = get_param(connectorPath, 'ServiceFunctionNames');
            serviceFunctionNames = eval(serviceFunctionNamesStr);
            
            portDefArgumentsStr = get_param(connectorPath, 'ServicePortDefinedArguments');
            portDefArguments = eval(portDefArgumentsStr);
            
            if isempty(applicationFcnNames)
                % no application functions to implement
                return
            end
            
            % determine BSW component type
            [~,bswCompType] = autosar.bsw.ServiceComponent.parseOperationAndPortNames(...
                serviceFunctionNames{1});
            
            % get the prototype and argument specs for the application
            % functions we are providing implementations for
            applicationFcnsInfo = autosar.bsw.ServiceComponent.getApplicationFcnsInfo(...
                applicationFcnNames, bswCompType);
            if isempty(applicationFcnsInfo)
                return
            end
            
            % build RTE layer
            defaultPos = [105 115 327 215];
            gap = 140;
            for fcnIdx = 1:length(applicationFcnsInfo)
                simulinkFcnBlk = autosar.bsw.ServiceComponent.defineCaller(...
                    connectorPath, applicationFcnsInfo(fcnIdx),...
                    serviceFunctionNames{fcnIdx}, portDefArguments{fcnIdx});
                
                % set position and size
                set_param(simulinkFcnBlk, 'Position',...
                    [defaultPos(1) defaultPos(2)+(gap*(fcnIdx-1))...
                    defaultPos(3) defaultPos(4)+(gap*(fcnIdx-1))]);
                autosar.mm.mm2sl.layout.BlockBeautifier.beautifyBlock(simulinkFcnBlk);
            end
        end
        
        
        function simulinkFcnBlk = defineCaller(destPath, callerInfo,...
                serviceFunctionName, portDefArgument)
            
            % Add a Simulink function block for this caller
            [simulinkFcnBlk, inArgHandles, outArgHandles] =...
                autosar.bsw.ServiceComponent.createOrGetSimulinkFcnForCaller(...
                callerInfo, destPath);
            
            % Add the call to BSW inside the Simulink Function
            
            % get operation name
            [operationName, bswCompType] = autosar.bsw.ServiceComponent.parseOperationAndPortNames(...
                serviceFunctionName);
            
            % get service function prototype
            if autosar.bsw.DemDiagnosticMonitor.FunctionPrototypeWithEventIdMap.isKey(operationName)
                serviceFcnPrototype = autosar.bsw.DemDiagnosticMonitor.FunctionPrototypeWithEventIdMap(operationName);
            elseif autosar.bsw.DemDiagnosticInfo.FunctionPrototypeWithEventIdMap.isKey(operationName)
                serviceFcnPrototype = autosar.bsw.DemDiagnosticInfo.FunctionPrototypeWithEventIdMap(operationName);
            elseif autosar.bsw.NvMService.FunctionPrototypeWithBlockIdMap.isKey(operationName)
                serviceFcnPrototype = autosar.bsw.NvMService.FunctionPrototypeWithBlockIdMap(operationName);
            elseif autosar.bsw.NvMAdmin.FunctionPrototypeWithBlockIdMap.isKey(operationName)
                serviceFcnPrototype = autosar.bsw.NvMAdmin.FunctionPrototypeWithBlockIdMap(operationName);
            else
                assert(false, 'Unrecognized operation: %s', operationName);
            end
            serviceFcnPrototype = sprintf(serviceFcnPrototype, bswCompType);
            
            % Now add the caller block
            % no need to specify argument specs for caller because the
            % Simulink function is in scope, so Simulink will auto-deduce
            % it.
            fcnCallerName = [serviceFunctionName '(caller)'];
            fcnCallerBlk = [simulinkFcnBlk '/' fcnCallerName];
            if isempty(find_system(simulinkFcnBlk, 'SearchDepth', 1,...
                    'BlockType', 'FunctionCaller', 'Name', fcnCallerName))
                add_block('built-in/FunctionCaller', fcnCallerBlk,...
                    'FunctionPrototype', serviceFcnPrototype,...
                    'Position', [335 200 595 240]);
            end
            autosar.mm.mm2sl.layout.BlockBeautifier.beautifyBlock(fcnCallerBlk);
            
            % If BSW module uses ids, add a constant block to specify id
            % value
            bswModuleUsesIds = any(strcmp(bswCompType, {'Dem', 'NvM'}));
            if bswModuleUsesIds
                idConstantName = 'Id';
                idConstantBlock = [simulinkFcnBlk '/' idConstantName];
                if isempty(find_system(simulinkFcnBlk, 'SearchDepth', 1,...
                        'BlockType', 'Constant', 'Name', idConstantName))
                    add_block('built-in/Constant',idConstantBlock);
                end
                set_param(idConstantBlock, 'Value', num2str(portDefArgument),...
                    'OutDatatypeStr', 'uint16',...
                    'Position', [265 205 295 235]);
                autosar.mm.mm2sl.layout.BlockBeautifier.beautifyBlock(idConstantBlock);
                
                % event id is always the first argument for Dem
                lh = get_param(idConstantBlock, 'LineHandles');
                if (lh.Outport==-1) ||...
                        ~strcmp(getfullname(get_param(lh.Outport, 'DstBlockHandle')), fcnCallerBlk)
                    add_line(simulinkFcnBlk, [idConstantName '/1'],...
                        [fcnCallerName '/1'],...
                        'autorouting', 'on');
                end
                autosar.mm.mm2sl.MRLayoutManager.homeBlk(idConstantBlock);
            end
            
            % connect the argument blocks to the inports/outports of the
            % caller block
            if bswModuleUsesIds
                startIdx = 2;
            else
                startIdx = 1;
            end
            ph = get_param(fcnCallerBlk, 'PortHandles');
            inArgIdx = 1;
            for pIdx = startIdx:length(ph.Inport)
                inArgHandle = inArgHandles(inArgIdx);
                argBlockName = get_param(inArgHandle, 'Name');
                lh = get_param(inArgHandle, 'LineHandles');
                if (lh.Outport==-1) ||...
                        ~strcmp(getfullname(get_param(lh.Outport, 'DstBlockHandle')), fcnCallerBlk)
                    autosar.mm.mm2sl.layout.LayoutHelper.addLine(simulinkFcnBlk,...
                        [argBlockName '/1'],...
                        [fcnCallerName '/' num2str(pIdx)]);
                end
                autosar.mm.mm2sl.MRLayoutManager.homeBlk(inArgHandle);
                inArgIdx = inArgIdx+1;
            end
            
            for pIdx = 1:length(ph.Outport)
                outArgHandle = outArgHandles(pIdx);
                argBlockName = get_param(outArgHandle, 'Name');
                lh = get_param(outArgHandle, 'LineHandles');
                if (lh.Inport==-1) ||...
                        ~strcmp(getfullname(get_param(lh.Inport, 'SrcBlockHandle')), fcnCallerBlk)
                    autosar.mm.mm2sl.layout.LayoutHelper.addLine(simulinkFcnBlk,...
                        [fcnCallerName '/' num2str(pIdx)],...
                        [argBlockName '/1']);
                end
                autosar.mm.mm2sl.MRLayoutManager.homeBlk(outArgHandle);
            end
        end
        
        % Return info (function prototype and argument specs) for
        % application functions in model
        function fcnsInfo = getApplicationFcnsInfo(applicationFcnNames, bswCompType)
            fcnsInfo = [];
            
            if isempty(applicationFcnNames)
                return
            end
            
            % Deduce the full application function prototypes. Assume the
            % application callers use the same argument names and
            % specifications as the ones we ship for BSW blocks
            fcnInfo = struct('ApplicationFcnPrototype','',...
                'InArgSpecs','',...
                'OutArgSpecs','',...
                'EnumDatatype', '');
            for fcnIdx = 1:length(applicationFcnNames)
                applicationFcnName = applicationFcnNames{fcnIdx};
                [operationName,portName] = autosar.bsw.ServiceComponent.parseOperationAndPortNames(...
                    applicationFcnName);
                
                switch (bswCompType)
                    case 'Dem'
                        if autosar.bsw.DemDiagnosticMonitor.FunctionPrototypeMap.isKey(operationName)
                            serviceClassName = 'autosar.bsw.DemDiagnosticMonitor';
                        elseif autosar.bsw.DemDiagnosticInfo.FunctionPrototypeMap.isKey(operationName)
                            serviceClassName = 'autosar.bsw.DemDiagnosticInfo';
                        else
                            % user might be calling BSW functionality that we don't
                            % support yet
                            continue;
                        end
                    case 'NvM'
                        if autosar.bsw.NvMService.FunctionPrototypeMap.isKey(operationName)
                            serviceClassName = 'autosar.bsw.NvMService';
                        elseif autosar.bsw.NvMAdmin.FunctionPrototypeMap.isKey(operationName)
                            serviceClassName = 'autosar.bsw.NvMAdmin';
                        else
                            % user might be calling BSW functionality that we don't
                            % support yet
                            continue;
                        end
                    otherwise
                        assert(false, 'Unsupported bsw component type: %s', bswCompType);
                end
                
                % fill in the application functions details
                serviceFcnPrototype = eval([serviceClassName '.FunctionPrototypeMap(''' operationName ''')']);
                fcnInfo.ApplicationFcnPrototype = sprintf(serviceFcnPrototype, portName);
                fcnInfo.InArgSpecs = eval([serviceClassName '.InputArgSpecMap(''' operationName ''')']);
                fcnInfo.OutArgSpecs = eval([serviceClassName '.OutputArgSpecMap(''' operationName ''')']);
                fcnInfo.EnumDatatype = eval([serviceClassName '.EnumDatatypeMap(''' operationName ''')']);
                fcnsInfo = [fcnsInfo fcnInfo]; %#ok<AGROW>
            end
        end
        
        % creates a Simulink function that implements the interface of
        % the caller
        function [simulinkFcnBlk, inArgHandles, outArgHandles] =...
                createOrGetSimulinkFcnForCaller(callerInfo, destPath)
            
            % parse the application function prototype
            [inArgNames, outArgNames, fcnName] =...
                autosar.validation.ClientServerValidator.getFcnInOutParamNames(...
                callerInfo.ApplicationFcnPrototype);
            
            % Add the Simulink function block if one does not exist
            simulinkFcnBlk = [destPath '/' fcnName];
            if isempty(find_system(destPath, 'SearchDepth', 1,...
                    'LookUnderMasks', 'all','FollowLinks', 'on',...
                    'BlockType', 'SubSystem',  'Name', fcnName))
                
                add_block('built-in/SubSystem',...
                    [destPath '/' fcnName]);
                
                add_block('built-in/TriggerPort',...
                    [simulinkFcnBlk '/' fcnName],...
                    'TriggerType', 'function-call',...
                    'StatesWhenEnabling', 'held',... % workaround g1617519
                    'IsSimulinkFunction', 'on',...
                    'FunctionVisibility', 'global',...
                    'FunctionName', fcnName,...
                    'Position', [ 400 20 420 40]);
            else
                assert(strcmp(get_param(simulinkFcnBlk, 'IsSimulinkFunction'), 'on'),...
                    '%s is not a Simulink function', simulinkFcnBlk);
            end
            
            % Add the argument blocks
            inArgSpecs = regexp(callerInfo.InArgSpecs, '\s*,\s*', 'split');
            outArgSpecs = regexp(callerInfo.OutArgSpecs, '\s*,\s*', 'split');
            
            inArgPosition = [85 60 170 80];
            gap = 80;
            inArgHandles = [];
            for argIdx=1:length(inArgNames)
                inArgName = inArgNames{argIdx};
                inArgSpec = inArgSpecs{argIdx};
                argH = autosar.bsw.ServiceComponent.createOrGetArgumentBlk(...
                    simulinkFcnBlk, inArgName, inArgSpec,...
                    callerInfo.EnumDatatype, true);
                
                % set position and size
                set_param(argH, 'Position',...
                    [inArgPosition(1) inArgPosition(2)+(gap*argIdx)...
                    inArgPosition(3) inArgPosition(4)+(gap*argIdx)]);
                autosar.mm.mm2sl.layout.BlockBeautifier.beautifyBlock(argH);
                inArgHandles = [inArgHandles argH]; %#ok<AGROW>
            end
            
            
            outArgPosition = [685 60 735 80];
            outArgHandles = [];
            for argIdx=1:length(outArgNames)
                outArgName = outArgNames{argIdx};
                outArgSpec = outArgSpecs{argIdx};
                
                argH = autosar.bsw.ServiceComponent.createOrGetArgumentBlk(...
                    simulinkFcnBlk, outArgName, outArgSpec,...
                    callerInfo.EnumDatatype, false);
                
                % set position and size
                set_param(argH, 'Position',...
                    [outArgPosition(1) outArgPosition(2)+(gap*argIdx)...
                    outArgPosition(3) outArgPosition(4)+(gap*argIdx)]);
                autosar.mm.mm2sl.layout.BlockBeautifier.beautifyBlock(argH);
                outArgHandles = [outArgHandles argH]; %#ok<AGROW>
            end
        end
        
        
        % deletes RTE layer i.e. all global Simulink functions
        function deleteRTELayer(sys)
            subsystems = find_system(sys, 'SearchDepth', 1,...
                'LookUnderMasks', 'all', 'FollowLinks', 'on',...
                'BlockType', 'SubSystem', 'IsSimulinkFunction', 'on');
            for subIdx = 1:length(subsystems)
                subsys = subsystems{subIdx};
                trigPort = find_system(subsys, 'SearchDepth', 1, ...
                    'FollowLinks', 'on', 'BlockType', 'TriggerPort');
                if strcmp(get_param(trigPort,'FunctionVisibility'),'global')
                    delete_block(subsys);
                end
            end
        end
        
        % Add argument block to specified simulinkFcnPath and set datatype
        % based on argSpec
        function argH = createOrGetArgumentBlk(simulinkFcnPath, argName, argSpec,...
                enumDatatype, isArgIn)
            
            portDimStr = '1';
            if contains(argSpec, '%s')
                outDatatype = enumDatatype;
            else
                value = eval(argSpec);
                outDatatype = class(value);
                if numel(value) > 1
                    portDimStr = mat2str(size(value));
                end
            end
            
            if strcmp(outDatatype, 'logical')
                outDatatype = 'boolean';
            end
            
            if isArgIn
                argBlkType = 'ArgIn';
            else
                argBlkType = 'ArgOut';
            end
            
            argBlock = [simulinkFcnPath '/' argName];
            if isempty(find_system(simulinkFcnPath, 'SearchDepth', 1,...
                    'BlockType', argBlkType, 'Name', argName))
                argH = add_block(['built-in/' argBlkType], argBlock,...
                    'ArgumentName', argName);
            else
                argH = get_param(argBlock, 'Handle');
            end
            set_param(argH, 'OutDataTypeStr', outDatatype,...
                'PortDimensions', portDimStr);
            autosar.mm.mm2sl.layout.BlockBeautifier.beautifyBlock(argH);
        end
        
        % get the application function names using compiled information
        % from CompiledSimulinkFunctions
        function fcnNames = getApplicationFcnNames(model)
            fcnNames = {};
            compiledFcns = autosar.bsw.ServiceComponent.getCompiledCallersAndFunctions(model);
            
            if isempty(compiledFcns)
                return
            end
            
            % return the callers that are not implemented
            for fcnIdx = 1:length(compiledFcns)
                compiledFunction = compiledFcns(fcnIdx);
                if isempty(compiledFunction.FunctionBlock)
                    fcnNames{end+1} = compiledFunction.FunctionName; %#ok<AGROW>
                end
            end
            fcnNames = unique(fcnNames);
        end
        
        % get the compiled functions information
        function compiledFcns = getCompiledCallersAndFunctions(model)
            assert(autosar.validation.CompiledModelUtils.isCompiled(model),...
                'Getting application names is only supported when model is compiled');
            
            compiledFcns = [];
            
            % We are only interested in callers inside Model blocks
            % (which represent AUTOSAR Components or Compositions)
            allModelBlocks = find_system(model, 'FollowLinks', 'on',...
                'LookUnderMasks', 'all', 'BlockType', 'ModelReference');
            
            % filter our basic software reference libraries from the list
            modelBlocks = [];
            for blkIdx = 1:length(allModelBlocks)
                modelBlock = allModelBlocks{blkIdx};
                if ~any(strcmp(autosar.bsw.ServiceComponent.getRefModelName(modelBlock),...
                        autosar.bsw.ServiceComponent.ReferenceImplLibNames))
                    modelBlocks{end+1} = modelBlock; %#ok<AGROW>
                end
            end
            
            if isempty(modelBlocks)
                return
            end
            
            % get the compiled functions & callers info
            for blkIdx = 1:length(modelBlocks)
                modelBlock = modelBlocks{blkIdx};
                refModelName = autosar.bsw.ServiceComponent.getRefModelName(modelBlock);
                if strcmp(get_param(modelBlock, 'SimulationMode'), 'Normal')
                    % if model block is in normal mode, then we can get
                    % CompiledSimulinkFunctions from the loaded referenced
                    % model
                    if bdIsLoaded(refModelName)
                        slFcnsCatalog = get_param(refModelName, 'CompiledSimulinkFunctions');
                        compiledFcns = [compiledFcns slFcnsCatalog.CompFunctions]; %#ok<AGROW>
                    end
                else
                    % not in normal mode, deserialize the information from
                    % the model's ModelRefCompileInfo.xml file
                    modelRefInfoFile = coder.internal.modelRefUtil(refModelName,...
                        'getModelRefInfoFileName', 'SIM',...
                        get_param(model, 'SystemTargetFile'));
                    if exist(modelRefInfoFile, 'file') == 2
                        parser = mf.zero.io.XmlParser;
                        parsedContents = parser.parseFile(modelRefInfoFile);
                        compiledFcns = [compiledFcns...
                            parsedContents.simulinkFunctions.compSimulinkFunctionCatalog.CompFunctions]; %#ok<AGROW>
                    end
                end
            end
        end
        
        function refModelName = getRefModelName(hBlock)
            lIsProtected = strcmp(get_param(hBlock, 'ProtectedModel'), 'on');
            if lIsProtected
                lModelFile = get_param(hBlock, 'ModelFile');
                refModelName = strrep(lModelFile, '.slxp', '');
            else
                refModelName = get_param(hBlock, 'ModelName');
            end
        end
        
        % Simulink.FunctionGraphCatalog implementation
        function fcnNames = getApplicationFcnNames_graph(modelH)
            assert(ishandle(modelH), 'modelH should be a handle');
            
            functionCatalog = Simulink.FunctionGraphCatalog(modelH, 'ListFunctionsAndCallers');
            fcnNames = {};
            for fcnIdx = 1:length(functionCatalog)
                fcn = functionCatalog{fcnIdx};
                if ~isempty(fcn{2})
                    fcnName = functionCatalog{fcnIdx}{1};
                    if ~isempty(fcnName)
                        fcnNames{end+1} = fcnName; %#ok<AGROW>
                    end
                end
            end
            fcnNames = unique(fcnNames);
        end
        
        function isDem = isDemOperation(operationName)
            isDem = autosar.bsw.DemDiagnosticMonitor.FunctionPrototypeWithEventIdMap.isKey(operationName) ||...
                autosar.bsw.DemDiagnosticInfo.FunctionPrototypeWithEventIdMap.isKey(operationName);
        end
        
        function isNvM = isNvMOperation(operationName)
            isNvM = autosar.bsw.NvMService.FunctionPrototypeMap.isKey(operationName) ||...
                autosar.bsw.NvMAdmin.FunctionPrototypeMap.isKey(operationName);
        end
        
    end
end

% LocalWords:  BSW Nv bsw autosarstandard sync'ing autorouting MService MAdmin
% LocalWords:  dem nvm fim autosarlicensed autosarspkglib SVG
