classdef MetaModelValidator < autosar.validation.PhasedValidator
% MetaModelValidator
    
%   Copyright 2013-2017 The MathWorks, Inc.
    properties (Access=private)
        %runnableToEventMap;
    end

    methods (Access=protected)
        
        function verifyInitial(this, hModel)

            m3iModel = autosar.api.Utils.m3iModel(hModel);
            
            commonValidator = autosar.validation.MetaModelCommonValidator(hModel);
            commonValidator.verify(m3iModel);

            arRoot = m3iModel.RootPackage.front();
            m3iComp = autosar.mm.Model.findChildByName(arRoot,...
                                                       arRoot.ComponentQualifiedName);

            compValidator = autosar.validation.MetaModelMappedComponentValidator(hModel);
            compValidator.verify(m3iComp);
            
            this.verifyXmlOptions(hModel);
        end
        
        function verifyPostProp(this, hModel)
            m3iModel = autosar.api.Utils.m3iModel(hModel);
            arRoot = m3iModel.RootPackage.front();
            m3iComp = autosar.mm.Model.findChildByName(arRoot,...
                arRoot.ComponentQualifiedName);
            this.verifyRunnables(hModel, m3iComp);
            this.verifyEvents(hModel, m3iComp);
            this.verifyRunnableCanBeInvokedConcurrently(m3iComp, arRoot);
            this.verifyNoInternalBehaviorContaineeShortNameClash(hModel, m3iComp);  
        end
        
    end
    
    methods (Static, Access=private)        

        function verifyRunnables(hModel, compObj)
            activeMapping = autosar.api.Utils.modelMapping(hModel);
            
            % Check first whether mapped runnable exists before we do all the
            % validation on meta-model runnables and events
            runnableSeq = autosar.mm.Model.findObjectByMetaClass(compObj.Behavior,...
                                                              Simulink.metamodel.arplatform.behavior.Runnable.MetaClass(), false, false);
                                                          

            functionCallInports = activeMapping.FcnCallInports;
            for ii=1:runnableSeq.size()
                runnable = runnableSeq.at(ii);
                if ~isempty(functionCallInports)
                    for jj = 1 : numel(functionCallInports)
                        if strcmp(runnable.Name, functionCallInports(jj).MappedTo.Runnable)
                            if runnable.Events.isEmpty()
                                mException = MException('RTW:autosar:validateEventForExportedFunction', ...
                                    DAStudio.message('RTW:autosar:validateEventForExportedFunction', runnable.Name));
                                throw(mException);                                
                            end
                            break;
                        end
                    end
                end
            end

            irvSeq = autosar.mm.Model.findObjectByMetaClass(compObj.Behavior,...
                Simulink.metamodel.arplatform.behavior.IrvData.MetaClass(), false, false);
            
            isExportedFunctionModel = ...
                autosar.validation.ExportFcnValidator.isExportFcn(hModel);
            
            if isExportedFunctionModel && ...
               (~isempty(activeMapping.FcnCallInports) || ...
                ~isempty(activeMapping.ServerFunctions))
                if runnableSeq.size() ~= numel(activeMapping.FcnCallInports) ...
                        + numel(activeMapping.ServerFunctions) + ...
                        + numel(activeMapping.ResetFunctions) + ...
                        numel(activeMapping.TerminateFunctions) + 1
                    % Don't allow more runnables than entry point functions
                    mException = MException('autosarstandard:validation:multiRunnableTooManyRunnables', ...
                                            DAStudio.message('autosarstandard:validation:multiRunnableTooManyRunnables'));
                    throw(mException);
                end

                % Check IRVs for MultiRunnable case
                if irvSeq.size() ~= (numel(activeMapping.DataTransfers) + numel(activeMapping.RateTransition))
                    % Don't allow more IRVs than data transfers
                    mException = MException('autosarstandard:validation:multiRunnableTooManyIRVs', ...
                                            DAStudio.message('autosarstandard:validation:multiRunnableTooManyIRVs'));
                    throw(mException);
                end
            else
                % Validate asynchronous exported functions and periodic runnables.
                expectedRunnalbes = numel(activeMapping.FcnCallInports) + ...     % Function call runnables
                                    length(activeMapping.StepFunctions) + ...     % Periodic runnables
                                    length(activeMapping.ResetFunctions) + ...    % Reset runnables
                                    numel(activeMapping.TerminateFunctions) + ... % Terminate runnables
                                    1;                                            % Initialization function
                
                % Assume step model
                % Multitasking
                if runnableSeq.size() ~= expectedRunnalbes
                    % Don't allow extra runnables
                    mException = MException('autosarstandard:validation:multiRunnableTooManyRunnables', ...
                                            DAStudio.message('autosarstandard:validation:multiRunnableTooManyRunnables'));
                    throw(mException);
                end
                
                if irvSeq.size() ~= length(activeMapping.RateTransition)
                    % Don't allow more IRVs than data transfers
                    mException = MException('autosarstandard:validation:multitaskingTooManyIRVs', ...
                                            DAStudio.message('autosarstandard:validation:multitaskingTooManyIRVs'));
                    throw(mException);
                end
            end
            
        end    

        function verifyEvents(hModel, m3iComp)
            
            activeMapping = autosar.api.Utils.modelMapping(hModel);
            schemaVersion = get_param(hModel, 'AutosarSchemaVersion');

            if m3iComp.Behavior.isvalid()
                for runIdx = 1:m3iComp.Behavior.Runnables.size()
                    m3iRun = m3iComp.Behavior.Runnables.at(runIdx);
                    [messageID, message] = autosar.validation.MetaModelValidator.verifyEventsForRunnable(...
                        hModel, m3iRun.Name, m3i.mapcell(@(x) x, m3iRun.Events), activeMapping, schemaVersion);
                    if ~isempty(messageID)
                        mException = MException(messageID, message);
                        throw(mException);
                    end
                end
            end
        end
       
        % Function: verifyRunnableCanBeInvokedConcurrently ====================
        % Abstract: 
        %   Verify that runnables (if configured) can be invoked concurrently.
        function verifyRunnableCanBeInvokedConcurrently(m3iComp, arRoot)
        %
            classStr = 'Simulink.metamodel.arplatform.behavior.OperationInvokedEvent';
            messageID = 'RTW:autosar:validateRunnableCanBeInvokedConcurrently';
            if m3iComp.Behavior.isvalid()
                diag = autosar.mm.util.XmlOptionsAdapter.get(...
                    arRoot, 'CanBeInvokedConcurrentlyDiagnostic');
                for runIdx = 1:m3iComp.Behavior.Runnables.size()
                    m3iRun       = m3iComp.Behavior.Runnables.at(runIdx);
                    m3iRunEvents = m3iRun.Events;
                    
                    if (m3iRunEvents.size() == 0)
                        cannotBeConcurrent = true;                        
                    else
                        cannotBeConcurrent = ~isa(m3iRunEvents.at(1), classStr);
                    end
                    
                    if (m3iRun.canBeInvokedConcurrently && cannotBeConcurrent)
                        message = DAStudio.message(messageID, m3iRun.Name);
                        if strcmp(diag, 'Error')
                            mException = MException(messageID, message);
                            throw(mException);
                        else
                             autosar.mm.util.MessageReporter.createWarning(messageID);
                        end
                    end
                end
            end
        end
        
        function verifyNoInternalBehaviorContaineeShortNameClash(hModel, m3iComp)
        % Function: verifyInternalBehaviorShortName =============================
        % Abstract: Verify that for all elements that will end up in internal behavior
        % there are no short name clashes.
            
            m3iBehavior = m3iComp.Behavior;
            if ~m3iBehavior.isvalid()
                return
            end

            runnableNames = m3i.mapcell(@(obj) obj.Name, m3iBehavior.Runnables);
            eventNames    = m3i.mapcell(@(obj) obj.Name, m3iBehavior.Events);
            irvNames      = m3i.mapcell(@(obj) obj.Name, m3iBehavior.IRV);
            
            dsmPIMNames         = autosar.validation.DataObjectValidator.getDSMPIMNames(hModel);
            internalCalPrmNames = autosar.validation.DataObjectValidator.getInternalCalPrmNames(hModel);
            constMemPrmNames    = autosar.validation.DataObjectValidator.getConstMemoryPrmNames(hModel);
            staticMemVarNames   = autosar.validation.DataObjectValidator.getStaticMemoryVarNames(hModel);

            internalBehaviorShortNames = [runnableNames eventNames irvNames dsmPIMNames internalCalPrmNames constMemPrmNames staticMemVarNames];
            [~, unique_indices] = unique( internalBehaviorShortNames );
            duplicates = unique( internalBehaviorShortNames( setdiff( 1:length( internalBehaviorShortNames), ...
                                                              unique_indices ) ) );
            if ~isempty(duplicates)
                msg = DAStudio.message('RTW:autosar:internalBehavShortNameClash', duplicates{1});
                DAStudio.error('RTW:fcnClass:finish', msg);
            end

            % check for case insensitive matches
            autosar.validation.AutosarUtils.checkShortNameCaseClash(internalBehaviorShortNames);
        end


        function verifyXmlOptions(hModel)
            
            m3iModel = autosar.api.Utils.m3iModel(hModel);
            arRoot = m3iModel.RootPackage.front();
            schemaVersion = get_param(hModel, 'AutosarSchemaVersion');
            [componentPackage, ~] = autosar.mm.sl2mm.ModelBuilder.getNodePathAndName(arRoot.ComponentQualifiedName);
            [foundDuplicate, errmsg] = autosar.mm.util.checkAmbigousXmlOptions(m3iModel,...
                componentPackage, arRoot.DataTypePackage, arRoot.InterfacePackage,...
                arRoot.InternalBehaviorQualifiedName, arRoot.ImplementationQualifiedName,...
                AR4.isCompatibleSchemaVer(schemaVersion));
            if foundDuplicate
                mException = MException('Simulink:Engine:RTWCGAutosarValidateError', errmsg);
                throw(mException);
                
            end
        end
       
    end
    
    methods (Static, Access=public)
        
        function msgIDAndHoles = verifyModeSwitchInterface(m3iModeSwitchInterface)
            % VERIFYMODESWITCHINTERFACE Find inconsistencies in an MS
            % Interface object.
            % According to AUTOSAR RTE spec, ms interfaces must have
            % exactly one mode group, and that mode group must have exactly
            % one initial mode.
            msgIDAndHoles = {};
            if isempty(m3iModeSwitchInterface.ModeGroup)
                msgIDAndHoles = {'autosarstandard:validation:msInterfaceMissingModeGroup', m3iModeSwitchInterface.Name};
            end
        end
        
        function [messageID, message] = verifyPRPort(portName, schemaVersion)
            messageID = '';
            message = '';
            
            % PRPorts are not supported for AUTOSAR Schema version less than 4.1
            if str2double(schemaVersion) < 4.1
                messageID = 'RTW:autosar:disallowedPRPort';
                message = DAStudio.message('RTW:autosar:disallowedPRPort', portName);
            end
        end
        
        function [messageID, message] = verifyEventsForRunnable(hModel, ...
                runnableName, eventArray, mapping, schemaVersion)
            messageID = '';
            message = '';
            numTimingEvents = 0;
            numDataReceivedEvents = 0;
            numDataReceiveErrorEvents = 0;
            numModeSwitchEvents = 0;
            numInitEvents = 0;
            numOpInvokedEvents = 0;
            numExternalTriggerOccurredEvents = 0;
            functionCallInports = mapping.FcnCallInports;
            for ii = 1 : numel(eventArray)
                if isa(eventArray{ii}, 'Simulink.metamodel.arplatform.behavior.TimingEvent')
                    numTimingEvents = numTimingEvents + 1;
                elseif isa(eventArray{ii}, 'Simulink.metamodel.arplatform.behavior.DataReceivedEvent')
                    numDataReceivedEvents = numDataReceivedEvents + 1;
                elseif isa(eventArray{ii}, 'Simulink.metamodel.arplatform.behavior.DataReceiveErrorEvent')
                    numDataReceiveErrorEvents = numDataReceiveErrorEvents + 1;
                elseif isa(eventArray{ii}, 'Simulink.metamodel.arplatform.behavior.ModeSwitchEvent')
                    numModeSwitchEvents = numModeSwitchEvents + 1;
                    %[constr_1098] Mode switch and mode disabling. A SwcModeSwitchEvent shall
                    % not simultaneously reference to the same ModeDeclaration in both the roles mode
                    % and disabledMode.                    
                    % If [constr_1098] would not apply it might happen that a RunnableEntity would be
                    % triggered by a SwcModeSwitchEvent and on the same time it would be suppressed
                    % by the mode disabling.
                    if eventArray{ii}.DisabledMode.size() > 0
                        for jj = 1 : eventArray{ii}.instanceRef.size()
                            for kk = 1 : eventArray{ii}.DisabledMode.size()
                                if eventArray{ii}.instanceRef.at(jj).Mode == eventArray{ii}.DisabledMode.at(kk).Mode
                                    messageID = 'RTW:autosar:sameModeForTriggerAndDisabledMode';
                                    message = DAStudio.message(messageID, eventArray{ii}.Name);
                                    return;
                                end
                            end
                        end
                    end
                elseif isa(eventArray{ii}, 'Simulink.metamodel.arplatform.behavior.InitEvent')
                    % [constr_1260] No mode disabling for InitEvents d An InitEvent shall not have
                    % a reference to a ModeDeclaration in the role disabledMode.
                    % Rationale: the concept of RunnableEntity triggered by an InitEvent is (as mentioned
                    % before) orthogonal to the mode concept and therefore shall be implemented
                    % independent of modes.
                    if eventArray{ii}.DisabledMode.size() > 0
                        messageID = 'RTW:autosar:noDisabledModeForInitEvent';
                        message = DAStudio.message(messageID, eventArray{ii}.Name);
                        return;
                    end
                    numInitEvents = numInitEvents + 1;
                elseif isa(eventArray{ii}, 'Simulink.metamodel.arplatform.behavior.OperationInvokedEvent')
                    numOpInvokedEvents = numOpInvokedEvents + 1;
                elseif isa(eventArray{ii}, 'Simulink.metamodel.arplatform.behavior.ExternalTriggerOccurredEvent')
                    numExternalTriggerOccurredEvents = numExternalTriggerOccurredEvents + 1;
                end
            end
            
            % InitEvent is not supported for AUTOSAR Schema version less than 4.1
            if str2double(schemaVersion) < 4.1
               if numInitEvents > 0
                  messageID = 'RTW:autosar:invalidInitEvent';
                  message = DAStudio.message(messageID, runnableName);
                  return;
               end
            end
            if str2double(schemaVersion) < 4.0
               if numExternalTriggerOccurredEvents > 0
                  messageID = 'autosarstandard:validation:invalidExternalTriggerOccurredEvent';
                  message = DAStudio.message(messageID, runnableName);
                  return;
               end
            end
            
            mappedRunnable = false;
            
            % Init Runnable
            if ~isempty(mapping.InitializeFunctions) &&...
                    ~isempty(mapping.InitializeFunctions(1).MappedTo.Runnable) &&...
                    strcmp(runnableName, mapping.InitializeFunctions(1).MappedTo.Runnable)
                % Init Runnable can only have one event (ModeSwitchEvent) (0-1)
                if numDataReceivedEvents > 0 || numDataReceiveErrorEvents > 0 ||...
                        numTimingEvents > 0 || numOpInvokedEvents > 0 ||...
                        (numModeSwitchEvents + numInitEvents) > 1
                    messageID = 'autosarstandard:validation:validateEventForInitFunction';
                    message = DAStudio.message(messageID, runnableName);
                    return;
                end
                mappedRunnable = true;
                
            % Terminate Runnable
            elseif ~isempty(mapping.TerminateFunctions) &&...
                    ~isempty(mapping.TerminateFunctions(1).MappedTo.Runnable) &&...
                    strcmp(runnableName, mapping.TerminateFunctions(1).MappedTo.Runnable)
                % Terminate Runnable can only have one event (ModeSwitchEvent) (0-1)
                if numTimingEvents > 0 || ...
                        numModeSwitchEvents + numDataReceivedEvents + ...
                        numDataReceiveErrorEvents + numOpInvokedEvents + ...
                        numInitEvents + numExternalTriggerOccurredEvents > 1
                    messageID = 'autosarstandard:validation:validateEventForTerminateFunction';
                    message = DAStudio.message(messageID, runnableName);
                    return;
                end
                mappedRunnable = true;
                
            % Step Runnable (Single Tasking or Multitasking)            
            elseif ~isempty(mapping.StepFunctions) 
                
                for idx = 1:length(mapping.StepFunctions)
                    if ~isempty(mapping.StepFunctions(idx).MappedTo.Runnable) && ...
                            strcmp(runnableName, mapping.StepFunctions(idx).MappedTo.Runnable)
                        % Step Runnable must have only one TimingEvent (1-1).
                        if numTimingEvents == 0
                            messageID = 'RTW:autosar:validateEventForStepFunction';
                            message = DAStudio.message(messageID, runnableName);
                            return;
                        end
                        
                        if numDataReceivedEvents > 0 || numDataReceiveErrorEvents > 0 ||...
                                numTimingEvents > 1 ||...
                                numModeSwitchEvents > 0 || numInitEvents > 0 ||...
                                numOpInvokedEvents > 0
                            if length(mapping.StepFunctions)==1
                                messageID = 'RTW:autosar:moreThanOneTimingEvent';
                                message = DAStudio.message(messageID);
                            else
                                messageID = 'RTW:autosar:moreThanOneTimingEventPerStepRunnable';
                                message = DAStudio.message(messageID, runnableName);
                            end                            
                            return;
                        end 
                        if length(mapping.StepFunctions)==1
                            if strcmp(get_param(mapping.Name, 'SampleTimeConstraint' ), 'STIndependent')
                                messageID = 'RTW:autosar:ensureSTIndependentIsOff';
                                message = DAStudio.message(messageID);
                                return;
                            end
                        end                        
                        mappedRunnable = true;
                        
                    end
                end               
            end

            if ~mappedRunnable
                for jj = 1 : numel(mapping.ResetFunctions)
                    if strcmp(runnableName, mapping.ResetFunctions(jj).MappedTo.Runnable)
                        if numTimingEvents > 0
                            messageID = 'autosarstandard:validation:validateEventForResetFunction';
                            message = DAStudio.message(messageID, runnableName);
                            return;
                        end
                        
                        mappedRunnable = true;
                        break;
                    end
                end
            end
            
            if ~mappedRunnable
                for jj = 1 : numel(functionCallInports)
                    if strcmp(runnableName, functionCallInports(jj).MappedTo.Runnable)
                        % Number of DataReceivedEvents in a Runnable (0-n)
                        % Number of DataReceiveErrorEvents in a Runnable (0-n)
                        % Number of TimingEvents in a Runnable (0-1)
                        % Number of OperationInvokedEvents (0)
                        if numTimingEvents > 1
                            messageID = 'RTW:autosar:validateEventTypeForRunnable';
                            message = DAStudio.message(messageID, runnableName,...
                                                       'TimingEvent');
                            return;
                        end
                        if ~autosar.validation.ExportFcnValidator.isExportFcn(hModel) && ...
                                numTimingEvents > 0
                            messageID = 'autosarstandard:validation:noTimingEventForNonPeriodicRunnable';
                            message = DAStudio.message(messageID, runnableName);
                            return;
                        end                        
                        % Number of ModeSwitchEvents in a Runnable (0-1)
                        if  numModeSwitchEvents > 1
                            messageID = 'RTW:autosar:validateEventTypeForRunnable';
                            message = DAStudio.message(messageID, runnableName,...
                                                       'ModeSwitchEvent');
                            return;
                        end
                        % Number of InitEvents in a Runnable (0-1)
                        if  numInitEvents > 1
                            messageID = 'RTW:autosar:validateInitEventForRunnable';
                            message = DAStudio.message(messageID, runnableName);
                            return;
                        end
                        % Number of OperationInvokedEvents in a Runnable (0)
                        if numOpInvokedEvents > 0
                            messageID = 'autosarstandard:validation:validateOperationInvokedEventForRunnable';
                            message = DAStudio.message(messageID, runnableName);
                            return;
                        end
                        
                        mappedRunnable = true;
                        break;
                    end
                end
                if ~mappedRunnable
                    for jj = 1 : numel(mapping.ServerFunctions)
                        if strcmp(runnableName, mapping.ServerFunctions(jj).MappedTo.Runnable)
                            % Number of Timing/Init/ModeSwitch/DataReceivedEvent/DataReceiveErrorEvent (0)
                            if numTimingEvents > 0 || numModeSwitchEvents > 0 ||...
                                    numInitEvents > 0 || numDataReceivedEvents > 0 ||...
                                    numDataReceiveErrorEvents > 0
                                messageID = 'RTW:autosar:validateOpInvokedEventForRunnable';
                                message = DAStudio.message(messageID, runnableName);
                                return;
                            end
                            if numOpInvokedEvents > 1
                                % Multiple OpInvokedEvents are allowed only
                                % with compatible clientserver operations
                                operation = eventArray{1}.instanceRef.Operations;
                                arguments = eventArray{1}.instanceRef.Operations.Arguments;
                                for index=2:length(eventArray)
                                    % check arguments
                                    if arguments.size ~= eventArray{index}.instanceRef.Operations.Arguments.size
                                        messageID = 'RTW:autosar:validateIncompatibleOpInvokedEventForRunnable';
                                        message = DAStudio.message(messageID, runnableName);
                                        return;
                                    end
                                    for j=1:arguments.size
                                        if (eventArray{index}.instanceRef.Operations.Arguments.at(j).Direction...
                                            ~= operation.Arguments.at(j).Direction) || ...
                                                (~strcmp(eventArray{index}.instanceRef.Operations.Arguments.at(j).Name,...
                                                         operation.Arguments.at(j).Name))
                                            messageID = 'RTW:autosar:validateIncompatibleOpInvokedEventForRunnable';
                                            message = DAStudio.message(messageID, runnableName);
                                            return;
                                        end
                                    end
                                end
                            end
                            
                            mappedRunnable = true;
                            break;
                        end
                    end
                end
                
                if mappedRunnable
                    % ModeSwitchEvent activation type should be ?OnTransition?
                    % is not supported for AUTOSAR Schema version less than 4.0
                    if ~AR4.isCompatibleSchemaVer(schemaVersion)
                        for ii = 1 : numel(eventArray)
                            if isa(eventArray{ii}, 'Simulink.metamodel.arplatform.behavior.ModeSwitchEvent')
                                if eventArray{ii}.activation == Simulink.metamodel.arplatform.behavior.ModeActivationKind.OnTransition
                                    messageID = 'RTW:autosar:validateEventPropertyValue';
                                    message = DAStudio.message(messageID, eventArray{ii}.Name,...
                                        runnableName, 'Exported Function', 'Mode Activation',...
                                        '{OnEntry | OnExit}');
                                    return;
                                end
                            end
                        end
                    end
                end
            end
        end
    end
            
end

% LocalWords:   Thru metamodel arplatform  Behav clientserver autosarstandard
% LocalWords:  VERIFYMODESWITCHINTERFACE
