classdef SLPortValidator < autosar.validation.PhasedValidator
% SLPortValidator

%   Copyright 2013-2017 The MathWorks, Inc.
 
    methods (Access=protected)
        
        function verifyPostProp(this, hModel)
            assert(isscalar(hModel) && ishandle(hModel), 'hModel is not a handle');
            
            % Schema version is needed for verifyRootIODataTypes
            schemaVersionHandler = autosar.SchemaVersionHandler.instance();
            schemaVersionHandler.setSchemaVersion(get_param(hModel, 'AutosarSchemaVersion'));

            this.verifyPortSignalObjectsHavePimCSC(hModel);
            this.verifyRootIODataTypes(hModel);
            this.verifyVariantConditions(hModel);
            this.verifyServerSubsystemArg(hModel);
        end
       
    end

    
    methods (Static, Access=private)

        function verifyPortSignalObjectsHavePimCSC(hModel)
        % Function: verifyPortSignalObjectsHavePimCSC =============================
        % Abstract: [Compiled] verify port signal objects are PIM CSCs
            
            % Only need to check outports because inports don't reference signal objects
            ports = find_system(hModel, 'FindAll', 'on', 'LookUnderMasks','on',...
                                'FollowLinks','on', 'Type', 'port', 'PortType', 'outport');
            for portIdx = 1:length(ports)
                dataObj = get_param(ports(portIdx), 'CompiledSignalObject');
                isPIM_CSC = isa(dataObj,'AUTOSAR.Signal') && ...
                    dataObj.getIsAutosarPerInstanceMemory();
                isPIM_CSC2 = isa(dataObj,'Simulink.Signal') && ...
                    strcmp( dataObj.CoderInfo.StorageClass, 'Custom' ) && ...
                    isa(dataObj.CoderInfo.CustomAttributes, 'SimulinkCSC.AttribClass_AUTOSAR_PerInstanceMemory');
                if isPIM_CSC || isPIM_CSC2
                    msg = DAStudio.message(...
                        'RTW:autosar:PortWithPimCSC', ...
                        get_param(ports(portIdx), 'PortNumber'), ...
                        get_param(ports(portIdx), 'Parent'));
                    DAStudio.error('RTW:fcnClass:finish', msg);
                end
            end

        end
       
        function validateVP(m3iObj, arRoot, modelName, blockPath)
            % VALIDATEVP Validate that AUTOSAR VP condition matches a
            % Simulink propagated condition, for a given block and
            % corresponding m3i object.
            %
            % M3IOBJ -- m3i object corresponding to the block at BLOCKPATH
            % MODELNAME -- Name of the model in question
            % BLOCKPATH -- path to the block corresponding to M3IOBJ, e.g.
            %              'subsys/myblock'               
            function result = bool2str(val)
                if val
                    result = 'enabled';
                else
                    result = 'disabled';
                end
            end
            
            exportVPs = autosar.mm.util.XmlOptionsAdapter.get(arRoot, 'ExportPropagatedVariantConditions');
            if strcmp(exportVPs, 'All')
                return;
            end
            
            if isempty(m3iObj) || isempty(m3iObj.variationPoint)
                return
            else
                expected = autosar.mm.util.evaluateSystemConstantExpression(modelName, m3iObj.variationPoint.Condition);
            end
            isEnabled = strcmp(get_param(blockPath, 'CompiledIsActive'), 'on');
            if expected ~= isEnabled
                m3iCond = autosar.mm.util.extractCondExpressionFromM3iCondAccess(m3iObj.variationPoint.Condition);
                slCond = get_param(blockPath, 'CompiledLocalVVCE');
                if isempty(slCond)
                    slCond = 'unconditional';
                end
                DAStudio.error('autosarstandard:validation:VariantConditionMismatch', ...
                    blockPath, m3iCond, bool2str(expected), slCond, bool2str(isEnabled));
            end
        end
        
        function verifyVariantConditions(hModel)
            % VERIFYVARIANTCONDITIONS Compares a block's condition as 
            % specified in the m3i model's variation point, versus the
            % propagated condition from Simulink.  If the two do not match,
            % throws an error.
            modelName = get_param(hModel, 'Name');
            if ~autosar.api.Utils.isMapped(modelName)
                return
            end
            
            mapping = autosar.api.Utils.modelMapping(modelName);
            m3iModel = autosar.api.Utils.m3iModel(modelName);
            arRoot = m3iModel.RootPackage.front();
            m3iComp = autosar.mm.Model.findChildByName(arRoot, arRoot.ComponentQualifiedName);
            
            portDict = containers.Map();
            m3iPortLists = {m3iComp.ReceiverPorts; m3iComp.SenderPorts; ...
                 m3iComp.SenderReceiverPorts; m3iComp.ModeReceiverPorts; ...
                 m3iComp.ClientPorts; m3iComp.ServerPorts; ...
                 m3iComp.NvReceiverPorts; m3iComp.NvSenderPorts; ...
                 m3iComp.NvSenderReceiverPorts; m3iComp.ModeSenderPorts;};
            for ii = 1:length(m3iPortLists)
                m3iPortList = m3iPortLists{ii};
                for jj = 1:m3iPortList.size
                    m3iPort = m3iPortList.at(jj);
                    portDict(m3iPort.Name) = m3iPort;
                end
            end
            
            portMappings = [mapping.Inports mapping.Outports];
            for ii = 1:length(portMappings)
                portMap = portMappings(ii);
                
                if ~portMap.IsActive || isempty(portMap.MappedTo) || isempty(portMap.MappedTo.Port)
                    continue;
                end
                
                m3iObj = portDict(portMap.MappedTo.Port);
                autosar.validation.SLPortValidator.validateVP(m3iObj, arRoot, modelName, portMap.Block);
            end
            
            for ii = 1:length(mapping.FcnCallInports)
                fcnInportMap = mapping.FcnCallInports(ii);
                
                if ~fcnInportMap.IsActive || isempty(fcnInportMap.MappedTo) || isempty(fcnInportMap.MappedTo.Runnable)
                    continue;
                end
                
                m3iObj = Simulink.metamodel.arplatform.ModelFinder.findNamedItemInSequence(...
                    m3iComp.Behavior, ...
                    m3iComp.Behavior.Runnables, ...
                    fcnInportMap.MappedTo.Runnable,'Simulink.metamodel.arplatform.behavior.Runnable');
                autosar.validation.SLPortValidator.validateVP(m3iObj, arRoot, modelName, fcnInportMap.Block);
            end
        end
        
        function verifyRootIODataTypes(hModel)
        % Function: verifyRootIODataTypes =============================
        % Abstract: [Compiled] verify root IO data types are correct
            
            inpH = find_system(hModel, 'SearchDepth', 1, 'Type', 'block',...
                               'BlockType', 'Inport');
            outpH = find_system(hModel, 'SearchDepth', 1, 'Type', 'block',...
                                'BlockType', 'Outport');

            inoutH = [inpH; outpH];
            isInportArray = [ones(size(inpH)); zeros(size(outpH))];
            
            cs = autosar.validation.ConfigSetValidator.getActiveConfigSet(hModel);
            supportMatrixIOAsArray = strcmpi(...
                get_param(cs,'AutosarMatrixIOAsArray'), 'on');
            modelName = get_param(hModel, 'Name');
            enablePRPortChecking = false;
            
            if autosar.validation.SLPortValidator.isPRPortEnable(modelName)
                mapping = autosar.api.getSimulinkMapping(modelName);
                dataObj = autosar.api.getAUTOSARProperties(modelName, true);
                prportMap = containers.Map();
                enablePRPortChecking = true;
            end
            portInfo = cell(length(inoutH), 6);
            for i=1:length(inoutH)
                if isInportArray(i)
                    isInport = true;
                    portH = inoutH(i);
                    portHDims = get_param(portH, 'CompiledPortDimensions');
                    if ~supportMatrixIOAsArray && ...
                            (portHDims.Outport(1) > 1) && ...
                            (sum(portHDims.Outport(2:end) ~= 1) > 1)
                        msg = DAStudio.message('RTW:autosar:scalarOnly', ...
                                               get_param(portH, 'Name'));
                        DAStudio.error('RTW:fcnClass:finish', msg);
                    end

                    % Get the DataType and Width
                    portHDataTypes = get_param(portH, 'CompiledPortAliasedThruDataTypes');
                    portHDataTypesAliased = get_param(portH, 'CompiledPortDataTypes');
                    portHWidths = get_param(portH, 'CompiledPortWidths');
                    portInfo{i,1} = portHDataTypes.Outport{1};
                    portInfo{i,2} = portHWidths.Outport(1);
                    portInfo{i,3} = isInport;
                    portInfo{i,4} = get_param(portH, 'CompiledSampleTime');
                    portInfo{i,5} = arblk.convertPortNameToArgName(get_param(portH, 'Name'));
                    portInfo{i,6} = portHDataTypesAliased.Outport{1};
                    if enablePRPortChecking
                        autosar.validation.SLPortValidator.checkPRPort( ...
                            dataObj, mapping, prportMap, ...
                            get_param(portH, 'Name'), portInfo(i, :));
                    end
                else %Outport
                    isInport = false;
                    portH = inoutH(i);
                    portHDims = get_param(portH, 'CompiledPortDimensions');
                    if ~supportMatrixIOAsArray && ...
                            (portHDims.Inport(1) > 1) && ...
                            (sum(portHDims.Inport(2:end) ~= 1) > 1)
                        msg = DAStudio.message('RTW:autosar:scalarOnly', get_param(portH, 'Name'));
                        DAStudio.error('RTW:fcnClass:finish', msg);
                    end

                    % Get the DataType and Width
                    portHDataTypes = get_param(portH, 'CompiledPortAliasedThruDataTypes');
                    portHDataTypesAliased = get_param(portH, 'CompiledPortDataTypes');
                    portHWidths = get_param(portH, 'CompiledPortWidths');
                    portInfo{i,1} = portHDataTypes.Inport{1};
                    portInfo{i,2} = portHWidths.Inport(1);
                    portInfo{i,3} = isInport;
                    portInfo{i,4} = get_param(portH, 'CompiledSampleTime');
                    portInfo{i,5} = arblk.convertPortNameToArgName(get_param(portH, 'Name'));
                    portInfo{i,6} = portHDataTypesAliased.Inport{1};       
                    if enablePRPortChecking
                        autosar.validation.SLPortValidator.checkPRPort( ...
                            dataObj, mapping, prportMap, ...
                            get_param(portH, 'Name'), portInfo(i, :));
                    end
                end

                portName = get_param(portH, 'Name');
                portHandles= get_param(portH, 'PortHandles');
                if isInport
                    thePort = portHandles.Outport;
                else
                    thePort = portHandles.Inport;
                end

                % name & Rte_Type.h check
                dataTypeName = get_param(thePort, 'CompiledPortDataType');
                cs = autosar.validation.ConfigSetValidator.getActiveConfigSet(hModel);
                maxShortNameLength = get_param(cs, 'AutosarMaxShortNameLength');
                autosar.validation.AutosarUtils.checkDataType(hModel, ['Port ', portName], dataTypeName, maxShortNameLength, supportMatrixIOAsArray);

                % Variable-size signal check
                dimsMode = get_param(thePort, 'CompiledPortDimensionsMode');
                if any(dimsMode)
                    msg = DAStudio.message('RTW:autosar:variableSizeSignal',...
                                           portName);
                    DAStudio.error('RTW:fcnClass:finish', msg);
                end

                % Check that ports don't have non-auto storage class
                % For non-interactive validation of non-server, a hidden buffer block has
                % been inserted and in order to be consistent, we need to
                % look at its input port
                theRTWPort = thePort;
                csc = get_param(theRTWPort, 'CompiledRTWStorageClass');
                if ~isempty(csc) && ~strcmp(csc, 'Auto')
                    msg = DAStudio.message('RTW:autosar:nonAutoStorageClass',...
                                           portName);
                    DAStudio.error('RTW:fcnClass:finish', msg);
                end

            end
            
            if AR4.isCompatibleSchemaVer(autosar.SchemaVersionHandler.instance().getSchemaVersion)
                autosar.validation.SLPortValidator.checkApplImplClash(hModel, portInfo);
            end
        end
            
        function verifyServerSubsystemArg(hModel)
            allSubSystemsH = find_system(hModel, ...
                'Type', 'block', 'BlockType', 'SubSystem');
            for i = 1:length(allSubSystemsH)
                if autosar.validation.ExportFcnValidator.isServerSubSys(allSubSystemsH(i))
                    autosar.validation.SLPortValidator.checkSubsystemArgDataType(...
                        hModel, get_param(allSubSystemsH(i), 'Handle'));
                end
            end
        end
    end
    
    methods(Static, Access=private)      
        
        function checkSubsystemArgDataType(hModel, ssBlock)
            cs = autosar.validation.ConfigSetValidator.getActiveConfigSet(hModel);
            maxShortNameLength = get_param(cs, 'AutosarMaxShortNameLength');
            supportMatrixIOAsArray = strcmpi(...
                get_param(cs,'AutosarMatrixIOAsArray'), 'on');
            
            % Get the argins, and argouts of the ssBlock.
            inArgs = find_system(ssBlock, 'SearchDepth', 1, ...
                'BlockType', 'ArgIn');
            outArgs = find_system(ssBlock, 'SearchDepth', 1, ...
                'BlockType', 'ArgOut');
            allSSinout = [inArgs; outArgs];
            for ii = 1:length(allSSinout)
                blkObj = get_param(allSSinout(ii), 'Object');
                blkType = blkObj.BlockType;
                blkPortH = blkObj.PortHandles;
                blkPath = blkObj.getFullName;
                if strcmp(blkType, 'ArgIn')
                    thePort = blkPortH.Outport;
                else
                    thePort = blkPortH.Inport;
                end
                dataTypeName = get_param(thePort, 'CompiledPortDataType');
                if autosar.validation.AutosarUtils.isErrorArgument(hModel, blkPath)
                    autosar.validation.AutosarUtils.checkDataTypeForErrArg(hModel, ...
                        blkPath, dataTypeName, maxShortNameLength);
                else
                    autosar.validation.AutosarUtils.checkDataType(hModel, ...
                        blkPath, dataTypeName, maxShortNameLength, ...
                        supportMatrixIOAsArray);
                end
            end
            
        end
        
        function enable = isPRPortEnable(modelName)
        % Function: isPRPortEnable =============================
        % Abstract: check if prport is enabled
            if ~autosar.api.Utils.isMapped(modelName)
                enable = false;
                return
            end
            
            dataObj = autosar.api.getAUTOSARProperties(modelName, true);
            componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
            prports = dataObj.find(componentQualifiedName, 'DataSenderReceiverPort', ...
                'PathType', 'FullyQualified');
            
            if isempty(prports)
                enable = false;
            else
                enable = true;
            end
        end
        
        function checkPRPort(dataObj, mapping, prportMap, portName, portInfo)
            % Condition: version >= 4.1; not fcn call; not error status
            % NOTE error status and version will be checked later
            if strcmp(portInfo{1}, 'fcn_call')
                return;
            end
            % Get AUTOSAR info for the port
            if portInfo{3}
                [ARPortName, ARElementName, ARDataAccessMode] = ...
                            mapping.getInport(portName);
            else
                [ARPortName, ARElementName, ARDataAccessMode] = ...
                    mapping.getOutport(portName);
            end
            
            % Check if [ARPortName, ARElementName] mapped to PRPort
            compQName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
            portObj = dataObj.find(compQName, 'DataSenderReceiverPort', ...
                'PathType', 'FullyQualified', 'Name', ARPortName);
            if isempty(portObj)
                return;
            end
            interfaceName = dataObj.get(portObj{1}, 'Interface', ...
                'PathType', 'FullyQualified');
            des = dataObj.get(interfaceName, 'DataElements', ...
                'PathType', 'FullyQualified');
            foundOne = false;
            for i = 1: length(des)
                desName = dataObj.get(des{i}, 'Name');
                if strcmp(desName, ARElementName)
                    foundOne = true;
                end
            end
            % If not a PRPort then just return
            if ~foundOne
                return;
            end
            
            % Check DataAccessMode, can be either implicit, explicit or 
            % errorStatus
            if strncmpi(ARDataAccessMode, 'implicit', 8)
                ARDataAccessMode = 'implicit';
            elseif strncmpi(ARDataAccessMode, 'explicit', 8)
                ARDataAccessMode = 'explicit';
            elseif strcmpi(ARDataAccessMode, 'errorstatus')
                % error status is ignored because we can't pair it
                return; 
            elseif strcmpi(ARDataAccessMode, 'IsUpdated')
                % PR ports are not valid for IsUpdated access mode.
                DAStudio.error('autosarstandard:validation:isUpdatedInvalidSenderReceiverPort', ...
                    portName);
            elseif strncmpi(ARDataAccessMode, 'EndToEnd', 8)
                ARDataAccessMode = 'explicit';
            else
                DAStudio.error('RTW:autosar:invalidPRPortDAM', ...
                    portName, ARPortName, ARElementName);
            end
            
            % 1st time meet a PRPort, insert; 
            % 2nd time meet a PRPort, get the pair and check.
            key = [ARPortName, ARElementName];
            if ~prportMap.isKey(key)
                prportMap(key) = {portName, portInfo, ARDataAccessMode};%#ok<NASGU>   
            else           
                thatVal = prportMap(key);
                thatPortName = thatVal{1};
                thatPortInfo = thatVal{2};
                thatDAM = thatVal{3};
                
                % NOTE check order is important, DO NOT change the order
                
                % Check port type, they cannot be both inport or both outport
                if portInfo{3} == thatPortInfo{3}
                    DAStudio.error('RTW:autosar:invalidPRPortType', ...
                        portName, thatPortName, ARPortName, ARElementName);
                end
                
                % Check DataAccessMode, they should be the same
                if ~strcmp(ARDataAccessMode, thatDAM)
                    DAStudio.error('RTW:autosar:differentPRPortDAM', ...
                        portName, thatPortName, ARPortName, ARElementName);
                end
                
                % Check Data types, aliased type and dimension,
                % they should be exactly the same
                if ~strcmp(portInfo{1}, thatPortInfo{1}) || ...
                        ~strcmp(portInfo{6}, thatPortInfo{6}) || ...
                        portInfo{2} ~= thatPortInfo{2}
                    DAStudio.error('RTW:autosar:invalidPRPortDataType', ...
                        portName, thatPortName, ARPortName, ARElementName);
                end
                
                % Remove key so that at the end we can check if there's any
                % dangling prport
                prportMap.remove(key);
            end           
        end
        
        function checkApplImplClash(hModel, portInfo)
        % Function: checkApplImplClash =============================================
        % Abstract: Helper function to check application to implementation data type
        % clashes.
            
            appl2impl = containers.Map();
            if autosar.api.Utils.isMapped(hModel)
                appl2impl = autosar.api.Utils.app2ImpMap(hModel);
            end
    
            alias2base = containers.Map();
            for ii=1:size(portInfo,1)
                baseDataType = portInfo{ii,1};
                aliasDataType = portInfo{ii,6};
                mprops = Simulink.getMetaClassIfValidEnumDataType(baseDataType);
                if ~isempty(mprops)
                    storageType = Simulink.data.getEnumTypeInfo(baseDataType, 'StorageType');
                    alias2base(aliasDataType) = storageType;
                else
                    alias2base(aliasDataType) = baseDataType;
                end
            end

            applDataTypes = appl2impl.keys;
            for ii=1:length(applDataTypes)
                applDataType = applDataTypes{ii};
                implDataType = appl2impl(applDataType);
        
                if alias2base.isKey(applDataType) && alias2base.isKey(implDataType)
                    applBaseType = alias2base(applDataType);
                    implBaseType = alias2base(implDataType);
        
                    if ~autosar.mm.util.BuiltInTypeMapper.isFixPtTypeEquivalent(applBaseType, implBaseType)
                        DAStudio.error('RTW:autosar:appl2implClash', applDataType, implDataType, applBaseType, implBaseType);
                    end
                end
            end
        end

    end
        
end

% LocalWords:  Attrib Thru Appl appl impl prport errorstatus nd argins argouts
