classdef Validator < autosar.validation.PhasedValidator
% Validator

%   Copyright 2013-2017 The MathWorks, Inc.

    properties (Access=private)
        cleanupObj
    end
    
    methods (Access=public)

        function verify(this, aiOrModelH, varargin)
        % NOTE: ai will still be passed in for UI validation for top-model
        % ServerOperation and VirtualSubsystemExportFunctions.
            cleanupVerifier = onCleanup(@() delete(this.cleanupObj));
            % Parse inputs
            p = inputParser;
            p.KeepUnmatched = false;
            p.FunctionName = 'verify';
            p.addParameter( 'ValidationPhase', 'All', @(x)any(validatestring(x,{'Initial','PostProp','Final','All'})));
            p.parse( varargin{:} );
            validationPhase = p.Results.ValidationPhase;
            
            % Check the modeling style first
            autosar.validation.ClientServerValidator.checkValidRunnableConfig(aiOrModelH);
                        
            if strcmp(validationPhase, 'All')
                % Explicitly call Initial validation phase
                this.verifyPhase('Initial', aiOrModelH);

                % Compile model which also calls PostProp and Final validation stages
                this.cleanupObj = autosar.validation.CompiledModelUtils.forceCompiledModelForRTW(aiOrModelH);
                
                % Sanity check that model was compiled
                assert(~isempty(this.cleanupObj), 'Expected non-empty cleanup object');
            else
                this.verifyPhase(validationPhase, aiOrModelH);
            end
        end
        
        function verifyPhase(~, validationPhase, hModel)
        % Override verifyPhase because at top-level we call all validators regardless of
        % validation phase and let the individual validators react to the
        % different phases.
            
            assert((ischar(hModel) || isStringScalar(hModel)) || ishandle(hModel), 'Expected a model handle object');

            if ~autosar.api.Utils.isMapped(hModel)
                DAStudio.error('Simulink:Engine:RTWCGAutosarEmptyConfigurationError');
            end
            
            configSetValidator = autosar.validation.ConfigSetValidator();
            configSetValidator.verifyPhase(validationPhase, hModel);
            
            configSetFinalValidator = autosar.validation.ConfigSetFinalValidator();
            configSetFinalValidator.verifyPhase(validationPhase, hModel);
                
            slPortValidator = autosar.validation.SLPortValidator();
            slPortValidator.verifyPhase(validationPhase, hModel);
            
            dataObjectValidator = autosar.validation.DataObjectValidator();
            dataObjectValidator.verifyPhase(validationPhase, hModel);

            % Legacy client block checks
            clientOperationValidator = autosar.validation.ClientOperationValidator();
            clientOperationValidator.verifyPhase(validationPhase, hModel);

            if autosar.validation.ExportFcnValidator.isExportFcn(hModel)
                exportFcnValidator = autosar.validation.ExportFcnValidator();
                
                if autosar.validation.ExportFcnValidator.isTopModelExportFcn(hModel)
                    % Top model export functions
                    exportFcnValidator.verifyPhase(validationPhase, hModel, hModel);
                    
                elseif autosar.validation.ExportFcnValidator.isVirtualSubsysExportFcn(hModel)
                    % Classic virtual subsystem export functions
                    newSubsysHdl = get_param(hModel, 'NewSubsystemHdlForRightClickBuild');
                    origSSHdl    = get_param(hModel, 'SubsystemHdlForRightClickBuild');
                    
                    try
                        exportFcnValidator.verifyPhase(validationPhase, newSubsysHdl, origSSHdl);
                    catch ME
                        % Make sure error message refers to actual model not hidden model
                        message = autosar.validation.ExportFcnValidator.convertErrMsgToUseActModel(ai, ME.message);
                        newME = MException(ME.identifier, message);
                        newME.throw();
                    end
                else
                    assert(false, 'Did not recognize ExportFcn mode');
                end
            end
                
            % Check mapping before metamodel
            if autosar.api.Utils.isMapped(hModel)                    
                % Basic Software checks
                basicSoftwareValidator = autosar.validation.BasicSoftwareValidator();
                basicSoftwareValidator.verifyPhase(validationPhase, hModel);
                
                simulinkMappingValidator = autosar.validation.SimulinkMappingValidator();
                simulinkMappingValidator.verifyPhase(validationPhase, hModel);
                
                % Client-Server Modeling checks only valid if the model has
                % a mapping.
                clientServerValidator = autosar.validation.ClientServerValidator();
                clientServerValidator.verifyPhase(validationPhase, hModel);

                metaModelValidator = autosar.validation.MetaModelValidator();
                metaModelValidator.verifyPhase(validationPhase, hModel);
            end
            
        end
        
    end

end

% LocalWords:  ai validators HAnd metamodel ARGUI
