classdef OperationBuilder < m3i.Visitor
    % OPERATIONBUILDER
    %
    % Copyright 2017 The MathWorks, Inc.
    properties (SetAccess = immutable, GetAccess = private)
        SourceFile
        TypeBuilder
    end
    
    properties (Access = private)
        CodeInfo % Handle to CodeInfo
    end
    
    methods (Access = public)
        function obj = OperationBuilder(lSourceFile, lCodeInfo, lTypeBuilder)
            % OPERATIONBUILDER
            obj.SourceFile = lSourceFile;
            obj.CodeInfo = lCodeInfo;
            obj.TypeBuilder = lTypeBuilder;
            
            % Register visitor for the AUTOSAR side
            obj.registerVisitor('mmVisit', 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.behavior.OperationBlockingAccess', @mmWalkOperationBlockingAccess, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.instance.OperationPortInstanceRef', @mmWalkOperationPortInstanceRef, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.interface.Operation', @mmWalkOperation, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.interface.ArgumentData', @mmWalkArgumentData, 'mmVisit');
        end
        
        function processOperation(obj, m3iOperation, lCallObject)
            % PROCESSOPERATION
            obj.apply('mmVisit', m3iOperation, lCallObject);
        end
        
        function processOperationBlockingAccess(obj, m3iBlockingAccess, lTimingInterface)
            % PROCESSOPERATIONBLOCKINGACCESS
            obj.applySeq('mmVisit', m3iBlockingAccess, lTimingInterface);
        end
    end
    
    
    methods (Access = public)
        function ret = mmVisitM3IObject(~, m3iObj, varargin)
            % MMVISITM3IOBJECT
            %
            % Note: this cannot be a true static method, because M3I is
            % expecting a particular function signature.  Default visitor
            % must be public.
            %
            ret = [];
            assert(false, 'Failure in: %s\n Unsupported M3I object class %s', mfilename('-class'), class(m3iObj));
        end
    end
    
    methods (Access = private)    
        function ret = mmWalkOperationBlockingAccess(obj, m3iBlockingAccess, lTimingInterface)
            % MMWALKOPERATIONBLOCKINGACCESS
            ret = [];
            obj.applySeq('mmVisit', m3iBlockingAccess.instanceRef, lTimingInterface);
        end
        
        function ret = mmWalkOperationPortInstanceRef(obj, m3iOperationInstance, lTimingInterface)
            % MMWALKOPERATIONPORTINSTANCEREF
            ret = [];
            m3iPort = m3iOperationInstance.Port;
            lCallObject = RTW.AutosarClientCall;
            lCallObject.PortName = m3iPort.Name;
            lCallObject.Timing = lTimingInterface;
            obj.apply('mmVisit', m3iOperationInstance.Operations, lCallObject);
            obj.CodeInfo.ServerCallPoints(end+1) = lCallObject;
        end
        
        
        function ret = mmWalkOperation(obj, m3iOperation, lCallObject)
            % MMWALKOPERATION
            ret = [];
            
            % This method can be reached by two paths - one for client
            % calls and one for server functions.  
            lIsCaller = isa(lCallObject, 'RTW.AutosarClientCall');
            if lIsCaller
                % For client calls, we need to create a new prototype
                % object.
                lPrototype = coder.types.Prototype;
                lPrototype.Name = m3iOperation.Name;
                lPrototype.SourceFile = obj.SourceFile;
                lCallObject.Prototype = lPrototype;
            else
                % For server functions, the prototype has already been
                % partly filled in with the function name, which is not
                % otherwise directly available at this level from the
                % m3iOperation object.  So we need to stick with what's
                % already been populated and augment it.
                lPrototype = lCallObject.Prototype;
            end
            obj.applySeq('mmVisit', m3iOperation.Arguments, lIsCaller, lPrototype);
            
        end
        
        function ret = mmWalkArgumentData(obj, m3iArgumentData, lIsCaller, lPrototype)
            % MMWALKARGUMENTDATA
            ret = [];
            m3iType = m3iArgumentData.Type;
            switch m3iArgumentData.Direction.toString
                case 'In'
                    lIOType = 'INPUT';
                    assert(~isempty(m3iType));
                    lM3iTypeName = m3iType.Name;
                    lCodeInfoBaseType = obj.TypeBuilder.processType(m3iType);
                    lArgProp = 'Arguments';
                    lArgumentName = m3iArgumentData.Name;
                case 'Out'
                    lIOType  = 'OUTPUT';
                    assert(~isempty(m3iType));
                    lM3iTypeName = m3iType.Name;
                    lCodeInfoBaseType = obj.TypeBuilder.processType(m3iType);
                    lArgProp = 'Arguments';
                    lArgumentName = m3iArgumentData.Name;
                case 'Error'
                    lIOType  = 'OUTPUT';
                    assert(isempty(m3iType));
                    lM3iTypeName = 'UInt8';
                    lCodeInfoBaseType = obj.TypeBuilder.getUInt8();
                    lArgProp = 'Return';
                    lArgumentName = ''; % No name for return type
                otherwise
                    assert(false, 'Direction should be In or Out or Error');
            end
            
            % If a parameter is not a matrix, we need to load it into a
            % 2-D matrix if part of a function caller interface
            if lIsCaller
                lCodeInfoMatrixType = obj.TypeBuilder.wrapScalarWithMatrix(lCodeInfoBaseType, lM3iTypeName);
            else
                lCodeInfoMatrixType = lCodeInfoBaseType;
            end
            
            if strcmp(m3iArgumentData.Direction.toString, 'Out')
                % Outputs are pointers.
                lCodeInfoType = coder.types.Pointer;
                lCodeInfoType.BaseType = lCodeInfoMatrixType;
            else
                lCodeInfoType = lCodeInfoMatrixType;
            end
            
            lArgument = coder.types.Argument;
            lArgument.Type = lCodeInfoType;
            lArgument.IOType = lIOType;
            lArgument.Name = lArgumentName;
            lPrototype.(lArgProp)(end+1) = lArgument;
        end
    end
end