classdef AutosarExternalCodeImportHook < coder.internal.xrel.CodeImportHook
    % AUTOSAREXTERNALCODEIMPORTHOOK
    %
    %   Copyright 2017 The MathWorks, Inc.
    
    methods (Access=public, Static)
        function entryHook(lModelName)
            import coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook;
            if ~AutosarExternalCodeImportHook.isAutosarCodeImport(lModelName)
                return;
            end
            
            % Make sure block reduction is OFF
            set_param(lModelName, 'BlockReduction', 'off');
            
            % Make sure block IO reuse is OFF
            set_param(lModelName, 'OptimizeBlockIOStorage', 'off');
            
            % Make sure that model-to-code traceability is OFF
            lParamsToSwitchOff = { ...
                'IncludeHyperlinkInReport', ...
                'GenerateTraceInfo', ...
                'GenerateTraceReport', ...
                'GenerateTraceReportSl', ...
                'GenerateTraceReportSf', ...
                'GenerateTraceReportEml', ...
                'GenerateCodeReplacementReport'};
            
            for kParam = 1:numel(lParamsToSwitchOff)
                set_param(lModelName, lParamsToSwitchOff{kParam}, 'off');
            end
            
            set_param(lModelName, 'Dirty', 'off');
        end
        
        
        function preTlcHook(lModelName)
            import coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook;
            if ~AutosarExternalCodeImportHook.isAutosarCodeImport(lModelName)
                return;
            end
            lRTWGenSettings = get_param(lModelName, 'RTWGenSettings');
            lModelReferenceBuild = isfield(lRTWGenSettings, 'mdlRefRelativeSimDir');
            if lModelReferenceBuild
                DAStudio.error('PIL:xrelimport:TopModelBuildOnly', lModelName);
            end
        end
        
        function postTlcHook(lModelName, lBuildInfo)
            import coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook;
            import coder.internal.xrel.AUTOSAR.ModelBlockSourcesHelper;
            
            [lIsAutosarCodeImport, lControlBlockHandle] = AutosarExternalCodeImportHook.isAutosarCodeImport(lModelName);
            if ~lIsAutosarCodeImport
                return;
            end
            
            lBuildFolderInfo = RTW.getbuildDir(lModelName);
            lBuildFolder = lBuildFolderInfo.BuildDirectory;
            lStubFolder = fullfile(lBuildFolder, 'stub');
            
            % Expunge generated code from buildInfo
            for lKind = {'Src', 'Inc'}
                lGroup = get(lBuildInfo.(lKind{:}).Files, 'Group');
                lFileBuildDir = strcmp(lGroup, 'BuildDir') | strcmp(lGroup, 'ModelSources');
                lBuildInfo.(lKind{:}).Files(lFileBuildDir) = [];
            end
            
            % Get the list of source and include files that need to be
            % added to the current BuildInfo
            bReplaceStartDir = true;
            lModelBlockSources = ModelBlockSourcesHelper.createFromBlock(lControlBlockHandle, bReplaceStartDir);
            [lSourcePaths, lIncludePaths, ...
                lSourceFiles, lIncludeFiles, ...
                lSourceStubs, lIncludeStubs, ...
                lIncludeProxies] = lModelBlockSources.getData;
            
            % Write proxy include files to the RTE headers for files where
            % the custom code is expecting a header file with a different
            % name.
            for kFile = 1:numel(lIncludeProxies)
                AutosarExternalCodeImportHook.writeProxy(lIncludeProxies{kFile}, lStubFolder);
            end
            
            % Write proxy model.h include file if necessary, and add to
            % BuildInfo.
            lModelH = [lModelName, '.h'];
            lDir = coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook.cellDir(lIncludePaths);
            lDirNames = {lDir.name};
            if ~any(strcmp(lModelH, lDirNames))
                % If lModelName.h is not in the list of sources that we are
                % importing, write a proxy and add it to BuildInfo.
                % Otherwise it can be added in the normal way.
                AutosarExternalCodeImportHook.writeModelHProxy(lModelH, lBuildFolder, lIncludeFiles);
                lBuildInfo.addIncludeFiles(lModelH, lBuildFolder, 'BuildDir');
            end
            
            doCopy = true;
            
            if doCopy
                % Copy source files and add to BuildInfo
                AutosarExternalCodeImportHook.copyToBuildFolder(lSourcePaths, lSourceFiles, lBuildFolder);
                AutosarExternalCodeImportHook.copyToBuildFolder(lIncludePaths, lIncludeFiles, lBuildFolder);
                AutosarExternalCodeImportHook.copyToBuildFolder(lSourcePaths, lSourceStubs, lStubFolder);
                AutosarExternalCodeImportHook.copyToBuildFolder(lIncludePaths, lIncludeStubs, lStubFolder);
                lBuildInfo.addSourceFiles(lSourceFiles, lBuildFolder, 'BuildDir');
                lBuildInfo.addSourceFiles(lSourceStubs, lStubFolder, 'BuildDir');
                lBuildInfo.addIncludeFiles(lIncludeFiles, lBuildFolder, 'BuildDir');
                lBuildInfo.addIncludeFiles(lIncludeStubs, lStubFolder, 'BuildDir');
            else
                lSourceFileFolders = AutosarExternalCodeImportHook.getFilePaths(lSourcePaths, lSourceFiles);
                lIncludeFileFolders = AutosarExternalCodeImportHook.getFilePaths(lIncludePaths, lIncludeFiles);
                lSourceStubFolders = AutosarExternalCodeImportHook.getFilePaths(lSourcePaths, lSourceStubs);
                lIncludeStubFolders = AutosarExternalCodeImportHook.getFilePaths(lIncludePaths, lIncludeStubs);
                lBuildInfo.addSourceFiles(lSourceFiles, lSourceFileFolders, 'BuildDir');
                lBuildInfo.addSourceFiles(lSourceStubs, lSourceStubFolders, 'BuildDir');
                lBuildInfo.addIncludeFiles(lIncludeFiles, lIncludeFileFolders, 'BuildDir');
                lBuildInfo.addIncludeFiles(lIncludeStubs, lIncludeStubFolders, 'BuildDir');
                lBuildInfo.addIncludePaths(lIncludePaths, 'BuildDir');
                lBuildInfo.addSourcePaths(lSourcePaths, 'BuildDir');
            end
            lBuildInfo.addIncludeFiles(lIncludeProxies, lStubFolder, 'BuildDir');
            
            % Get the stuff from userdata
            lUserData = get_param(lControlBlockHandle, 'UserData');
            lInitRunnable = lUserData.InitRunnable;
            lXrelInfo = lUserData.XrelInfo;
            
            % Ignore the CodeInfo from code generation, and build our own
            lCodeInfo = load('codeInfo');
            lCodeInfo.codeInfo = RTW.ComponentInterface;            
            m3iModel = autosar.api.Utils.m3iModel(lModelName);
            m3iComp = autosar.mm.Model.findChildByName(m3iModel, m3iModel.RootPackage.at(1).ComponentQualifiedName);
            lCodeInfoWriter = coder.internal.xrel.AUTOSAR.mm2CodeInfo.CodeInfoBuilder(lCodeInfo.codeInfo, m3iModel, m3iComp, lInitRunnable, lXrelInfo);
            lCodeInfoWriter.updateCodeInfo;
            lCodeInfo.expInports = lCodeInfoWriter.getExpInports;
            
            % Needed to write SIL header
            lCodeInfo.codeInfo.GraphicalPath = lModelName;
            save('codeInfo.mat', '-struct', 'lCodeInfo');
            
            % Patch the m3i model if there are any C-typed PIMs - this can
            % only be done after we've written CodeInfo
            nPIMs = m3iComp.Behavior.PIM.size;
            if nPIMs > 0
                lAutosarSchemaVersion = get_param(lModelName, 'AutosarSchemaVersion');
                lMaxShortNameLength = get_param(lModelName, 'AutosarMaxShortNameLength');
                m3iBehavior = m3iComp.Behavior;
                tran = M3I.Transaction(m3iModel);
                lCPimUtils = coder.internal.xrel.AUTOSAR.CTypePimUtils(lModelName, ...
                    lAutosarSchemaVersion, ...
                    lMaxShortNameLength, ...
                    m3iModel, m3iBehavior);
                for kPIM = 1:nPIMs
                    m3iPIM = m3iComp.Behavior.PIM.at(kPIM);
                    lCPimUtils.processPIM(m3iPIM, lCodeInfo.codeInfo);
                end
                tran.commit();
            end
        end        
        
        function accelModeSimulationCallback(lModelName)
            % Callback to indicate that accelerator mode is not permitted
            import coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook;
            h = coder.internal.ModelCodegenMgr.getInstance(lModelName);
            if ~isempty(h)
                assert(isa(h, 'coder.internal.ModelCodegenMgr'));
                isAccelOrRaccel = ...
                    any(ismember(h.MdlRefSimModeForXRelBlock, {'Accelerator', 'accelerator', 'rapid-accelerator'}));
                if isAccelOrRaccel
                    DAStudio.error('PIL:xrelimport:SILPILOnly', lModelName);
                end
            end
        end
        
        function normalModeSimulationCallback(lModelName)
            % Called by StartFcn hook.
            DAStudio.error('PIL:xrelimport:SILPILOnly', lModelName);
        end
                
        function writeModelHProxy(lFile, lFolder, lFilesToInclude)
            % Write proxy header file lFile in folder lFolder
            lFileName = fullfile(lFolder, lFile);
            lWriter = rtw.connectivity.CodeWriter.create(...
                'language', 'C', ...
                'filename', lFileName, ...
                'callCBeautifier', true);
            for kFiles = 1:numel(lFilesToInclude)
                lWriter.wLine('#include "%s"', lFilesToInclude{kFiles});
            end
            lWriter.close;
        end
        
        function writeProxy(lFile, lFolder)
            % Write proxy header file lFile in folder lFolder
            lFileName = fullfile(lFolder, lFile);
            lWriter = rtw.connectivity.CodeWriter.create(...
                'language', 'C', ...
                'filename', lFileName, ...
                'callCBeautifier', true);
            lWriter.wLine('#include "Rte_Type.h"');
            lWriter.close;
        end
        
        function copyToBuildFolder(lPaths, lFiles, lDest)
            % Copy lFiles (cellstr, name only) from search path lPaths
            % (cellstr) to folder lDest (char)
            fhOpToPerform = @(lDirElem)copyfile(fullfile(lDirElem.folder, lDirElem.name), lDest);
            fhElseIfNotFound = @(currName)~(exist(fullfile(lDest, currName), 'file') == 2);
            coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook.processFilePaths(lPaths, lFiles, fhOpToPerform, fhElseIfNotFound);
        end
        
        function lFilePaths = getFilePaths(lPaths, lFiles)
            % Get the path for each of the files and return as a cell array
            fhOpToPerform = @(lDirElem)lDirElem.folder;
            fhElseIfNotFound = @(currName)true;
            lFilePaths = coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook.processFilePaths(lPaths, lFiles, fhOpToPerform, fhElseIfNotFound);
        end
        
        function lRetVal = processFilePaths(lPaths, lFiles, fhOpToPerform, fhElseIfNotFound)
            % Look over files and run fhOpToPerform on the dir entry
            lDir = coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook.cellDir(lPaths);
            lDirNames = {lDir.name};
            nFiles = numel(lFiles);
            lRetVal = cell(1, nFiles);
            for kFile = 1:nFiles
                currName = lFiles{kFile};
                idx = find(strcmp(lDirNames, currName), 1, 'first');
                if ~isempty(idx)
                    lRetVal = fhOpToPerform(lDir(idx));
                elseif fhElseIfNotFound(currName)
                    DAStudio.error('PIL:xrelimport:RequiredFileNotFound', currName, lModelName);
                end
            end
        end
        
        function lDir = cellDir(lPathCell)
            % Concatenated directory listing for all folders in cellstr
            % lPathCell
            lDir = struct('name', {}, 'folder', {}, 'date', {}, 'bytes', {}, 'isdir', {}, 'datenum', {});
            for kElem = 1:numel(lPathCell)
                lDir = [lDir; dir(lPathCell{kElem})]; %#ok<AGROW>
            end
        end             
    end
end
