classdef AutosarImportModelManager < handle
    % AUTOSARIMPORTMODELMANAGER
    %
    %   Utility class for cross-release import of AUTOSAR models
    %
    %   Copyright 2017 The MathWorks, Inc.
    methods (Access = public)
        function obj = AutosarImportModelManager()
            % Empty constructor
        end
        
        function blockHandles = doImport(obj, lXrelModelName, lXrelInfo, lOriginalFolder, lDataDictionary, lXILMode, lXrelModelManager)
            % DOIMPORT
            %   Main logic for cross-release import of AUTOSAR code.
            %
            % Get an instance of the XRelArtifactUtils class (which
            % provides search methods)
            lUtils = coder.internal.xrel.AUTOSAR.XRelArtifactUtils(lXrelInfo);
            
            % Use the arxml.importer to generate a stub model
            lARXMLImporterObj = obj.getARXMLImporterFromArtifact(lXrelInfo);
            lComponentName = obj.getComponentNameForModel(lARXMLImporterObj, lXrelInfo.ComponentName);
            m3iModel = obj.getM3iModelFromArxmlImporter(lARXMLImporterObj);
            m3iComp = autosar.mm.Model.findChildByName(m3iModel, lComponentName);
            lPeriodicTimes = ...
                autosar.mm.mm2sl.PeriodicRunnablesModelingStyleDeterminer.collectPeriodicRunnableSampleTimes(m3iComp.Behavior.Runnables);
            lNoPeriodic = isempty(lPeriodicTimes);
            lPeriodicByFcnCall = lUtils.hasExplicitPeriodicTasking;
            lDefinesSimulinkFunctions = lUtils.definesSimulinkFunctions;
            lForceFunctionCall = lDefinesSimulinkFunctions || lPeriodicByFcnCall || lNoPeriodic;
            
            % Need to map back from the InitializationRunnable symbol to
            % the initialization runnable name
            lInitializationRunnable = '';
            for kRunnable = 1:m3iComp.Behavior.Runnables.size
                m3iRunnable = m3iComp.Behavior.Runnables.at(kRunnable);
                if strcmp(m3iRunnable.symbol, lUtils.getAutosarInitRunnableSymbol)
                    lInitializationRunnable = m3iRunnable.Name;
                    break;
                end
            end
            assert(~isempty(lInitializationRunnable), 'Could not find initialization runnable');
            
            hModel = obj.doModelImport(lARXMLImporterObj, ...
                'DataDictionary', lDataDictionary, ...
                'ForceFunctionCallSubsystem', lForceFunctionCall, ...
                'InitializationRunnable', lInitializationRunnable);
            
            % Configure the configset of the stub model
            %
            % Preserve the AUTOSAR Schema, or feature support in the
            % temporary model may not match the features required by the
            % original code.
            %
            lAutosarSchemaVersion = get_param(hModel, 'AutosarSchemaVersion');
            
            % Preserve the max AUTOSAR identifier length
            lAutosarMaxShortNameLength = get_param(hModel, 'AutosarMaxShortNameLength');
            
            % Preserve multi-tasking (or multi-rate models will fail to
            % build)
            lEnableMultiTasking = get_param(hModel, 'EnableMultiTasking');
            lCodeInterfacePackaging = get_param(hModel, 'CodeInterfacePackaging');
            
            % Apply the cross-release config set (from the user's harness
            % model)
            lXrelModelManager.setupConfigset(get(hModel, 'Name'));
            
            % Re-set the config set parameters from the stub model that
            % must be preserved.
            set_param(hModel, 'EnableMultiTasking', lEnableMultiTasking);
            set_param(hModel, 'AutosarSchemaVersion', lAutosarSchemaVersion);
            set_param(hModel, 'AutosarMaxShortNameLength', lAutosarMaxShortNameLength);
            set_param(hModel, 'CodeInterfacePackaging', lCodeInterfacePackaging);
            set_param(hModel, 'SaveFormat', 'Array');   % For async function calls
            
            % Get the sources from the artifact
            lSourcesHelper = ...
                coder.internal.xrel.AUTOSAR.ModelBlockSourcesHelper.createFromArtifact(...
                lXrelInfo);
            
            % Re-order the ports
            obj.doPortReordering(hModel, lUtils, 'Inport');
            obj.doPortReordering(hModel, lUtils, 'Outport');
            
            % Re-name AFTER re-ordering
            obj.doPortRenaming(hModel, lUtils, 'Inport');
            obj.doPortRenaming(hModel, lUtils, 'Outport');
            
            % Re-name function callers
            obj.doFunctionRenaming(hModel, lXrelInfo.SimulinkFunctions);
            
            % Update task priorities
            obj.updateAsyncTaskPriorities(hModel, lUtils);
                        
            % Configure the stub model with the sources
            hBlock = obj.configureModel(hModel, ...
                lComponentName, ...
                lSourcesHelper, ...
                'ProxyModelName', lXrelModelName);
            
            % Temporarily attach some stuff to UserData
            tempStruct.XrelInfo = lXrelInfo;
            tempStruct.InitRunnable = lUtils.getAutosarInitRunnableSymbol;
            set_param(hBlock, 'UserData', tempStruct);
            set_param(hBlock, 'UserDataPersistent', 'on');
            
            % Make sure data access mode is correct
            obj.doSRInterfaceMatching(hModel, lUtils, 'Outport');
            obj.doSRInterfaceMatching(hModel, lUtils, 'Inport');
            
            % Replace all terminators in the auto-generated ARXML model
            % with a library block containing a terminator and a signal
            % conversion block.
            %
            % This ensures that the inputs are marked as
            % DirectFeedThrough in binfo.
            obj.replaceTerminators(hModel);
            
            % Remove enums defined in base workspace that
            % conflict with session enums.
            obj.cleanUpEnums(hModel);
            
            % C-type PIMs don't get blocks added by createComponentAsModel,
            % so we need to do this ourselves.
            obj.addCTypePIMBlocks(hModel, m3iComp, lUtils);
            
            % Update bus types with dimensions from artifact (unflatten)
            obj.updateStructTypes(lDataDictionary, lUtils);
            
            % Unflatten values in parameter objects (including in
            % bus elements)
            lParameters = lUtils.getParameters;
            for kParam = 1:numel(lParameters)
                obj.updateParamDimensions(lUtils, ...
                    lDataDictionary, ...
                    lParameters(kParam));
            end
            
            % Create the model block
            if coder.internal.xrel.feature('ProtectedAUTOSAR')
                blockHandles = obj.doProtect( ...
                    hModel, lOriginalFolder, lXILMode);
            else
                blockHandles = obj.createModelBlock( ...
                    hModel, lOriginalFolder, lXILMode);
            end 
        end
    end
    
    methods (Access = private, Static)
        function [hModel, bSuccess] = doModelImport(lARXMLImporterObj, varargin)
            % DOMODELIMPORT
            %   Import the required component using the arxml.import object
            %   passed in as lARXMLImporterObj.
            %
            %   Create a cleanup object so that the model is closed when
            %   the AutosarImportModelManager goes out of scope.
            %
            p = inputParser;
            p.addParameter('InitializationRunnable', '', @ischar);
            p.addParameter('ComponentName', '', @ischar);
            p.addParameter('DataDictionary', '', @ischar);
            p.addParameter('ForceFunctionCallSubsystem', false, @islogical);
            p.parse(varargin{:});
            
            lInitializationRunnable = p.Results.InitializationRunnable;
            lComponentName = p.Results.ComponentName;
            lDataDictionary = p.Results.DataDictionary;
            lForceFunctionCallSubsystem = p.Results.ForceFunctionCallSubsystem;
            
            if isempty(lComponentName)
                % If no component name is specified, pick the first one in
                % the ARXML files
                lComponentNameList = lARXMLImporterObj.getComponentNames;
                assert(numel(lComponentNameList) == 1);
                lComponentToImport = lComponentNameList{1};
            else
                lComponentToImport = lComponentName;
            end
            
            % Pass data dictionary through if necessary
            lVarArgs = {};
            if ~isempty(lDataDictionary)
                lVarArgs = {'DataDictionary', lDataDictionary};
            end
            
            % If client server interfaces or we have any function caller
            % inputs, then we can't have periodic runnables
            if lForceFunctionCallSubsystem
                % Call createComponentAsModel with periodic runnables
                % modeled as function call subsystems
                lVarArgs = [lVarArgs, ...
                    {'ModelPeriodicRunnablesAs', 'FunctionCallSubsystem'}];
            else
                % Call createComponentAsModel with periodic runnable
                % specifically modeled as Atomic Subsystems - this is
                % default behavior, but we don't want the information
                % message to be displayed that indicates this since we do
                % not expose this option as a manual user choice.
                lVarArgs = [lVarArgs, ...
                    {'ModelPeriodicRunnablesAs', 'AtomicSubsystem'}];
            end
            
            % Call arxml.importer
            [hModel, bSuccess] = lARXMLImporterObj.createComponentAsModel( ...
                lComponentToImport, ...
                'NameConflictAction', 'makenameunique', ...
                'OpenModel', false, ...
                'InitializationRunnable', lInitializationRunnable, ...
                lVarArgs{:});
        end
        
        function hBlock = configureModel(hModel, lComponentName, lSourcesHelper, varargin)
            % CONFIGUREMODEL
            %   Add the control block to the model, populated with the
            %   sources and include files contained within lSourcesHelper.
            %
            %   Set the tag so that the build hooks can identify that the
            %   model must be treated specially when building.
            %
            %   Then, save the model.
            %   Optionally re-name to ProxyModelName when saving.
            p = inputParser;
            p.addParameter('ProxyModelName', '', @ischar);
            p.parse(varargin{:});
            lProxyModelName = p.Results.ProxyModelName;
            lModelName = get_param(hModel, 'Name');
            hBlock = lSourcesHelper.addBlockToModel(lModelName);
            set_param(hBlock, 'ComponentName', lComponentName);
            set_param(lModelName, 'Tag', coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook.XRelModelTag);
            
            % Use our custom TLC template to filter the model file from
            % being written!
            set_param(lModelName, 'ERTCustomFileTemplate', 'xrel_autosar_modelblock_customization.tlc');
            
            % Use compact file packaging - we don't want the generated
            % files anyway.
            if ~strcmp(get_param(lModelName, 'CodeInterfacePackaging'), 'Reusable function')
                set_param(lModelName, 'ERTFilePackagingFormat', 'Compact');
                
                % Set single instance model ref mode too even though we are
                % not a model ref (issues with accel build)
                set_param(lModelName, 'ModelReferenceNumInstancesAllowed', 'Single');
            end
            
            lDescription = sprintf('Cross-Release Model Block: %s', ...
                lModelName);
            set_param(lModelName, 'Description', lDescription);
            set_param(hModel, 'GenCodeOnly', 'on');
            set_param(hModel, 'UseOnlyExistingSharedCode', 'none');
            
            % Set all logging off or we have trouble when the model has
            % variants.
            set_param(hModel, 'SaveTime', 'off');
            set_param(hModel, 'SaveState', 'off');
            set_param(hModel, 'SaveOutput', 'off');
            set_param(hModel, 'SaveFinalState', 'off');
            set_param(hModel, 'SignalLogging', 'off');
            set_param(hModel, 'DSMLogging', 'off');
            set_param(hModel, 'LoggingToFile', 'off');
            set_param(hModel, 'ReturnWorkspaceOutputs', 'off');
            
            % Preserve port interface
            set_param(hModel, 'AutosarMatrixIOAsArray', 'on');

            % Make sure SimGenImportedTypeDefs is ON or Stateflow will
            % cause problems building SIM target for the stub model
            set_param(lModelName, 'SimGenImportedTypeDefs', 'on');
            
            % Don't want warnings - we know we have some unused data and
            % events with the arxml imported model.
            set_param(lModelName, 'SFUnusedDataAndEventsDiag', 'none');

            % Don't want shared constants to be generated by the stub
            % model, or we could get a phantom shared library dependency
            % which can cause issues with MSVC/Nmake on Windows.
            set_param(lModelName, 'GenerateSharedConstants', 'off');
            
            % Need to save the model or we cannot protect it
            if isempty(lProxyModelName)
                save_system(lModelName);
            else
                save_system(lModelName, lProxyModelName);
            end
        end
        
        function doPortReordering(hModel, lXrelArtifactUtils, lPortType)
            % DOPORTREORDERING
            %
            % Re-order the ports in the skeleton model imported by
            % arxml.importer based on the information from the
            % cross-release artifact (lArtifactPortInfo).
            %
            % Inputs:
            %       hModel:     Handle to the skeleton model
            %       lArtifactPortInfo:  XrelInfo.Inports or XrelInfo.Outports
            %       lPortType:  Inport or Outport
            import coder.internal.xrel.AUTOSAR.AutosarImportModelManager;
            
            % Get list of ports
            isInport = strcmp(lPortType, 'Inport');
            [lModelDataPorts, lModelDataPortNames] = AutosarImportModelManager.getModelDataPortHandlesAndNames(hModel, isInport);
            [lModelFcnPorts, lModelFcnPortNames] = AutosarImportModelManager.getModelFcnPortHandlesAndNames(hModel, isInport);
            
            % Use the AUTOSAR Simulink mapping to match the ports to avoid
            % making assumptions about the naming scheme used in the
            % skeleton model.
            slMap = autosar.api.getSimulinkMapping(hModel);
            arProps = autosar.api.getAUTOSARProperties(hModel);
            lDataPortMap = AutosarImportModelManager.getDataPortSlMapping(slMap, lModelDataPorts, lModelDataPortNames, isInport);
            lFcnPortMap = AutosarImportModelManager.getFcnPortSlMapping(slMap, arProps, lModelFcnPorts, lModelFcnPortNames);
            
            % Can compute these up front
            lIsUpdated = strcmp('IsUpdated', {lDataPortMap.DataAccessMode});
            lIsErrorStatus = strcmp('ErrorStatus', {lDataPortMap.DataAccessMode});
            
            % Log which ports we've visited
            lModelDataPortWasInArtifact = false(1, numel(lModelDataPorts));
            lModelFcnPortWasInArtifact = false(1, numel(lModelFcnPorts));
            
            % Renumber each port in turn
            for kPort = 1:lXrelArtifactUtils.getNumPorts(lPortType)
                lcurrPortNum = num2str(kPort);
                if lXrelArtifactUtils.isFunctionCall(lPortType, kPort)
                    lExpectSymbol = lXrelArtifactUtils.getPortSymbol(lPortType, kPort);
                    lIdx = find(strcmp(lExpectSymbol, lFcnPortMap), 1, 'first');
                    assert(~isempty(lIdx), 'Could not find function port for runnable with symbol %s', lExpectSymbol);
                    set_param(lModelFcnPorts(lIdx), 'Port', lcurrPortNum);
                    lModelFcnPortWasInArtifact(lIdx) = true;
                else
                    % For the graphical port of lPortType and index kPort,
                    % get the AUTOSAR port name, data element name and data
                    % access mode from the cross-release artifact.
                    [lAutosarPort, lAutosarDataElement, lDataAccessMode] = ...
                        lXrelArtifactUtils.getAutosarPortMapping(lPortType, kPort);
                    
                    lHasMatch = strcmp(lAutosarPort, {lDataPortMap.PortName}) ...
                        & strcmp(lAutosarDataElement, {lDataPortMap.DataElementName});
                    if strcmp(lDataAccessMode, 'IsUpdated')    
                        lHasMatch = lHasMatch & lIsUpdated;
                    elseif strcmp(lDataAccessMode, 'ErrorStatus')
                        lHasMatch = lHasMatch & lIsErrorStatus;
                    else
                        lHasMatch = lHasMatch & ~lIsUpdated & ~lIsErrorStatus;
                    end
                    
                    lIdx = find(lHasMatch);
                    assert(numel(lIdx) <= 1, 'Found multiple matches for %s: %s, DataElement: %s, AccessMode: %s', lPortType, lAutosarPort, lAutosarDataElement, lDataAccessMode);
                    assert(numel(lIdx) == 1, 'Found no matches for %s: %s, DataElement: %s, AccessMode: %s', lPortType, lAutosarPort, lAutosarDataElement, lDataAccessMode);
                    set_param(lModelDataPorts(lIdx), 'Port', lcurrPortNum);
                    lModelDataPortWasInArtifact(lIdx) = true;
                end
            end
            
            % Assert that we've not missed any function ports
            assert(all(lModelFcnPortWasInArtifact), 'Not all function call ports in stub model were mapped to function call ports from cross release artifact');
            
            % We might get unwanted ErrorStatus ports when triggered by
            % error events
            for kDataPort = 1:numel(lModelDataPorts)
                if isInport
                    if ~lModelDataPortWasInArtifact(kDataPort)
                        if lIsErrorStatus(kDataPort)
                            % Remove it
                            lTag = 'PORT_TO_REMOVE';
                            set_param(lModelDataPorts(kDataPort), 'Tag', lTag);
                            
                            % Should be able to just swap with a
                            % ground block?
                            replace_block(hModel, 'Tag', lTag, 'built-in/Ground', 'noprompt');
                            
                            % We know about it and have dealt with it.
                            % Mark as true.
                            lModelDataPortWasInArtifact(kDataPort) = true;
                        end
                    end
                end
            end
            assert(all(lModelDataPortWasInArtifact), 'Not all data ports in stub model were mapped to function call ports from cross release artifact');
        end
        
        function [lModelDataPorts, lModelDataPortNames]  = getModelDataPortHandlesAndNames(hModel, isInport)
            % GETMODELDATAPORTHANDLESANDNAMES
            if isInport
                % Ports with a signal output
                lModelDataPorts = find_system(hModel, 'SearchDepth', 1, 'BlockType', 'Inport', 'OutputFunctionCall', 'off');
                lModelDataPortNames = get_param(lModelDataPorts, 'Name');
            else
                % Assume all ports take a signal
                lModelDataPorts = find_system(hModel, 'SearchDepth', 1, 'BlockType', 'Outport');
                lModelDataPortNames = get_param(lModelDataPorts, 'Name');
            end
            
            % Single element returns from get_param won't be cells, so wrap them.
            if ~iscell(lModelDataPortNames)
                lModelDataPortNames = {lModelDataPortNames};
            end
        end
        
        function [lModelFcnPorts, lModelFcnPortNames]  = getModelFcnPortHandlesAndNames(hModel, isInport)
            % GETMODELFCNPORTHANDLESANDNAMES
            if isInport
                % Ports with a function call output
                lModelFcnPorts = find_system(hModel, 'SearchDepth', 1, 'BlockType', 'Inport', 'OutputFunctionCall', 'on');
                lModelFcnPortNames = get_param(lModelFcnPorts, 'Name');
            else
                lModelFcnPorts = [];
                lModelFcnPortNames = {};
            end
            if ~iscell(lModelFcnPortNames)
                lModelFcnPortNames = {lModelFcnPortNames};
            end
        end
        
        function lDataPortMap = getDataPortSlMapping(slMap, lModelDataPorts, lModelDataPortNames, isInport)
            % GETDATAPORTSLMAPPING
            %
            % Pre-allocate struct array populated with empty elements
            nModelDataPorts = numel(lModelDataPorts);
            lDataPortMap = struct('PortName', cell(1, nModelDataPorts), ...
                'DataElementName', cell(1, nModelDataPorts), ...
                'DataAccessMode', cell(1, nModelDataPorts));
            if isInport
                fhData = @slMap.getInport;
            else
                fhData = @slMap.getOutport;
            end
            for kDataPort = 1:nModelDataPorts
                [lDataPortMap(kDataPort).PortName, ...
                    lDataPortMap(kDataPort).DataElementName, ...
                    lDataPortMap(kDataPort).DataAccessMode] = fhData(lModelDataPortNames{kDataPort});
            end
        end
        
        function lFcnPortMap = getFcnPortSlMapping(slMap, arProps, lModelFcnPorts, lModelFcnPortNames)
            % GETFCNPORTSLMAPPING
            swc = get(arProps,'XmlOptions','ComponentQualifiedName');
            ib = get(arProps,swc,'Behavior');
            nModelFcnPorts = numel(lModelFcnPorts);
            lFcnPortMap = cell(1, nModelFcnPorts);
            for kFcnPort = 1:nModelFcnPorts
                % This gets the "Name", but not the "Symbol"
                lRunnableName = slMap.getFunction(lModelFcnPortNames{kFcnPort});
                
                % Convert the "Name" to a "Symbol", because that's what we
                % are able to guarantee from CodeInfo in the old release.
                lRunnablePath = arProps.find(ib, 'Runnable', 'Name', lRunnableName);
                assert(numel(lRunnablePath) == 1);  
                lFcnPortMap{kFcnPort} = arProps.get(lRunnablePath{1}, 'symbol');
            end
        end
        
        function doPortRenaming(hModel, lXrelArtifactUtils, lBlockType)
            % DOPORTRENAMING
            %
            %   Re-name the root input/output ports in the skeleton model
            %   with the GraphicalNames from the cross release artifact.
            
            % Desired input/output port names will not conflict with each
            % other.
            %
            % But they might conflict with the automatic ones (or another
            % block in the diagram)
            %
            import coder.internal.xrel.AUTOSAR.AutosarImportModelManager;
            
            for kPort = 1:lXrelArtifactUtils.getNumPorts(lBlockType)
                lcurrPortNum = num2str(kPort);
                lGraphicalName = lXrelArtifactUtils.getGraphicalName(lBlockType, kPort);
                hPortBlock = find_system(hModel, 'SearchDepth', 1, 'BlockType', lBlockType, 'Port', lcurrPortNum);
                lExistingBlock = find_system(hModel, 'SearchDepth', 1, 'Name', lGraphicalName);
                lPortHasConflict = ~isempty(lExistingBlock);
                if lPortHasConflict
                    kRenamed = 0;
                    lNewName = [lGraphicalName, '_RENAMED'];
                    while ~isempty(find_system(hModel, 'SearchDepth', 1, 'Name', lNewName))
                        kRenamed = kRenamed+1;
                        lNewName = [lGraphicalName, '_RENAMED', num2str(kRenamed)];
                    end
                    set_param(lExistingBlock, 'Name', lNewName);
                end
                set_param(hPortBlock, 'Name', lGraphicalName);
                
                % Verify dimensions
                lCurrentDims = eval(get_param(hPortBlock, 'PortDimensions'));
                if ~lXrelArtifactUtils.isFunctionCall(lBlockType, kPort)
                    [lExpectDims, lIsBus] = lXrelArtifactUtils.getDimensions(lBlockType, kPort);
                    assert(prod(lCurrentDims) == prod(lExpectDims), 'Dims dont match');
                    set_param(hPortBlock, 'PortDimensions', mat2str(lExpectDims));
                    % Trace back to source
                    if strcmp(lBlockType, 'Outport') && lIsBus
                        AutosarImportModelManager.updateConstant(hPortBlock, lExpectDims);
                    end
                end
            end
        end
        
        
        function updateConstant(hPortBlock, lExpectDims)
            % UPDATECONSTANT
            %
            %   Update the constant block driving an output port where the
            %   original model expected an n-D rather than 1D signal.
            %
            if numel(lExpectDims) > 1    
                lLineHandles = get_param(hPortBlock, 'LineHandles');
               
                lSourceBlock = get(lLineHandles.Inport, 'SrcBlockHandle');
                lSourceBlockType = get(lSourceBlock, 'BlockType');
                
                if strcmp(lSourceBlockType, 'SubSystem')
                    lSourcePort = get(lLineHandles.Inport, 'SrcPortHandle');
                    lSourcePortNum = get(lSourcePort, 'PortNumber');
                    lSourceParent = get_param(get_param(lSourcePort, 'Parent'), 'Handle');
                    
                    lPortInParent = find_system(lSourceParent, ...
                        'SearchDepth', 1, ...
                        'BlockType', 'Outport', ...
                        'Port', num2str(lSourcePortNum));
                    
                    lLinesInParent = get_param(lPortInParent, 'LineHandles');
                    lSecondSource = get_param(lLinesInParent.Inport, 'SrcBlockHandle');
                    lSecondSourceType = get_param(lSecondSource, 'BlockType');
                    
                    if strcmp(lSecondSourceType, 'Constant')
                        lStrZeros = sprintf('zeros(%s)', mat2str(lExpectDims));
                        set_param(lSecondSource, 'Value', lStrZeros);
                    end 
                end
            end
        end
        
        function doSRInterfaceMatching(hModel, lXRelArtifactUtils, lBlockType)
            % DOSRINTERFACEMATCHING
            %   Make sure DataAccessMode is correct for root input and
            %   output ports.
            slMap=autosar.api.getSimulinkMapping(hModel);
            for kPort = 1:lXRelArtifactUtils.getNumPorts(lBlockType)
                if ~lXRelArtifactUtils.isFunctionCall(lBlockType, kPort)
                    lPort = lXRelArtifactUtils.getDataPortElemWithIndirection(lBlockType, kPort);
                    switch lBlockType
                        case 'Outport'
                            [lArPortName,lArDataElementName] = slMap.getOutport(lPort.GraphicalName);
                            mapCmd = @slMap.mapOutport;
                        case 'Inport'
                            [lArPortName,lArDataElementName] = slMap.getInport(lPort.GraphicalName);
                            mapCmd = @slMap.mapInport;
                        otherwise
                            assert(false, 'Should not get here');
                    end
                    
                    assert(isequal(lArPortName, lPort.AutosarPort), 'Port mismatch');
                    assert(isequal(lArDataElementName, lPort.AutosarDataElement), 'Data element mismatch');
                    mapCmd(lPort.GraphicalName, lArPortName, lArDataElementName, lPort.DataAccessMode);
                end
            end
        end
        
        
        function doFunctionRenaming(hModel, lArtifactFunctions)
            % DOFUNCTIONRENAMING
            %   Update function caller blocks to reflect the original
            %   Simulink model.
            %
            lBlockType = 'FunctionCaller';
            lFcnBlocks = find_system(hModel, 'BlockType', lBlockType);
            lFcnPrototypes = get_param(lFcnBlocks, 'FunctionPrototype');
            if ~iscell(lFcnPrototypes)
                lFcnPrototypes = {lFcnPrototypes};
            end
            
            lProtoRegexp = '(?<=(\w*\s*=\s*)?)(\w*)(?=(\(|$))';
            lFcnNames = regexp(lFcnPrototypes, lProtoRegexp, 'match', 'once');
            
            if isempty(lArtifactFunctions)
                lExpectList = {};
            else
                lExpectList = strcat( ...
                    { lArtifactFunctions.AUTOSARPortName }, ...
                    '_', ...
                    { lArtifactFunctions.AUTOSAROperationName });
            end
            
            % Iterate through the function caller blocks that we find - not
            % all functions will be called.
            for kFcnCaller = 1:numel(lFcnBlocks)   
                lCurrentFunction = lFcnNames{kFcnCaller};
                lIsMatched = strcmp(lCurrentFunction, lExpectList);
                if any(lIsMatched)
                    lIdx = find(lIsMatched, 1, 'first');    
                    oldFunctionPrototype = lFcnPrototypes{kFcnCaller};
                    newFunctionPrototype = regexprep( oldFunctionPrototype, ...
                        lProtoRegexp, ...
                        regexptranslate('escape', lArtifactFunctions(lIdx).CGFunctionName));
                    set_param( lFcnBlocks(kFcnCaller), 'FunctionPrototype', newFunctionPrototype);
                end
            end
        end
        
        function updateAsyncTaskPriorities(hModel, lXrelArtifactUtils)
            % UPDATEASYNCTASKPRIORITIES
            %
            %   Update async task priorities to the values in the original
            %   model (or potentially the cross-release model will not be a
            %   drop-in replacement)
            %
            for kPort = 1:lXrelArtifactUtils.getNumPorts('Inport')
                if lXrelArtifactUtils.isFunctionCall('Inport', kPort)
                    lOutputFunction = lXrelArtifactUtils.getOutputFunctionFromInport('Inport', kPort);
                    if strcmp(lOutputFunction.TimingMode, 'ASYNCHRONOUS')
                        lGraphicalName = lXrelArtifactUtils.getGraphicalName('Inport', kPort);
                        lPriority = num2str(lOutputFunction.Priority);
                        hPortBlock = find_system(hModel, 'SearchDepth', 1, 'Name', lGraphicalName);
                        hLine = get(hPortBlock, 'LineHandles');
                        hAsyncBlock = get(hLine.Outport, 'DstBlockHandle');
                        
                        % There might not be an async task block always
                        if ~isempty(hAsyncBlock) && ...
                                strcmp(get(hAsyncBlock, 'BlockType'), 'AsynchronousTaskSpecification')
                            set_param(hAsyncBlock, 'TaskPriority', lPriority);
                        end
                    end
                end
            end
        end
        
        function addCTypePIMBlocks(hModel, m3iComp, lUtils)
            % ADDCTYPEPIMBLOCKS
            
            % Iterate through C-typed PIMS
            lDWorks = cellfun(@lUtils.getStatesDWorkName, ...
                {'Block states', 'Block signals', 'Zero crossing states'}, ...
                'UniformOutput', false);
            
            % Delete entries that are empty
            lDWorks(cellfun(@isempty, lDWorks)) = [];
            
            [~, lControlBlockHandle] = coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook.isAutosarCodeImport(hModel);
            lControlBlockPosition = double(get_param(lControlBlockHandle, 'Position'));
            lModelName = get_param(hModel, 'Name');
            for kPim = 1:double(m3iComp.Behavior.PIM.size)
                m3iPim = m3iComp.Behavior.PIM.at(kPim);
                if ~isempty(lDWorks) && ismember(m3iPim.Name, lDWorks)
                    % Skip if this is one of the model DWorks
                    continue;
                end
                    
                lNewPos0 = [lControlBlockPosition(3)+60*kPim, lControlBlockPosition(2)+30];
                lNewPos = [lNewPos0, lNewPos0+[30 30]];
                
                % Get data type and dimensions from cross-release artifact
                lDataStoreInfo = lUtils.getDataStoreByGraphicalName(m3iPim.Name);
                if ~isempty(lDataStoreInfo)
                    if lDataStoreInfo.IsBus
                        lDataType = ['Bus: ', lDataStoreInfo.TypeName];
                    elseif lDataStoreInfo.IsEnum
                        lDataType = ['Enum: ', lDataStoreInfo.TypeName];
                    else
                        lDataType = lDataStoreInfo.TypeName;
                    end
                    add_block('built-in/DataStoreMemory', ...
                        [lModelName, '/', sprintf('CPIM_%d', kPim)], ...
                        'Position', lNewPos, ...
                        'DataStoreName', m3iPim.Name, ...
                        'OutDataTypeStr', lDataType, ...
                        'Dimensions', mat2str(lDataStoreInfo.Dimensions), ...
                        'StateSignalObject', AUTOSAR.Signal, ...
                        'StateStorageClass', 'PerInstanceMemory (Custom)');
                end
            end
        end
        
        function updateStructTypes(lDataDictionary, lUtils)
            % UPDATESTRUCTTYPES
            lXRelTypeUtils = coder.internal.xrel.XrelTypeUtils(lDataDictionary);
            
            lStructs = lUtils.getStructTypes;
            for kStruct = 1:numel(lStructs)
                [slObj, wasFound] = lXRelTypeUtils.getSimulinkObject(lStructs(kStruct).Name);
                if wasFound
                    % Check the dimensions of each element
                    assert(isa(slObj, 'Simulink.Bus'), 'Expected %s to be a Simulink.Bus object', lStructs(kStruct).Name);
                    lDidChange = false;
                    for kElem = 1:numel(slObj.Elements)
                        lFoundDims = slObj.Elements(kElem).Dimensions;
                        lExpectDims = lStructs(kStruct).Elements(kElem).Dimensions;
                        assert(prod(lFoundDims) == prod(lExpectDims));
                        if ~isequal(lFoundDims, lExpectDims)
                            slObj.Elements(kElem).Dimensions = lExpectDims;
                            lDidChange = true;
                        end
                    end
                    if lDidChange
                        lXRelTypeUtils.setSimulinkObject(lStructs(kStruct).Name, slObj);
                    end                    
                end
            end
        end
        

        function ret = hasNonScalarBusElement(lUtils, lName)
            % HASNONSCALARBUSELEMENT
            lArtifact = lUtils.getStructTypeByName(lName);
            if isempty(lArtifact)
                ret = false;
                return;
            end
            
            for kElem = 1:numel(lArtifact.Elements)
                if any(lArtifact.Elements(kElem).Dimensions ~= 1)
                    ret = true;
                    return;
                else
                    % Recurse
                    lElemType = lArtifact.Elements(kElem).DataType;
                    ret = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.hasNonScalarBusElement(lUtils, lElemType);
                end
            end
        end
        
        function updateParamDimensions(lUtils, lDataDictionary, lArtifact)
            % UPDATEPARAMDIMENSIONS
            %
            % Determine the name of the parameter based on the Data Access
            % Mode.
            if strcmp(lArtifact.DataAccessMode, 'Calibration')
                lName = [lArtifact.AutosarPort, '_', lArtifact.AutosarDataElement];
            else
                lName = lArtifact.AutosarPort;
            end
            if numel(lName) > 63
                lName = lName(1:63);
            end
            
            lXRelTypeUtils = coder.internal.xrel.XrelTypeUtils(lDataDictionary);
            
            % Unflatten any data that was marked as a matrix in the
            % Cross-Release Artifact.
            lIsMatrix = numel(lArtifact.Dimensions) > 1 && any(lArtifact.Dimensions ~=1);
            if lIsMatrix
                lObj = lXRelTypeUtils.getSimulinkObject(lName);
                if ~isempty(lObj)
                    assert(prod(lArtifact.Dimensions) == prod(lObj.Dimensions), 'Expected number of elements to be the same');
                    lObj.Value = reshape(lObj.Value, lArtifact.Dimensions);
                    lXRelTypeUtils.setSimulinkObject(lName, lObj);
                end
            end
            
            if lArtifact.IsBus
                % Walk through bus elements and unflatten any data that was
                % marked as matrix valued in the Cross-Release Artifact.
                lHasNonScalar = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.hasNonScalarBusElement(lUtils, lArtifact.BusTypeName);
                if lHasNonScalar
                    % Then we need to do something
                    lObj = lXRelTypeUtils.getSimulinkObject(lName);
                    if ~isempty(lObj)
                        lVal = lObj.Value;
                        lVal = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeVal( ...
                            lUtils, ...
                            lVal, ...
                            lArtifact.BusTypeName);
                        lObj.Value = lVal;
                        lXRelTypeUtils.setSimulinkObject(lName, lObj);
                    end
                end
            end
        end
        
        function lVal = reshapeVal(lUtils, lVal, lBusTypeName)
            % RESHAPEVAL
            lTypeArtifact = lUtils.getStructTypeByName(lBusTypeName);
            if isempty(lTypeArtifact) || ~isstruct(lVal)
                return;
            end
            for kElem = 1:numel(lTypeArtifact.Elements)
                if numel(lTypeArtifact.Elements(kElem).Dimensions) > 1 && ...
                        any(lTypeArtifact.Elements(kElem).Dimensions ~= 1)
                    lVal.(lTypeArtifact.Elements(kElem).Identifier) = ...
                        reshape( ...
                        lVal.(lTypeArtifact.Elements(kElem).Identifier), ...
                        lTypeArtifact.Elements(kElem).Dimensions);
                end
                lElemTypeName = lTypeArtifact.Elements(kElem).DataType;
                lVal.(lTypeArtifact.Elements(kElem).Identifier) = ...
                    coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeVal( ...
                    lUtils, ...
                    lVal.(lTypeArtifact.Elements(kElem).Identifier), ...
                    lElemTypeName);
            end
        end
        
                
        function replaceTerminators(hModel)
            % REPLACETERMINATORS
            %   Replace terminators with a terminator connected to a signal
            %   conversion block - avoids being optimized away.
            lBlockType = 'Terminator';
            lToReplaceTag = '###TerminatorsToReplace';
            lReplacementBlock = 'pil_xrel_autosar_lib/SignalConvertingTerminator';
            load_system('pil_xrel_autosar_lib');
            coder.internal.xrel.AUTOSAR.AutosarImportModelManager.replaceBlocks(hModel, lBlockType, lToReplaceTag, lReplacementBlock);
        end

        
        function cleanUpEnums(hModel)
            % CLEANUPENUMS
            lBlocksWithEnum = find_system(hModel, 'Regexp', 'on', 'OutDataTypeStr', 'Enum:*');
            lBaseVars = evalin('base', 'who');
            for kPort = 1:numel(lBlocksWithEnum)
                lCurrPort = lBlocksWithEnum(kPort);
                lDataType = get_param(lCurrPort, 'OutDataTypeStr');
                if startsWith(lDataType, 'Enum:')
                    lEnumName = regexprep(lDataType, '^Enum:\s*', '');
                    if ismember(lEnumName, lBaseVars)
                        evalin('base', ['clearvars ', lEnumName]);
                    end        
                end
            end
        end
        
        function hXrelBlock = createModelBlock(hModel, lDstFolder, lXILMode)
            % CREATEMODELBLOCK
            %
            %   Package the proxy model as a model block (do not protect)
            %
            lExistingSharedCode = get_param(hModel, 'ExistingSharedCode');
            if ~isempty(lExistingSharedCode)
                set_param(hModel, 'UtilityFuncGeneration', 'Shared Location');
            end
            save_system(hModel);
            lModelName = get_param(hModel, 'Name');
            lModelSavedName = get_param(hModel, 'FileName');
            
            % Create a model block in a blank system.
            hUntitledModel = new_system;
            hUntitledModelName = get_param(hUntitledModel, 'Name');
            
            hXrelBlock = add_block('built-in/ModelReference', ...
                [hUntitledModelName, '/', lModelName]);
            set_param(hXrelBlock, 'ModelName', lModelName);
            
            switch lXILMode
                case {'SIL'}
                    set_param(hXrelBlock, 'SimulationMode', 'Software-in-the-loop (SIL)');
                case 'PIL'
                    set_param(hXrelBlock, 'SimulationMode', 'Processor-in-the-loop (PIL)');
                otherwise
                    assert(false, 'lXILMode should be SIL or PIL');
            end
            set_param(hXrelBlock, 'CodeInterface', 'Top Model');
            
            % Close the model
            bdclose(lModelName);
            
            % Move the file
            movefile(lModelSavedName, lDstFolder);
        end
        
        function hXrelBlock = doProtect(hModel, lDstFolder, lXILMode)
            % DOPROTECT
            %   Package the proxy model as a protected model
            %
            lModelName = get_param(hModel, 'Name');
            lExistingSharedCode = get_param(hModel, 'ExistingSharedCode');
            if ~isempty(lExistingSharedCode)
                set_param(hModel, 'UtilityFuncGeneration', 'Shared Location');
            end
            save_system(lModelName);
            fhProtect = @()Simulink.ModelReference.protect(lModelName, ...
                'Mode', 'CodeGeneration', ...
                'CodeInterface', 'Top model', ...
                'OutputFormat', 'MinimalCode', ...
                'Harness', true, ...
                'Webview', true, ...
                'ObfuscateCode', false, ...
                'Path', lDstFolder);
            
            % Suppress model build information, consistent with ERT
            % workflow
            hProtHarness = ...
                coder.internal.xrel.AUTOSAR.AutosarImportModelManager.runCmdWithNoOutput(fhProtect);
            
            hXrelBlock = find_system(hProtHarness, 'BlockType', 'ModelReference');
            switch lXILMode
                case {'SIL'}
                    set_param(hXrelBlock, 'SimulationMode', 'Software-in-the-loop (SIL)');
                case 'PIL'
                    set_param(hXrelBlock, 'SimulationMode', 'Processor-in-the-loop (PIL)');
                otherwise
                    assert(false, 'lXILMode should be SIL or PIL');
            end
            bdclose(lModelName);
        end
        
        function lComponentName = getComponentNameForModel(lARXMLImporterObj, lXRelInfoComponentName)
            % GETCOMPONENTNAMEFORMODEL
            %   Get the fully qualified component name for the model
            %   described by the cross release artifact
            lComponents = lARXMLImporterObj.getComponentNames;
            lComponentName = lComponents{ ...
                strcmp(lComponents, lXRelInfoComponentName) ...             % Earlier releases serialize the fully qualified component name in CodeInfo.Name
                | endsWith(lComponents, ['/', lXRelInfoComponentName]) };   % More recently just the final part is used
            assert(~isempty(lComponentName), 'Could not find component %s', lXRelInfoComponentName);
        end
        
        function m3iModel = getM3iModelFromArxmlImporter(lARXMLImporterObj)
            % GETM3IMODELFROMARXMLIMPORTER
            %
            % Ensure that the m3iModel is loaded then get it.
            lARXMLImporterObj.getComponentNames;
            m3iModel = lARXMLImporterObj.getM3IModel;
        end
        
        function lARXMLImporterObj = getARXMLImporterFromArtifact(xrel_info)
            % GETARXMLIMPORTERFROMARTIFACT
            %   Process the cross-release artifact and find the arxml files
            %   in the build folder.  Create an ARXML importer object
            %   associated with these arxml files.
            %
            startDir = xrel_info.BuildDir;
            lARXMLImporterObj = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.getARXMLImporterFromBuildDir(startDir);
            
        end
    end
       
    methods (Static, Access = public)
        function lARXMLImporterObj = getARXMLImporterFromBuildDir(startDir)
            % GETARXMLIMPORTERFROMBUILDDIR
            %
            % Searches over the arxml files in startDir and creates an
            % arxml.importer object from them.
            
            % Some interface arxml files live in the stub folder
            lDirARXMLFiles = [dir(fullfile(startDir, '*.arxml'));  ...
                dir(fullfile(startDir, 'stub', '*.arxml'))];
            % Don't want any folders ending in .arxml!
            assert(~isempty(lDirARXMLFiles));
            lDirARXMLFiles([lDirARXMLFiles.isdir]) = [];
            assert(~isempty(lDirARXMLFiles));
            
            % Convert to a cell array
            fhDirStructToFullPath = @(x)fullfile(x.folder, x.name);
            lARXMLFiles = arrayfun(fhDirStructToFullPath, lDirARXMLFiles, 'UniformOutput', false);
            lARXMLImporterObj = arxml.importer(lARXMLFiles);
        end
    end
    
    
    methods (Access = private, Static)
        % Internal helper methods
        function replaceBlocks(hModel, lBlockType, lToReplaceTag, lReplacementBlock)
            % REPLACEBLOCKS
            lModelHandle = get_param(hModel, 'handle'); % Ensure is handle
            lModelName = get_param(hModel, 'Name');     % Ensure is name
            lToReplace = find_system(lModelHandle, 'LookUnderMasks', 'all', 'BlockType', lBlockType);
            for hBlock = lToReplace(:)'
                set_param(hBlock, 'Tag', lToReplaceTag);
            end
            replace_block(lModelName, 'Tag', lToReplaceTag, lReplacementBlock, 'noprompt');
        end
        
        function retVal = runCmdWithNoOutput(cmd) %#ok<INUSD>
            % runCmdWithNoOutput
            ME = [];
            retVal = [];
            output = evalc('try retVal = feval(cmd), catch ME, end');
            if ~isempty(ME)
                disp(output);
                throwAsCaller(ME);
            end
        end
    end
end
