classdef ModelBlockSourcesHelper < handle
    % MODELBLOCKSOURCESHELPER
    %
    %  Use to associate cross-release or imported external AUTOSAR files
    %  with a model
    %
    %   Copyright 2017 The MathWorks, Inc.
    %
    properties (SetAccess = immutable)
        StartDir;
        PathRelativeToAnchor;
        SourcePaths;
        IncludePaths;
        SourceFiles;
        IncludeFiles;
        SourceStubs;
        IncludeStubs;
        IncludeProxies;
    end
    
    methods (Access = public)
        function obj = ModelBlockSourcesHelper(lStartDir, ...
                lPathRelativeToAnchor, ...
                lSourcePaths, lIncludePaths, ...
                lSourceFiles, lIncludeFiles, ...
                lSourceStubs, lIncludeStubs, ...
                lIncludeProxies)
            % MODELBLOCKSOURCESHELPER
            %
            % Populate object properties
            %
            obj.StartDir = lStartDir;
            obj.PathRelativeToAnchor = lPathRelativeToAnchor;
            obj.SourcePaths = lSourcePaths;
            obj.IncludePaths = lIncludePaths;
            obj.SourceFiles = lSourceFiles;
            obj.IncludeFiles = lIncludeFiles;
            obj.SourceStubs = lSourceStubs;
            obj.IncludeStubs = lIncludeStubs;
            obj.IncludeProxies = lIncludeProxies;
        end
        
        function hBlock = addBlockToModel(obj, lModelName)
            % ADDBLOCKTOMODEL
            %
            % Add the control block to lModelName
            % Get the location of the upper-left-most block
            lSubsystemBlocks = find_system(lModelName, 'BlockType', 'SubSystem');
            lOrigin = [NaN NaN];
            for kBlock = 1:numel(lSubsystemBlocks)
                lPos = get_param(lSubsystemBlocks{kBlock}, 'Position');
                lOrigin = min([lOrigin;lPos(1:2)]);
            end
            lOrigin = lOrigin - [0 150];
            
            % Library must be open
            load_system('pil_xrel_autosar_lib');
            
            % Add the block
            hBlock = add_block('pil_xrel_autosar_lib/IMPORT_AUTOSAR_CODE', ...
                [lModelName, '/IMPORT_AUTOSAR_CODE'], ...
                'Position', [lOrigin, lOrigin+[400 100]]);
            
            obj.setSourcesAndIncludes(hBlock);
        end
        
        function setSourcesAndIncludes(obj, hBlock)
            % SETSOURCESANDINCLUDES
            %
            % Configure block with handle hBlock with the information about
            % the imported model.
            %
            % This could be custom code, or code associated with a
            % cross-release artifact.
            %
            % Not static, because we want to use obj. notation to access
            % static member functions.
            set_param(hBlock, 'StartDir', ['''', obj.StartDir, '''']);
            set_param(hBlock, 'PathRelativeToAnchor', ['''', obj.PathRelativeToAnchor, '''']);
            set_param(hBlock, 'SourcePaths', obj.cell2Str(obj.SourcePaths));
            set_param(hBlock, 'IncludePaths', obj.cell2Str(obj.IncludePaths));
            set_param(hBlock, 'SourceFiles', obj.cell2Str(obj.SourceFiles));
            set_param(hBlock, 'IncludeFiles', obj.cell2Str(obj.IncludeFiles));
            set_param(hBlock, 'SourceStubs', obj.cell2Str(obj.SourceStubs));
            set_param(hBlock, 'IncludeStubs', obj.cell2Str(obj.IncludeStubs));
            set_param(hBlock, 'IncludeProxies', obj.cell2Str(obj.IncludeProxies));
        end
        
        function [lSourcePaths, lIncludePaths, ...
                lSourceFiles, lIncludeFiles, ...
                lSourceStubs, lIncludeStubs, ...
                lIncludeProxies] = getData(obj)
            % GETDATA
            lSourcePaths = obj.SourcePaths;
            lIncludePaths = obj.IncludePaths;
            lSourceFiles = obj.SourceFiles;
            lIncludeFiles = obj.IncludeFiles;
            lSourceStubs = obj.SourceStubs;
            lIncludeStubs = obj.IncludeStubs;
            lIncludeProxies = obj.IncludeProxies;
        end
    end

    methods (Static, Access = public)
        function obj = createFromBlock(hBlock, bReplaceStartDir)
            % CREATEFROMBLOCK
            %
            % Create an instance of the ModelBlockSourcesHelper class from
            % a control block
            % Source files folder
            import coder.internal.xrel.AUTOSAR.ModelBlockSourcesHelper;
            
            lStartDir = eval(get_param(hBlock, 'StartDir'));
            lPathRelativeToAnchor = eval(get_param(hBlock, 'PathRelativeToAnchor'));
            lSourcePaths = eval(get_param(hBlock, 'SourcePaths'));
            lIncludePaths = eval(get_param(hBlock, 'IncludePaths'));
            lSourceFiles = eval(get_param(hBlock, 'SourceFiles'));
            lIncludeFiles = eval(get_param(hBlock, 'IncludeFiles'));
            lSourceStubs = eval(get_param(hBlock, 'SourceStubs'));
            lIncludeStubs = eval(get_param(hBlock, 'IncludeStubs'));
            lIncludeProxies = eval(get_param(hBlock, 'IncludeProxies'));
            
            assert(ischar(lStartDir));
            assert(ischar(lPathRelativeToAnchor));
            assert(iscell(lSourcePaths), 'SourcePaths should be cell');
            assert(iscell(lIncludePaths), 'IncludePaths should be cell');
            assert(iscell(lSourceFiles), 'SourceFiles should be cell');
            assert(iscell(lIncludeFiles), 'IncludeFiles should be cell');
            assert(iscell(lSourceStubs), 'SourceStubs should be cell');
            assert(iscell(lIncludeStubs), 'IncludeStubs should be cell');
            
            if bReplaceStartDir
                % Substitute $(START_DIR)
                lSourcePaths =  ModelBlockSourcesHelper.replaceStartDir(lSourcePaths, lStartDir);
                lIncludePaths =  ModelBlockSourcesHelper.replaceStartDir(lIncludePaths, lStartDir);
            end
            
            obj = ModelBlockSourcesHelper(lStartDir, ...
                lPathRelativeToAnchor, ...
                lSourcePaths, lIncludePaths, ...
                lSourceFiles, lIncludeFiles, ...
                lSourceStubs, lIncludeStubs, ...
                lIncludeProxies);
        end
        
        function obj = createFromArtifact(xrel_info)
            % CREATEFROMARTIFACT
            %
            % Create an instance of the ModelBlockSourcesHelper class from
            % a cross-release artifact.
            
            % Need to recover the anchor dir
            lBuildDir = xrel_info.BuildDir;
            lPathRelativeToAnchor = xrel_info.PathRelativeToAnchor;
            lStartDir = strrep(lBuildDir, lPathRelativeToAnchor, '');
            if endsWith(lStartDir, filesep)
                lStartDir(end) = [];
            end
            lAllSourcePaths = xrel_info.LCTSpecs.SrcPaths;
            lAllIncludePaths = xrel_info.LCTSpecs.IncPaths;
            
            % Substitute $(START_DIR) token back in
            %   - it will help to make relocation easier later on
            lStartDirRegexp = ['^', regexptranslate('escape', lStartDir)];
            lStartDirTokenRegexp = regexptranslate('escape', '$(START_DIR)');
            lAllSourcePaths = regexprep(lAllSourcePaths, ...
                lStartDirRegexp, ...
                lStartDirTokenRegexp);
            lAllIncludePaths = regexprep(lAllIncludePaths, ...
                lStartDirRegexp, ...
                lStartDirTokenRegexp);
            
            lSourcePathIsStub = endsWith(lAllSourcePaths, 'stub');
            lIncludePathIsStub = endsWith(lAllIncludePaths, 'stub');
            
            lSourcePaths = lAllSourcePaths(~lSourcePathIsStub);
            lIncludePaths = lAllIncludePaths(~lIncludePathIsStub);
            
            lSourceFiles = xrel_info.LCTSpecs.SourceFiles;
            lIncludeFiles = xrel_info.LCTSpecs.HeaderFiles;
            
            lSourceStubs = {};
            lIncludeStubs = {};            
            lIncludeProxies = {};
            
            obj = coder.internal.xrel.AUTOSAR.ModelBlockSourcesHelper(lStartDir, ...
                lPathRelativeToAnchor, ...
                lSourcePaths, lIncludePaths, ...
                lSourceFiles, lIncludeFiles, ...
                lSourceStubs, lIncludeStubs, ...
                lIncludeProxies);
        end
    end
    
    methods (Static, Access = private)
        function lStr = cell2Str(lCellStr)
            % CELL2STR
            %
            % Convert cell array of char to a string that we can eval to
            % recreate it.
            lCellStr = strcat('''', lCellStr, '''');
            lStr = ['{', strjoin(lCellStr, ','), '}'];
        end
        
        function lPathCell = replaceStartDir(lPathCell, lStartDir)
            % Replace $(START_DIR) with string in lStartDir for all
            % elements of cellstr lPathCell
            lPathCell = strrep(lPathCell, '$(START_DIR)', lStartDir);
        end
    end
end