classdef XRelArtifactUtils < handle
    % XRELARTIFACTUTILS
    %
    %   Wrapper around cross-release artifact providing search/find
    %   functionality needed when building CodeInfo for AUTOSAR.
    %
    %   Copyright 2017 The MathWorks, Inc.
    %
    properties (Access = private)
        XRelInfo;   % Raw cross-release artifact
    end
    
    methods (Access = public)
        function obj = XRelArtifactUtils(xrel_info)
            % XRELARTIFACTUTILS
            %   Wrap the cross-release artifact
            obj.XRelInfo = xrel_info;
        end
        
        function ret = isExplicitTaskingForRunnable(obj, lRunnableSymbol)
            % ISEXPLICITTASKINGFORRUNNABLE
            % Does this runnable symbol have a function caller input associated
            % with it?
            fhIsTask = @(x)strcmp(x.Symbol, lRunnableSymbol);
            lElem = obj.searchArtifact('OutputFunctions', fhIsTask);
            ret = strcmp(lElem.TaskingMode, 'EXPLICIT_TASKING') || ...
                strcmp(lElem.TimingMode, 'ASYNCHRONOUS');
        end
        
        function ret = getSampleOffsetForRunnable(obj, lRunnableSymbol)
            % GETSAMPLEOFFSETFORRUNNABLE
            fhIsTask = @(x)strcmp(x.Symbol, lRunnableSymbol);
            lElem = obj.searchArtifact('OutputFunctions', fhIsTask);
            assert(~isempty(lElem), 'Could not find runnable with symbol %s', lRunnableSymbol);
            ret = lElem.SampleOffset;
        end
        
        function ret = getPriorityForRunnable(obj, lRunnableSymbol)
            % GETPRIORITYFORRUNNABLE
            fhIsTask = @(x)strcmp(x.Symbol, lRunnableSymbol);
            lElem = obj.searchArtifact('OutputFunctions', fhIsTask);
            assert(~isempty(lElem), 'Could not find runnable with symbol %s', lRunnableSymbol);
            ret = lElem.Priority;
        end
        
        function ret = getInternalCalPrmDimensions(obj, lName)
            % GETINTERNALCALPRMDIMENSIONS
            fh = @(x)isfield(x, 'AutosarPort') && isfield(x, 'DataAccessMode') ...
                && strcmp(x.AutosarPort, lName) && strcmp(x.DataAccessMode, 'InternalCalPrm');
            lElem = obj.searchArtifact('Parameters', fh);
            assert(~isempty(lElem));
            ret = lElem.Dimensions;
        end
        
        function ret = getDataStoreByGraphicalName(obj, lName)
            % GETDATASTOREBYNAME
            fh = @(x)isfield(x, 'GraphicalName') && strcmp(x.GraphicalName, lName);
            ret = obj.searchArtifact('DataStores', fh);
        end
        
        function [lXrelPortStruct, lIdx] = getXRelInfoFromPortAndDataElement(obj, lField, lPort, lDataElement, lDataAccess)
            % GETXRELINFOFROMPORTANDDATALEMENT
            if ~iscell(lDataAccess)
                lDataAccess = {lDataAccess};
            end
            fhIsMatch = @(x)strcmp(x.AutosarPort, lPort) && ...
                strcmp(x.AutosarDataElement, lDataElement) && ...
                ismember(x.DataAccessMode, lDataAccess);
            [lXrelPortStruct, lIdx] = obj.searchArtifact(lField, fhIsMatch);
        end
        
        function lIdx = getExpInportIdxFromGraphicalName(obj, lGraphicalName)
            % GETEXPINPORTIDXFROMGRAPHICALNAME
            fhIsMatch = @(x)strcmp(x.GraphicalName, lGraphicalName);
            [~, lIdx] = obj.searchArtifact('ExternalInportMapping', fhIsMatch);
            assert(~isempty(lIdx), 'Could not find port');
        end
        
        function lInportNames = getDataInportGraphicalNames(obj)
            % GETDATAINPORTGRAPHICALNAMES
            %   Get data input ports (i.e., not the function callers)
            lInports = obj.XRelInfo.Inports;
            if isfield(lInports, 'GraphicalName')
                lInportNames = {lInports.GraphicalName};
            else
                lInportNames = {};
            end
        end
        
        function lOutportNames = getDataOutportGraphicalNames(obj)
            % GETDATAOUTPORTGRAPHICALNAMES
            lOutports = obj.XRelInfo.Outports;
            if isfield(lOutports, 'GraphicalName')
                lOutportNames = {lOutports.GraphicalName};
            else
                lOutportNames = {};
            end
        end
        
        function ret = getNumPorts(obj, lPortType)
            switch lPortType
                case 'Inport'
                    ret = numel(obj.XRelInfo.ExternalInportMapping);
                case 'Outport'
                    ret = numel(obj.XRelInfo.Outports);
                otherwise
                    assert(false, 'lPortType must be Inport or Outport')
            end
        end
        
        function ret = isFunctionCall(obj, lPortType, kPort)
            % ISFUNCTIONCALL
            switch lPortType
                case 'Inport'
                    ret = obj.XRelInfo.ExternalInportMapping(kPort).IsFunctionCall;
                case 'Outport'
                    ret = false;
                otherwise
                    assert(false, 'lPortType must be Inport or Outport')
            end
        end
        
        function ret = getPortSymbol(obj, lPortType, kPort)
            % GETPORTSYMBOL
            assert(strcmp(lPortType, 'Inport'), 'getPortSymbol only for Inports');
            ret = obj.XRelInfo.ExternalInportMapping(kPort).Symbol;
        end
        
        
        function lElem = getOutputFunctionFromInport(obj, lPortType, kPort)
            % GETOUTPUTFUNCTIONFROMINPORT
            assert(strcmp(lPortType, 'Inport'), ...
                'lPortType must be Inport for getting runnable name');
            lSymbol = obj.getPortSymbol(lPortType, kPort);
            
            fh = @(x)strcmp(x.Symbol, lSymbol);
            lElem = obj.searchArtifact('OutputFunctions', fh);
            assert(~isempty(lElem), 'Could not find OutputFunction with Symbol %s', lSymbol);
        end
        
        function lElem = getDataPortElemWithIndirection(obj, lPortType, kPort)
            % GETDATAPORTELEMWITHINDIRECTION
            switch lPortType
                case 'Inport'
                    assert(kPort <= numel(obj.XRelInfo.ExternalInportMapping), ...
                        'Index %d is greater than the number of ExternalInportMapping (%d) in the artifact', ...
                        kPort, numel(obj.XRelInfo.ExternalInportMapping));
                    assert(~obj.XRelInfo.ExternalInportMapping(kPort).IsFunctionCall, 'getAutosarPortMapping does not apply to function call ports');
                    lGraphicalName = obj.XRelInfo.ExternalInportMapping(kPort).GraphicalName;
                    
                    % Look up by graphical name
                    lElem = obj.getInportFromGraphicalName(lGraphicalName);
                case 'Outport'
                    assert(kPort <= numel(obj.XRelInfo.Outports), ...
                        'Index %d is greater than the number of Outports (%d) in the artifact', ...
                        kPort, numel(obj.XRelInfo.Outports));
                    lElem = obj.XRelInfo.Outports(kPort);
                otherwise
                    assert(false, 'lPortType must be Inport or Outport');
            end
        end
        
        function [lPort, lDataElement, lDataAccessMode] = getAutosarPortMapping(obj, lPortType, kPort)
            % GETAUTOSARPORTMAPPING
            lElem = obj.getDataPortElemWithIndirection(lPortType, kPort);
            lPort = lElem.AutosarPort;
            lDataElement = lElem.AutosarDataElement;
            lDataAccessMode = lElem.DataAccessMode;
        end
        
        function [lDimensions, lIsBus] = getDimensions(obj, lPortType, kPort)
            % GETDIMENSIONS
            lElem = obj.getDataPortElemWithIndirection(lPortType, kPort);
            lDimensions = lElem.Dimensions;
            lIsBus = lElem.IsBus;
        end

        
        function lGraphicalName = getGraphicalName(obj, lPortType, kPort)
            % GETGRAPHICALNAME
            switch lPortType
                case 'Inport'
                    assert(kPort <= numel(obj.XRelInfo.ExternalInportMapping), ...
                        'Index %d is greater than the number of ExternalInportMapping (%d) in the artifact', ...
                        kPort, numel(obj.XRelInfo.ExternalInportMapping));
                    lGraphicalName = obj.XRelInfo.ExternalInportMapping(kPort).GraphicalName;
                case 'Outport'
                    assert(kPort <= numel(obj.XRelInfo.Outports), ...
                        'Index %d is greater than the number of Outports (%d) in the artifact', ...
                        kPort, numel(obj.XRelInfo.Outports));
                    lGraphicalName = obj.XRelInfo.Outports(kPort).GraphicalName;
                otherwise
                    assert(false, 'lPortType must be Inport or Outport')
            end
        end
        
        function lInportElem = getInportFromGraphicalName(obj, lGraphicalName)
            % GETINPORTFROMGRAPHICALNAME
            fh = @(x)strcmp(x.GraphicalName, lGraphicalName);
            lInportElem = obj.searchArtifact('Inports', fh);
            assert(~isempty(lInportElem), 'Could not find Inport with GraphicalName %s', lGraphicalName);
        end
        
        function ret = getStructTypes(obj)
            % GETSTRUCTTYPES
            ret = obj.XRelInfo.StructTypes;
        end
        
        function ret = getParameters(obj)
            % GETPARAMETERS
            ret = obj.XRelInfo.Parameters;
        end
        
        function ret = getStructTypeByName(obj, lName)
            % GETSTRUCTTYPEBYNAME
            fh = @(x)isfield(x, 'Name') && strcmp(x.Name, lName);
            ret = obj.searchArtifact('StructTypes', fh);
        end
        
        function [ret, isAlias, isFix, isStruct, isEnum] = getTypeByName(obj, lName)
             % GETTYPEBYNAME
             fh = @(x)isfield(x, 'Name') && strcmp(x.Name, lName);
             
             isAlias = false;
             isFix = false;
             isStruct = false;
             isEnum = false;
             
             ret = obj.searchArtifact('AliasTypes', fh);
             if ~isempty(ret)
                 isAlias = true;
                 return;
             end
             ret = obj.searchArtifact('FixTypes', fh);
             if ~isempty(ret)
                 isFix = true;
                 return;
             end
             ret = obj.searchArtifact('StructTypes', fh);
             if ~isempty(ret)
                 isStruct = true;
                 return;
             end
             ret = obj.searchArtifact('EnumTypes', fh);
             if ~isempty(ret)
                 isEnum = true;
             end
        end
        
        function ret = getStatesDWorkName(obj, lStateType)
            % GETSTATESDWORKNAME
            switch lStateType
                case 'Block states'
                    ret = obj.XRelInfo.BlockStatesDWorkName;
                case 'Block signals'
                    ret = obj.XRelInfo.SignalStatesDWorkName;
                case 'Zero crossing states'
                    ret = obj.XRelInfo.ZCStatesDWorkName;
                otherwise
                    assert(false, 'Invalid lStateType - should be Block states, Block signals or Zero crossing states');
            end
        end
        
        function ret = getStatesDWorkType(obj, lStateType)
            % GETSTATESDWORKTYPE
            switch lStateType
                case 'Block states'
                    ret = obj.XRelInfo.BlockStatesDWorkType;
                case 'Block signals'
                    ret = obj.XRelInfo.SignalStatesDWorkType;
                case 'Zero crossing states'
                    ret = obj.XRelInfo.ZCStatesDWorkType;
                otherwise
                    assert(false, 'Invalid lStateType - should be Block states, Block signals or Zero crossing states');
            end
        end
        
        function ret = getStatesDWorkTag(obj, lStateType)
            % GETSTATESDWORKTAG
            switch lStateType
                case 'Block states'
                    ret = obj.XRelInfo.BlockStatesDWorkTag;
                case 'Block signals'
                    ret = obj.XRelInfo.SignalStatesDWorkTag;
                case 'Zero crossing states'
                    ret = obj.XRelInfo.ZCStatesDWorkTag;
                otherwise
                    assert(false, 'Invalid lStateType - should be Block states, Block signals or Zero crossing states');
            end
        end
        
        function ret = definesSimulinkFunctions(obj)
            % DEFINESSIMULINKFUNCTIONS
            ret = isfield(obj.XRelInfo, 'SimulinkFunctions') && ...
                ~isempty(obj.XRelInfo.SimulinkFunctions) && ...
                any(strcmp({obj.XRelInfo.SimulinkFunctions.IsDefined}, 'yes'));
        end
        
        
        function lPeriodicByFcnCall = hasExplicitPeriodicTasking(obj)
            % HASEXPLICITPERIODICTASKING
            lPeriodicByFcnCall = false;
            % In releases <= R2016a, periodic triggers could be
            % registered with ASYNCHRONOUS timing mode, but would have
            % a sample period > 0.
            for kRunnable = 1:numel(obj.XRelInfo.OutputFunctions)
                lRunnable = obj.XRelInfo.OutputFunctions(kRunnable);
                if strcmp(lRunnable.TimingMode, 'PERIODIC') && strcmp(lRunnable.TaskingMode, 'EXPLICIT_TASKING')
                    lPeriodicByFcnCall = true;
                    break;
                elseif strcmp(lRunnable.TimingMode, 'ASYNCHRONOUS') && (lRunnable.SamplePeriod ~= -1)
                    lPeriodicByFcnCall = true;
                    break;
                end
            end
        end
        
        function ret = getAutosarInitRunnableSymbol(obj)
            % GETAUTOSARINITRUNNABLESYMBOL
            ret = obj.XRelInfo.AutosarInitRunnableSymbol;
        end
    end
    
    methods (Access = private)
        function [ret, idx] = searchArtifact(obj, lField, fhCriterion)
            % SEARCHARTIFACT
            %   Private utility method for searching through the fields of
            %   the cross-release artifact.
            ret = [];
            idx = 0;
            nElem = numel(obj.XRelInfo.(lField));
            for kElem = 1:nElem
                lCurrElem = obj.XRelInfo.(lField)(kElem);
                if fhCriterion(lCurrElem)
                    ret = lCurrElem;
                    idx = kElem;
                    break;
                end
            end
        end
    end
end