% AUTOSARMODELVALIDATION
%
% Copyright 2017 The MathWorks, Inc.

classdef (Hidden = true) AUTOSARModelValidation < coder.internal.xrelexport.ValidationStrategy
    
    properties (GetAccess = public)
        CodeInterface = coder.internal.xrelexport.CodeInterface.AUTOSAR;
        ValidatesModelRefDir = false;
    end
    
    properties (Constant, Access = protected)
        GlobalIOErrorID = 'Coder:xrelexport:RequiresFunctionPrototypeControl';
    end
    
    methods (Access = public)
        function obj = AUTOSARModelValidation
            % empty constructor
        end
        
        function validate(obj, XrelMsg, config_set, CodeInfoWrapper, InfoStructWrapper)
            % do the initial part of the common validation
            obj.doInitialCommonValidation(XrelMsg, config_set, CodeInfoWrapper, InfoStructWrapper);

            % Can't support obsolete Server Operation modeling style
            if ~isempty(CodeInfoWrapper.getRealTimeModelDW)
                XrelMsg.error('Coder:xrelexport:DoesNotSupportReusableServerOperation');
            end
            
            % Cannot support error status ports
            % Cannot use arrayfun in R2010a for our MCOS classes
            lAllPorts = [CodeInfoWrapper.Outports, CodeInfoWrapper.Inports];
            lHasAUTOSARErrorStatus = false;
            fhTest = @(x)x.AUTOSARData.IsAutosarErrorStatus;
            for kPort = 1:numel(lAllPorts)
                if fhTest(lAllPorts(kPort))
                    lHasAUTOSARErrorStatus = true;
                    break;
                end
            end
            if lHasAUTOSARErrorStatus
                XrelMsg.error('Coder:xrelexport:DoesNotSupportAUTOSARErrorStatus');
            end
            
            % Cannot support model refs with autosar
            if ~isempty(InfoStructWrapper.ModelRefsAll)
                XrelMsg.error('Coder:xrelexport:DoesNotSupportAUTOSARModelRefs');
            end
            
            % Cannot support per-instance memories that need NVRAM Access
            if ~isempty(InfoStructWrapper.GeneralDataFromTLC)
                lGeneralDataFromTLC = InfoStructWrapper.GeneralDataFromTLC;
                if isfield(lGeneralDataFromTLC, 'AutosarPIMDSM')
                    lAutosarPIMDSM = lGeneralDataFromTLC.AutosarPIMDSM;
                    if ~isempty(lAutosarPIMDSM)
                        if iscell(lAutosarPIMDSM)
                            % Re-interpret as struct array
                            lAutosarPIMDSM = [lAutosarPIMDSM{:}];
                        end
                        if any([lAutosarPIMDSM.NeedsNVRAMAccess])
                            XrelMsg.error('Coder:xrelexport:DoesNotSupportAUTOSARPerInstanceMemoriesNVRAM');
                        end
                    end
                end
            end
            
            % Cannot support lookup tables
            if CodeInfoWrapper.HasAutosarLookupTables
                XrelMsg.error('Coder:xrelexport:DoesNotSupportAUTOSARLookupTables');
            end
            
            % Cannot support struct elements of matrix type
            if CodeInfoWrapper.HasMatrixStructElements
                % XXX
               % XrelMsg.error('Coder:xrelexport:DoesNotSupportAUTOSARMatrixStructElements');
            end
            
            % Cannot support QueuedExplicitSend or QueuedExplicitReceive
            % when actual code does not queue the data.
            lInconsistentQueued = false;
            fhTest = @(x)x.AUTOSARData.IsQueued && ~x.IsQueued;
            for kPort = 1:numel(lAllPorts)
                if fhTest(lAllPorts(kPort))
                    lInconsistentQueued = true;
                    break;
                end
            end
            if lInconsistentQueued
                XrelMsg.error('Coder:xrelexport:DoesNotSupportNonQueuedDataForQueuedDataElement');
            end
            
            % do the final part of the common validation
            obj.doFinalCommonValidation(XrelMsg, config_set, ...
                CodeInfoWrapper);
            
        end
        
    end
end
